/*
 *************************************************************************************
 *                                                                                   *
 * FRNT_END.H - the 'C' header file for the Sangoma S508/S514 adapter front-end API. *
 *                                                                                   *
 *************************************************************************************
*/
#ifndef _SDLA_FRONT_END_H_
#define _SDLA_FRONT_END_H_

#ifndef __WINDOWS__
# ifndef	PACKED
# define	PACKED __attribute__((packed))
# endif	/* PACKED */
#else
# undef PACKED
# define PACKED
#endif

/*
*************************************************************************
*			  DEFINES AND MACROS				*	
*************************************************************************
*/
/* front-end UDP command */
#define WAN_FE_GET_STAT			(WAN_FE_UDP_CMD_START + 0)
#define WAN_FE_SET_LB_MODE		(WAN_FE_UDP_CMD_START + 1)
#define WAN_FE_FLUSH_PMON		(WAN_FE_UDP_CMD_START + 2)
#define WAN_FE_GET_CFG			(WAN_FE_UDP_CMD_START + 3)

#define IS_TE1_56K(te_cfg)	(te_cfg.media == WANOPT_MEDIA_T1 || \
				 te_cfg.media == WANOPT_MEDIA_E1 || \
				 te_cfg.media == WANOPT_MEDIA_56K)?1:0

#ifdef WAN_KERNEL

#define FE_ASSERT(val)	if (val) return;
#define FE_ASSERT1(val)	if (val) return 1;

/* front-end configuration and access interface commands */
#define READ_FRONT_END_REGISTER		(WAN_FE_CMD_START+0)	/* 0x90 read from front-end register */
#define WRITE_FRONT_END_REGISTER	(WAN_FE_CMD_START+1)	/* 0x91 write to front-end register */
#define READ_FRONT_END_STATISTICS	(WAN_FE_CMD_START+2)	/* 0x92 read the front-end statistics */
#define FLUSH_FRONT_END_STATISTICS	(WAN_FE_CMD_START+3)	/* 0x93 flush the front-end statistics */


/* adapter configuration interface commands */
#define SET_ADAPTER_CONFIGURATION	(WAN_INTERFACE_CMD_START+0)	/* 0xA0 set adapter configuration */
#define READ_ADAPTER_CONFIGURATION	(WAN_INTERFACE_CMD_START+1)	/* 0xA1 read adapter configuration */

/* front-end command */
#define WAN_FE_GET_STAT			(WAN_FE_UDP_CMD_START + 0)
#define WAN_FE_SET_LB_MODE		(WAN_FE_UDP_CMD_START + 1)
#define WAN_FE_FLUSH_PMON		(WAN_FE_UDP_CMD_START + 2)
#define WAN_FE_GET_CFG			(WAN_FE_UDP_CMD_START + 3)

/* return codes from interface commands */
#define LGTH_FE_CFG_DATA_INVALID       0x91 /* the length of the FE_RX_DISC_TX_IDLE_STRUCT is invalid */
#define LGTH_ADAPTER_CFG_DATA_INVALID  0x91 /* the length of the passed configuration data is invalid */
#define INVALID_FE_CFG_DATA            0x92 /* the passed SET_FE_RX_DISC_TX_IDLE_CFG data is invalid */
#define ADPTR_OPERATING_FREQ_INVALID   0x92 /* an invalid adapter operating frequency was selected */
#define PROT_CFG_BEFORE_FE_CFG         0x93 /* set the protocol-level configuration before setting the FE configuration */

#define SET_FE_RX_DISC_TX_IDLE_CFG      0x98 /* set the front-end Rx discard/Tx idle configuration */
#define READ_FE_RX_DISC_TX_IDLE_CFG     0x99 /* read the front-end Rx discard/Tx idle configuration */
#define SET_TE1_SIGNALING_CFG		0x9A /* set the T1/E1 signaling configuration */
#define READ_TE1_SIGNALING_CFG	0x9B /* read the T1/E1 signaling configuration */


#define COMMAND_INVALID_FOR_ADAPTER    0x9F /* the command is invalid for the adapter type */

 
/* ---------------------------------------------------------------------------------
 * Constants for the SET_FE_RX_DISC_TX_IDLE_CFG/READ_FE_RX_DISC_TX_IDLE_CFG commands
 * --------------------------------------------------------------------------------*/

#define NO_ACTIVE_RX_TIME_SLOTS_T1   24 /* T1 - no active time slots used for reception */
#define NO_ACTIVE_TX_TIME_SLOTS_T1   24 /* T1 - no active time slots used for transmission */
#define NO_ACTIVE_RX_TIME_SLOTS_E1   32 /* E1 - no active time slots used for reception */
#define NO_ACTIVE_TX_TIME_SLOTS_E1   31 /* E1 - no active time slots used for transmission (channel 0 reserved for framing) */

/* Read/Write to front-end register */
#define READ_REG(reg)		card->wandev.read_front_end_reg(card, reg)
#define WRITE_REG(reg, value)	card->wandev.write_front_end_reg(card, reg, (unsigned char)(value))

/* the structure used for the SET_FE_RX_DISC_TX_IDLE_CFG/READ_FE_RX_DISC_TX_IDLE_CFG command */
typedef struct {
        unsigned short lgth_Rx_disc_bfr PACKED; /* the length of the Rx discard buffer */
        unsigned short lgth_Tx_idle_bfr PACKED; /* the length of the Tx idle buffer */
                                                /* the transmit idle data buffer */
        unsigned char Tx_idle_data_bfr[NO_ACTIVE_TX_TIME_SLOTS_E1] PACKED;
} FE_RX_DISC_TX_IDLE_STRUCT;
                                         

/* ----------------------------------------------------------------------------
 *                       Constants for front-end access
 * --------------------------------------------------------------------------*/

/* the structure used for the READ_FRONT_END_REGISTER/WRITE_FRONT_END_REGISTER command */
typedef struct {
	unsigned short register_number PACKED; /* the register number to be read from or written to */
	unsigned char register_value PACKED;	/* the register value read/written */
} FRONT_END_REG_STRUCT;

typedef struct {
	unsigned char opp_flag PACKED;	/* opp flag */
	
	union {
		struct {
			unsigned char RR8_56k PACKED;						/* register #8 value - 56K CSU/DSU */
			unsigned char RR9_56k PACKED;						/* register #9 value - 56K CSU/DSU */
			unsigned char RRA_56k PACKED;						/* register #A value - 56K CSU/DSU */	
			unsigned char RRB_56k PACKED;						/* register #B value - 56K CSU/DSU */
			unsigned char RRC_56k PACKED;						/* register #C value - 56K CSU/DSU */
		} stat_56k;
	} FE_U;

} FRONT_END_STATUS_STRUCT;


/* -----------------------------------------------------------------------------
 *            Constants for the READ_FRONT_END_STATISTICS command
 * ---------------------------------------------------------------------------*/

/* the front-end statistics structure */
typedef struct {
	unsigned long FE_interrupt_count PACKED;   /* the number of front-end interrupts generated */
	unsigned long FE_app_timeout_count PACKED; /* the number of front-end interrupt application timeouts */
} FE_STATISTICS_STRUCT;



/* --------------------------------------------------------------------------------
 * Constants for the SET_ADAPTER_CONFIGURATION/READ_ADAPTER_CONFIGURATION commands
 * -------------------------------------------------------------------------------*/

/* the adapter configuration structure */
typedef struct {
	unsigned short adapter_type PACKED;	/* type of adapter */
	unsigned short adapter_config PACKED;	/* miscellaneous adapter configuration options */
	unsigned long operating_frequency PACKED;	/* adapter operating frequency */
} ADAPTER_CONFIGURATION_STRUCT;



typedef unsigned char (WRITE_FRONT_END_REG_T)(void*, unsigned short, unsigned char);
typedef unsigned char (READ_FRONT_END_REG_T)(void*, unsigned short);


/* settings for the 'adapter_type' */
#define S5141_ADPTR_1_CPU_SERIAL	0x0011	/* S5141, single CPU, serial */
#define S5142_ADPTR_2_CPU_SERIAL	0x0012	/* S5142, dual CPU, serial */
#define S5143_ADPTR_1_CPU_FT1		0x0013	/* S5143, single CPU, FT1 */
#define S5144_ADPTR_1_CPU_T1E1		0x0014	/* S5144, single CPU, T1/E1 */
#define S5145_ADPTR_1_CPU_56K		0x0015	/* S5145, single CPU, 56K */
#define S5147_ADPTR_2_CPU_T1E1		0x0017  /* S5147, dual CPU, T1/E1 */
#define S5148_ADPTR_1_CPU_T1E1		0x0018	/* S5148, single CPU, T1/E1 */
#define S518_ADPTR_1_CPU_ADSL		0x0018	/* S518, adsl card */
#define S519_ADPTR_XILINX		0x0019	/* S519, XILINX card */

/*
** Sangoma Front-End interface structure 
*/
typedef struct {
	unsigned long	(*get_fe_service_status)(void*);	/* In-Service or Not (T1/E1/56K) */
	void		(*print_fe_alarm)(void*,unsigned long);	/* Print Front-End alarm (T1/E1/56K) */
	char*	(*print_fe_act_channels)(void*);	/* Print Front-End alarm (T1/E1/56K) */
} sdla_fe_iface_t;


#endif	/* WAN_KERNEL */

#endif
