/*
 * Copyright (c) 1997, 1998, 1999
 *	Alex Feldman <al.feldman@sangoma.com>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Alex Feldman.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Alex Feldman AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Alex Feldman OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: wanrouter.h,v 1.13 2003/10/01 22:04:39 sangoma Exp $
 */

/*****************************************************************************
 * wanrouter.h	Definitions for the WAN Multiprotocol Router Module.
 *		This module provides API and common services for WAN Link
 *		Drivers and is completely hardware-independent.
 *
 * Author: 	Alex Feldman <al.feldman@sangoma.com>
 *		Nenad Corbic <ncorbic@sangoma.com>
 *		Gideon Hack 	
 *
 * ============================================================================
 * Dec 03, 1999	Alex Feldman	Updated to FreeBSD 3.2 
 * Jan 02, 1997	Gene Kozin	Initial version (based on wanpipe.h).
 ****************************************************************************
 */

#ifndef	_WANROUTER_H
#define	_WANROUTER_H

#include <net/wanpipe_version.h>
#include <net/wanpipe_defines.h>
#include <net/wanpipe_abstr.h>
#include <net/sdla_te1.h>
#include <net/sdla_56k.h>
#include <net/wanpipe_cfg.h>

#define	ROUTER_NAME	"wanrouter"	/* in case we ever change it */
#define	ROUTER_IOCTL	'W'		/* for IOCTL calls */
#define	ROUTER_MAGIC	0x524D4157L	/* signature: 'WANR' reversed */

#define WANROUTER_MAJOR_VER 2
#define WANROUTER_MINOR_VER 1

/* NLPID for packet encapsulation (ISO/IEC TR 9577) */
#define	NLPID_IP	0xCC	/* Internet Protocol Datagram */
#define	NLPID_SNAP	0x80	/* IEEE Subnetwork Access Protocol */
#define	NLPID_CLNP	0x81	/* ISO/IEC 8473 */
#define	NLPID_ESIS	0x82	/* ISO/IEC 9542 */
#define	NLPID_ISIS	0x83	/* ISO/IEC ISIS */
#define	NLPID_Q933	0x08	/* CCITT Q.933 */

/* Some usefull typedefs */
typedef char devname_t[WAN_DRVNAME_SZ+1];

/* Maximum size of rx/tx packets */
#define MAX_PACKET_SIZE 5000

/****** Data Types **********************************************************/

/*----------------------------------------------------------------------------
 * WAN Link Status Info (for ROUTER_STAT IOCTL).
 */
typedef struct wandev_stat
{
	unsigned state;		/* link state */
	unsigned ndev;		/* number of configured interfaces */

	/* link/interface configuration */
	unsigned connection;	/* permanent/switched/on-demand */
	unsigned media_type;	/* Frame relay/PPP/X.25/SDLC, etc. */
	unsigned mtu;		/* max. transmit unit for this device */

	/* physical level statistics */
	unsigned modem_status;	/* modem status */
	unsigned rx_frames;	/* received frames count */
	unsigned rx_overruns;	/* receiver overrun error count */
	unsigned rx_crc_err;	/* receive CRC error count */
	unsigned rx_aborts;	/* received aborted frames count */
	unsigned rx_bad_length;	/* unexpetedly long/short frames count */
	unsigned rx_dropped;	/* frames discarded at device level */
	unsigned tx_frames;	/* transmitted frames count */
	unsigned tx_underruns;	/* aborted transmissions (underruns) count */
	unsigned tx_timeouts;	/* transmission timeouts */
	unsigned tx_rejects;	/* other transmit errors */

	/* media level statistics */
	unsigned rx_bad_format;	/* frames with invalid format */
	unsigned rx_bad_addr;	/* frames with invalid media address */
	unsigned tx_retries;	/* frames re-transmitted */
	unsigned reserved[16];	/* reserved for future use */
} wandev_stat_t;

/* Front-End status */
enum fe_status {
        FE_UNITIALIZED = 0x00,
        FE_DISCONNECTED,
        FE_CONNECTED
};

/* 'modem_status' masks */
#define	WAN_MODEM_CTS	0x0001	/* CTS line active */
#define	WAN_MODEM_DCD	0x0002	/* DCD line active */
#define	WAN_MODEM_DTR	0x0010	/* DTR line active */
#define	WAN_MODEM_RTS	0x0020	/* RTS line active */

/* modem status changes */
#define WAN_DCD_HIGH			0x08
#define WAN_CTS_HIGH			0x20

typedef struct wan_conf
{
    devname_t devname;
    void      *arg;
} wan_conf_t;

/* IOCTL codes for /dev/wanrouter entries (up to255) */
#define ROUTER_SETUP    _IOW(ROUTER_IOCTL, 1, wan_conf_t) /* configure device  */
#define ROUTER_DOWN     _IOWR(ROUTER_IOCTL, 2, wan_conf_t)/* shut down device  */
#define ROUTER_STAT     _IOW(ROUTER_IOCTL, 3, wan_conf_t) /* get device status */
#define ROUTER_IFNEW    _IOW(ROUTER_IOCTL, 4, wan_conf_t) /* add interface     */
#define ROUTER_IFDEL    _IOW(ROUTER_IOCTL, 5, wan_conf_t) /* del interface     */
#define ROUTER_IFSTAT   _IOW(ROUTER_IOCTL, 6, wan_conf_t) /* get interface status */
#define ROUTER_DEBUGGING _IOW(ROUTER_IOCTL, 7, wan_conf_t)/* get interface status */
#define ROUTER_VER 	_IOWR(ROUTER_IOCTL, 8, wan_conf_t) /* get router version */
#define ROUTER_PROCFS 	_IOWR(ROUTER_IOCTL, 9, wan_conf_t) /* get procfs info */
/* ALEX_DEBUG */
#define ROUTER_DEBUG_READ 	_IOWR(ROUTER_IOCTL, 10, wan_conf_t) /* get dbg_msg */
#define ROUTER_USER     _IOW(ROUTER_IOCTL, 16, u_int)     /* driver specific calls */
#define ROUTER_USER_MAX _IOW(ROUTER_IOCTL, 31, u_int)

#define SIOC_WANPIPE_MONITOR  _IOWR('i', 150, struct ifreq) /* get monitor statistics */

#ifdef WAN_KERNEL
/****** Kernel Interface ****************************************************/

/*
 * Ethernet statistics collection data
 */ 
struct net_device_stats 
{
	unsigned long rx_packets;	/* total packets received    */
	unsigned long tx_packets;	/* total packets transmited  */
	unsigned long rx_bytes;		/* total bytes received    */
	unsigned long tx_bytes;		/* total bytes transmited  */
	unsigned long rx_errors;	/* bad packet received       */      
	unsigned long tx_errors;	/* packet transmit problems  */
	unsigned long rx_dropped;	/* no space in buffers       */
	unsigned long tx_dropped;	/* no space available        */
	unsigned long multicast;	/* multicast packet received */
	unsigned long collisions;   

	/* detailed rx_errors */
	unsigned long rx_length_errors;
	unsigned long rx_over_errors;	/* receiver ring off buff overflow */
	unsigned long rx_crc_errors;	/* recv'd pkt with crc error       */
	unsigned long rx_frame_errors;	/* recv'd frame alignment error    */
	unsigned long rx_fifo_errors;	/* recv'r fifo overrun             */
	unsigned long rx_missed_errors;	/* receiver missed packet          */

	/* detailed tx_errors */
	unsigned long tx_aborted_errors;
	unsigned long tx_carrier_errors;
	unsigned long tx_fifo_errors;
	unsigned long tx_heartbeat_errors;
	unsigned long tx_window_errors;

	/* for cslip etc */
	unsigned long	rx_compressed;
	unsigned long	tx_compressed;
};

struct ifnet_list
{
    struct ifnet *ifp;
    struct ifnet_list *pnext;
};

/*----------------------------------------------------------------------------
 * WAN device data space.
 */
typedef struct wan_device
{
	unsigned		magic;		/* magic number */
	char* 			name;		/* -> WAN device name (ASCIIZ) */
	void* 			private;	/* -> driver private data */
	int 			config_id;	/* Configuration ID */
				
	/****** hardware configuration ******/
	char 			S514_cpu_no;	/* PCI CPU Number */
	caddr_t 		maddr;		/* dual-port memory address */
#ifdef __OpenBSD__
#if 0
	bus_space_handle_t	memh;		/* dual-port memory address handle */
	bus_space_tag_t		memt;		/* dual-port memory address tag    */
	bus_space_handle_t	ioh;		/* adapter I/O port handle */
	bus_space_tag_t		iot;		/* adapter I/O port tag	   */
#endif
#endif
	int 			irq;		/* interrupt request level */
	int 			dma;		/* DMA request level */
	unsigned 		bps;		/* data transfer rate */
	unsigned 		mtu;		/* max physical transmit unit size */
	unsigned 		udp_port;	/* UDP port for management */
        unsigned char 		ttl;		/* Time To Live for UDP security */
	unsigned 		enable_tx_int; 	/* Transmit Interrupt enabled or not */
	int 			new_if_cnt;	/* Number of opened if per device */
	int 			del_if_cnt;	/* Number of deleted if per device */
	char 			interface;	/* RS-232/V.35, etc. */
	char 			clocking;	/* external/internal */
	char 			line_coding;	/* NRZ/NRZI/FM0/FM1, etc. */
	char 			station;	/* DTE/DCE, primary/secondary, etc. */
	char 			connection;	/* permanent/switched/on-demand */
	char 			signalling;	/* Signalling RS232 or V35 */
	char 			read_mode;	/* read mode: Polling or interrupt */
	unsigned char 		piggyback;      /* Piggibacking a port */
	wan_spinlock_t		lock;		/* Spin lock */ 
	/****** status and statistics *******/
	char 			state;		/* device state */
	char 			api_status;	/* device api status */
	struct net_device_stats	stats;	/* interface statistics */
	unsigned 		reserved[16];		/* reserved for future use */
	/*unsigned char critical;*/		/* critical section flag */
	/****** device management methods ***/
	int (*setup) (struct wan_device *wandev, wandev_conf_t *conf);
	int (*shutdown) (struct wan_device *wandev, wandev_conf_t* conf);
	int (*update) (struct wan_device *wandev);
	int (*ioctl) (struct wan_device *wandev, u_long cmd, caddr_t arg);
	int (*new_if) (struct wan_device *wandev, struct ifnet *ifp,
		wanif_conf_t *conf);
	int (*del_if) (struct wan_device *wandev, struct ifnet *ifp);
					/****** maintained by the router ****/
	struct wan_device* 	next;		/* -> next device */
	struct ifnet_list*	if_head;	/* network interface */
	unsigned 		ndev;		/* number of interfaces */

	sdla_te_cfg_t		te_cfg;		/* TE1 hardware configuration */
	unsigned long 		te_alarm;	/* TE1 alarm */
	unsigned long 		cur_te_alarm;	/* TE1 alarm (proc fs) */
	pmc_pmon_t		te_pmon;	/* TE PMON counters */
	unsigned char 		te_lb_cmd;	/* Received loopback command */	
	unsigned long 		te_lb_time;	/* Time when loopback command received */
	unsigned char 		te_lb_tx_cmd;	/* Received loopback command */	
	unsigned long 		te_lb_tx_cnt;	/* Time when loopback command received */
	unsigned char 		te_critical;	/* TE1 T1/E1 critical flag */	
 	unsigned char		te_timer_cmd;	/* TE1 T1/E1 timer command */
	unsigned long   	k56_alarm;	/* 56K alarms */
	wan_timer_info_t	te_timer;
	unsigned char 		front_end_status;
	unsigned char 		RLOS_56k;
	unsigned char 		RR8_reg_56k;
	unsigned char 		RRA_reg_56k;
	unsigned char 		RRC_reg_56k;
	unsigned char 		prev_RRC_reg_56k;
	unsigned char 		delta_RRC_reg_56k;
        WRITE_FRONT_END_REG_T* 	write_front_end_reg;
	READ_FRONT_END_REG_T* 	read_front_end_reg;
	void (*te_enable_timer)	(void* card_id);	/* TE1 Enable hardware timer */
	void (*te_report_rbsbits) (void* card_id, int channel, unsigned char rbsbits);
	void (*te_report_alarms) (void* card_id, unsigned long alarams);
	void (*te_link_state) (void* card_id);

	unsigned char		card_type;
	int (*debugging) (struct wan_device *wandev);
	int (*debug_read) (void *, void*);
	
} wan_device_t;

/* Public functions available for device drivers */
extern int register_wan_device(wan_device_t *wandev);
extern int unregister_wan_device(char *name);

/* Proc interface functions. These must not be called by the drivers! */
extern int wanrouter_proc_add(wan_device_t *wandev);
extern int wanrouter_proc_delete(wan_device_t *wandev);

/* extern int wanrouter_ioctl( struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg); */

/* Public Data */
extern wan_device_t *router_devlist;	/* list of registered devices */

/* Miscellaneous utilities function */
extern char* in_ntoa(uint32_t in);

#endif	/* KERNEL || _KERNEL */
#endif	/* _WANROUTER_H */
