/*
 * Copyright (c) 1997, 1998, 1999
 *	Alex Feldman <al.feldman@sangoma.com>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Alex Feldman.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Alex Feldman AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Alex Feldman OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: sdla_isa.c.OpenBSD,v 1.5 2003/08/01 14:55:07 sangoma Exp $
 */

/*****************************************************************************
 * sdla_isa.c	SDLA ISA Support Module.  Main module.
 *
 *		This module is a compeled with kernel sources.
 *
 * Author:	Alex Feldman	<al.feldman@sangoma.com>
 ****************************************************************************
 */

#undef SDLA_DEV_DEBUG	/* uncomment this for debug messages */
#undef INTR_TEST	/* uncomment this to check that driver RX INTR */

/*#include "sdla.h"*/		/* number of Sangoma cards */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/syslog.h>
#include <sys/errno.h>
#include <sys/mbuf.h>
#include <sys/proc.h>
#include <sys/sockio.h>
#include <sys/socket.h>
#include <sys/kernel.h>
#include <sys/device.h>
#include <machine/bus.h>

#include <net/sdlasfm.h>   
#include <net/sdlapci.h>   
#include <net/sdla_dev.h>  /* special header for S508 board */

/* hardware related constant */
#define MAX_CARDS         8
#define S508_NUMPORTS     2
#define S508_ID_BYTE      0x11
#define S508_IO_PORTS	  4

/**************************************************************************/
/* SDLA ISA device related entry point 					  */
int 	sdla_isa_match	__P((struct device*, void*, void*));
void 	sdla_isa_attach __P((struct device*, struct device*, void*));
int 	sdla_isa_intr(void*);

struct cfattach sdla_isa_ca = {
	sizeof(struct sdla_softc), sdla_isa_match, sdla_isa_attach
};

/* For debugging convenience */
#ifdef SDLA_ISA_DEBUG
 #define dlog(x) log x
#else
 #define dlog(x)
#endif /* SDLA_ISA_DEBUG */

/******* ISA Device Driver Entry Points ********************************/

int sdla_isa_match(parent, match, aux)
	struct device*	parent;
	void*	match, *aux;
{
	struct isa_attach_args*	ia = aux;
	bus_space_tag_t		iot = ia->ia_iot;
	bus_space_handle_t		ioh;
	int 			ret = 1;

	if (ia->ia_irq == -1){	/* IRQ default */
		return 0;
	}
	if (ia->ia_iobase == -1){	/* I/O port default */
		return 0;
	}
	/* Make sure this is a valid I/O address */

	if (!sdla_probe(SDLA_S508)){
		return 0;
	}

	/* Map I/O space */
	if (bus_space_map(iot, ia->ia_iobase, S508_IO_PORTS, 0, &ioh)){
		return 0;
	}

	if (bus_space_read_1(iot, ioh, SDLA_S508_OFFSET) != S508_ID_BYTE){
		ret = 0;
	}

	bus_space_unmap(iot, ioh, S508_IO_PORTS);
	return ret;
}

/*
 **************************************************************************
 * sdla_isa_attach() - called if the probe is succeeded. Attaches wanpipe?
 *                    device to the stack.
 */
void sdla_isa_attach(parent, self, aux)
	struct device* parent, *self;
	void*	aux;
{
	struct sdla_softc*	sc = (struct sdla_softc*)self;
	sdladev_t*		isa_cfg = NULL;
	struct isa_attach_args*	ia = aux;
	bus_space_handle_t	ioh, memh;
	int			memsize;
	
	printf("\n");
	sc->ia		= *ia;
	
	/* Map I/O space */
	if (bus_space_map(ia->ia_iot, ia->ia_iobase, S508_IO_PORTS, 0, &ioh)){
		printf("sdla?: can't map i/o space!\n");
		return;
	}
	memsize = 0x2000;
	if (bus_space_map(ia->ia_memt, ia->ia_maddr, memsize, 0, &memh)){
		printf("sdla?: can't map memory address!\n");
		return;
	}

	/* Attach sdla_softc to our structure */
	isa_cfg = sdla_attach(sc, 0x0);
	isa_cfg->type 		= SDLA_S508;
	isa_cfg->adapter_type	= 0x00;
	isa_cfg->vendor_id	= V3_VENDOR_ID;
	isa_cfg->isr		= NULL;
	isa_cfg->arg		= NULL;
	isa_cfg->irq 		= ia->ia_irq;
	isa_cfg->u.isa.ioport 	= ia->ia_iobase;
	isa_cfg->u.isa.memh 	= memh;
	isa_cfg->u.isa.ioh 	= ioh;
#if 0
	isa_cfg->u.isa.maddr 	= (caddr_t)ia->ia_maddr;
	isa_cfg->u.isa.memt 	= ia->ia_memt;
	isa_cfg->u.isa.ioport 	= ia->ia_iobase;
	isa_cfg->u.isa.iot 	= ia->ia_iot;
	isa_cfg->irq 		= ia->ia_irq;
#endif	
	/* Establish the interrupt handler */
	sc->sc_ih = 
		isa_intr_establish(ia->ia_ic, ia->ia_irq, IST_EDGE, IPL_NET,
				   sdla_isa_intr, isa_cfg, "sdla");
	if (sc->sc_ih == NULL){
		printf("sdla_isa: couldn't establish interrupt handler\n"); 
    		return;
	}
}

/*
 ****************************************************************************
 * sdla_isa_intr() - Dummy ISA device interrupt handler.
 */
int sdla_isa_intr(void* arg)
{
	sdladev_t* 		isa_cfg = (sdladev_t*)arg;
	struct sdla_softc* 	sc = NULL;
 
#ifdef INTR_TEST
	log(LOG_INFO, "%s: Interrupt from S508 Adapter!\n", SDLA_DEV_NAME); 
#endif

	if (isa_cfg == NULL || isa_cfg->type != SDLA_S508){
		return 0;
	}

	sc = isa_cfg->sc;
	if (sc == NULL){
		log(LOG_ERR, "%s: Uncorrect addapter type!\n", SDLA_DEV_NAME); 
		return 0; 
	}
	if (isa_cfg->isr && (isa_cfg->arg != NULL)){
		isa_cfg->isr(isa_cfg->arg);
	}
	return 1;
}

/****** End *****************************************************************/
