/*
 * Copyright (c) 1997, 1998, 1999
 *	Alex Feldman <al.feldman@sangoma.com>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Alex Feldman.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Alex Feldman AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Alex Feldman OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: sdla_chdlc.c,v 1.47 2003/10/14 20:46:49 sangoma Exp $
 */

/*****************************************************************************
 * sdla_chdlc.c	WANPIPE(tm) Multiprotocol WAN Link Driver. Cisco HDLC module.
 *
 * Author: 	Alex Feldman  <al.feldman@sangoma.com>
 *
 * ============================================================================
 * Sep 18, 2003 Alex Feldman		o Fixed ALTQ bug (IFQ_ENQUEUE)
 * Nov 28, 2001 Alex Feldman		o Added T1/E1 support (TE1).
 * Mar 10, 2000 Alex Feldman RX_CORR	o If receive buffer corrupted in
 *					  rx_intr(), disable interrupt!!!
 * Dec 02, 1999 Alex Feldman		o Updated to FreeBSD 3.2
 * Aug 07, 1998	David Fong		o Initial version.
 ****************************************************************************
 */
/*
****************************************************************************
****		For Debug purpose (only OpenBSD)			****
****************************************************************************
*/

#ifdef __OpenBSD__
# undef WANDEBUG_CHDLC	/* Uncomment this line for Debug output */
# undef ENABLE_IPV6	/* Uncomment this line for Enabling IPv6 support */
#endif

#include <net/wanpipe_config.h>
#include <net/wanpipe_includes.h>

#define _GNUC_
#include <net/wanpipe.h>		/* WANPIPE common user API definision */
#include <net/sdla_chdlc.h>		/* CHDLC firmware API definitions */
/****** Defines & Macros ****************************************************/

#ifdef	_DEBUG_
#define	STATIC
#define WANDEBUG_CHDLC
#else
#define	STATIC		static
#endif

/* reasons for enabling the timer interrupt on the adapter */
#define TMR_INT_ENABLED_UDP     0x0001
#define TMR_INT_ENABLED_UPDATE  0x0002
#define TMR_INT_ENABLED_TE  	0x0004
#define TMR_INT_ENABLED_CONFIG 	0x0008
#define TMR_INT_ENABLED_TEST 	0x0010

#define	CHDLC_DFLT_DATA_LEN	1500		/* default MTU */
#define CHDLC_HDR_LEN		1

#define IFF_POINTTOPOINT 0x10


#define	CHDLC_DFLT_DATA_LEN	1500		/* default MTU */
#define CHDLC_HDR_LEN		1

#define IFF_POINTTOPOINT 0x10

#define WANPIPE   0x00
#define API       0x01
#define CHDLC_API 0x01

#define PORT(x)  (x == 0 ? "PRIMARY" : "SECONDARY")

#define intr_permis flags.interrupt_info_struct.interrupt_permission
#define intr_type flags.interrupt_info_struct.interrupt_type

/******Data Structures*****************************************************/

/* This structure is placed in the private data area of the device structure.
 * The card structure used to occupy the private area but now the following 
 * structure will incorporate the card structure along with CHDLC specific data
 */

typedef struct chdlc_private_area
{
	wanpipe_common_t	common;
	sdla_t		*card;
	int 		TracingEnabled;		/* For enabling Tracing */
	unsigned long 	curr_trace_addr;	/* Used for Tracing */
	unsigned long 	start_trace_addr;
	unsigned long 	end_trace_addr;
	unsigned long 	base_addr_trace_buffer;
	unsigned long 	end_addr_trace_buffer;
	unsigned short 	number_trace_elements;
	unsigned  	available_buffer_space;

	unsigned long 	router_start_time;
	unsigned char 	route_status;
	unsigned char 	route_removed;
	unsigned long 	tick_counter;		/* For 5s timeout counter */
	unsigned long 	router_up_time;
	uint32_t        IP_address;
	uint32_t        IP_netmask;
	unsigned char   mc;			/* Multicast support on/off */
	unsigned 	short udp_pkt_lgth;
	char 		udp_pkt_src;
	char 		udp_pkt_data[sizeof(wan_udp_pkt_t)];
	unsigned short 	timer_int_enabled;
	char 		update_comms_stats;             /* update comms stats */
} chdlc_private_area_t;

/* Route Status options */
#define NO_ROUTE	0x00
#define ADD_ROUTE	0x01
#define ROUTE_ADDED	0x02
#define REMOVE_ROUTE	0x03

/* Array of Sangoma card structure */
extern sdla_t sdla_card_array[];

/* variable for keeping track of enabling/disabling FT1 monitor status */
static int rCount = 0;

#ifdef __OpenBSD__
extern int ticks;
#endif
/****** Function Prototypes *************************************************/

/* WAN link driver entry points. These are called by the WAN router module. */
static int update (wan_device_t* wandev);
static int new_if (wan_device_t* wandev, struct ifnet* ifp, wanif_conf_t* conf);
static int del_if (wan_device_t* wandev, struct ifnet* ifp);

/* Network device interface */
static int chdlc_close(struct ifnet *ifp);

/* ifnet entry points */
static int 	chdlc_output		/* output routine (enqueue) */
		__P((struct ifnet *, struct mbuf *, struct sockaddr *,
		     struct rtentry *));
static 	void	chdlc_start		/* initiate output routine */
		__P((struct ifnet *));
static	int	chdlc_ioctl		/* ioctl routine */
		__P((struct ifnet *, u_long, caddr_t));
static	void	chdlc_watchdog		/* timer routine */
		__P((struct ifnet *));
static	int	chdlc_init		/* init routine */
		__P((struct ifnet *, struct ifaddr *));

/* CHDLC Firmware interface functions */
static int set_adapter_config (sdla_t* card);
static int chdlc_configure 	(sdla_t* card, void* data);
static int chdlc_comm_enable 	(sdla_t* card);
static int chdlc_comm_disable 	(sdla_t* card);
static int chdlc_read_version 	(sdla_t* card, char* str, int);
static int chdlc_set_intr_mode 	(sdla_t* card, unsigned mode);
#if 0
static int chdlc_read_intr_mode (sdla_t* card);
#endif
static int chdlc_send(sdla_t* card, void* data, unsigned len);
static int chdlc_read_comm_err_stats (sdla_t* card, COMMS_ERROR_STATS_STRUCT* stats);
static int chdlc_read_configure(sdla_t *card, CHDLC_CONFIGURATION_STRUCT* cfg);
static int chdlc_read_op_stats (sdla_t* card, CHDLC_OPERATIONAL_STATS_STRUCT* op_stats);

/* Miscellaneous CHDLC Functions */
static int set_chdlc_config (sdla_t *card);
static void init_chdlc_tx_rx_buff( sdla_t* card, struct ifnet *ifp);
static int chdlc_error (sdla_t *card, int err, wan_mbox_t *mbox);
static int process_chdlc_exception(sdla_t *card);
static int process_global_exception(sdla_t *card);
static int update_comms_stats(sdla_t* card, struct ifnet* ifp); 
static int configure_ip (sdla_t* card);
static int unconfigure_ip (sdla_t* card);
static void process_route(void *arg);
static int chdlc_config(struct ifnet* ifp);

/* Interrupt handlers */
static void wpc_isr (sdla_t* card);
static void rx_intr (sdla_t* card);
static void timer_intr(sdla_t *);

/* Miscellaneous functions */
static int chk_bcast_mcast_addr(sdla_t* card, struct ifnet* ifp,
							    struct mbuf *m);
static int reply_udp(unsigned char *data, unsigned int mbox_len );
#ifdef CHDLC_OLD_INTR_TEST
static int chdlc_intr_test(sdla_t* card);
#endif
static int udp_pkt_type(caddr_t pkt, sdla_t* card);
static int store_udp_mgmt_pkt(char udp_pkt_src, sdla_t* card,
                                struct mbuf *m, struct ifnet* ifp,
                                chdlc_private_area_t* chdlc_priv_area);
static int process_udp_mgmt_pkt(sdla_t* card, struct ifnet *ifp,
				chdlc_private_area_t* chdlc_priv_area); 
static unsigned short calc_checksum(char*, int);

/* TE1 */
static WRITE_FRONT_END_REG_T write_front_end_reg;
static READ_FRONT_END_REG_T  read_front_end_reg;
static void chdlc_enable_timer(void* card_id);
static void chdlc_handle_front_end_state(void* card_id);

/* Miscellaneous utilities function */
#ifdef WANDEBUG_CHDLC
static void chdlc_dump_mailbox(wan_mbox_t *);
static void chdlc_dump_udp(wan_udp_pkt_t *, unsigned int);
#endif

extern struct mbuf* sdla_get(struct ifnet*, char*, unsigned);
static int  Intr_test_counter;

static int chdlc_udp_ioctl(struct ifnet* ifp, caddr_t data);
static int process_udp_cmd(struct ifnet* ifp, wan_udp_hdr_t* udp_hdr);

static int chdlc_debugging(sdla_t* card);
static unsigned long chdlc_crc_frames(sdla_t* card);
static unsigned long chdlc_abort_frames(sdla_t * card);
static unsigned long chdlc_tx_underun_frames(sdla_t* card);

/****** Public Functions ****************************************************/

/* 
 * ============================================================================
 * Cisco HDLC protocol initialization routine.
 *
 * This routine is called by the main WANPIPE module during setup.  At this
 * point adapter is completely initialized and firmware is running.
 *  o read firmware version (to make sure it's alive)
 *  o configure adapter
 *  o initialize protocol-specific fields of the adapter data space.
 *
 * Return:	0	o.k.
 *		< 0	failure.
 */
int wpc_init (sdla_t* card, wandev_conf_t* conf)
{
	CHDLC_CONFIGURATION_STRUCT	cfg;
	wan_mbox_t*			mbox;
	unsigned int 			timeout;
	unsigned char 			port_num;
	unsigned long 			max_permitted_baud = 0;
	union{
		char str[80];
	} u;
	int 	err, s = splimp();

	/* Verify configuration ID */
	if (conf->config_id != WANCONFIG_CHDLC && conf->config_id != WANCONFIG_MPPP){ 
		DEBUG_EVENT("%s: invalid configuration ID %u!\n",
		 			card->devname, conf->config_id);
		splx(s);
		return -EINVAL;
	}

	/* Find out which Port to use */
	if ((conf->comm_port == WANOPT_PRI) || (conf->comm_port == WANOPT_SEC)){
		if (card->next){
			if (conf->comm_port != card->next->u.c.comm_port){
				card->u.c.comm_port = conf->comm_port;
			}else{
				DEBUG_EVENT("%s: %s port busy: already in use! \n", 
						card->devname, PORT(conf->comm_port));
				splx(s);
				return -EINVAL;
			}
		}else{
			card->u.c.comm_port = conf->comm_port;
		}
	}else{ 
		DEBUG_EVENT("%s: Invalid Port selected!\n", card->devname);
		splx(s);
		return -EINVAL;
	}
	
	/* Initialize protocol-specific fields */
	if (card->type != SDLA_S514){
		if (card->u.c.comm_port == WANOPT_PRI){
			card->mbox_off	= 0x00;
		}else{
			card->mbox_off = SEC_BASE_ADDR_MB_STRUCT - 
					 PRI_BASE_ADDR_MB_STRUCT;
		}
	}else{
		/* for a S514 adapter, set a pointer to the actual mailbox in the */
		/* allocated virtual memory area */
		if (card->u.c.comm_port == WANOPT_PRI){
			card->mbox_off = PRI_BASE_ADDR_MB_STRUCT;
		}else{
			card->mbox_off = SEC_BASE_ADDR_MB_STRUCT;
		}
	}

	port_num = card->u.c.comm_port;
	mbox = &card->wan_mbox;

	if (!card->wandev.piggyback){
		u8 return_code;
	        /* The board will place an 'I' in the return code to indicate that it is
        	 * ready to accept commands.  We expect this to be completed in less
	         * than 1 second. 
        	 */
		timeout = SYSTEM_TICKS;
		do {
			return_code = 0x00;
			card->hw_iface.peek(card->hw, 
				card->mbox_off+offsetof(wan_mbox_t, wan_return_code),
				&return_code, 
				sizeof(unsigned char));
			if ((SYSTEM_TICKS - timeout) > 1*HZ) break;
		}while(return_code != 'I');
		if (return_code != 'I') {
			DEBUG_EVENT("%s: Initialization not completed by adapter\n", 
							card->devname);
        		DEBUG_EVENT("Please contact Sangoma representative.\n");
			splx(s);
			return -EIO;
		}
		DEBUG_INIT("%s: Received I from the board!!! Great\n",
							card->devname);
	}

	err = (card->hw_iface.check_mismatch) ? 
			card->hw_iface.check_mismatch(card->hw,conf->te_cfg.media) : -EINVAL;
	if (err){
		return err;
	}

	/* TE1 Make special hardware initialization for T1/E1 board */
	if (IS_TE1(conf->te_cfg)){
		
		memcpy(&card->wandev.te_cfg, &conf->te_cfg, sizeof(sdla_te_cfg_t));
		card->wandev.write_front_end_reg = write_front_end_reg;
		card->wandev.read_front_end_reg = read_front_end_reg;
		card->wandev.te_enable_timer = chdlc_enable_timer;
		card->wandev.te_link_state = chdlc_handle_front_end_state;
		conf->interface = 
			(card->wandev.te_cfg.media == WANOPT_MEDIA_T1) ? WANOPT_V35 : WANOPT_RS232;

		if (card->u.c.comm_port == WANOPT_PRI){
			conf->clocking = WANOPT_EXTERNAL;
		}

	}else if (IS_56K(conf->te_cfg)){

		memcpy(&card->wandev.te_cfg, &conf->te_cfg, sizeof(sdla_te_cfg_t));
		card->wandev.write_front_end_reg = write_front_end_reg;
		card->wandev.read_front_end_reg = read_front_end_reg;
		
		if (card->u.c.comm_port == WANOPT_PRI){
			conf->clocking = WANOPT_EXTERNAL;
		}

	}else{
		card->wandev.front_end_status = FE_CONNECTED;
	}


	/* Read firmware version.  Note that when adapter initializes, it
	** clears the mailbox, so it may appear that the first command was
	** executed successfully when in fact it was merely erased. To work
	** around this, we execute the first command twice.
	*/

	if (chdlc_read_version(card, u.str, 80)){
		splx(s);
		return -EIO;
	}
	if (conf->config_id == WANCONFIG_CHDLC){
		DEBUG_EVENT("%s: running Cisco HDLC firmware v%s\n", 
					card->devname, u.str); 
	}else{
		DEBUG_EVENT("%s: Running Raw HDLC firmware v%s for Multi-Port PPP protocol.\n", 
					card->devname,u.str); 
	}

	if (set_adapter_config(card)) {
		splx(s);
		return -EIO;
	}

	card->isr			= &wpc_isr;
	card->poll			= NULL;
	card->exec			= NULL;
	card->wan_debugging		= &chdlc_debugging;
	card->get_crc_frames	= &chdlc_crc_frames;
	card->get_abort_frames	= &chdlc_abort_frames;
	card->get_tx_underun_frames	= &chdlc_tx_underun_frames;
	card->wandev.update		= &update;
	card->wandev.new_if		= &new_if;
	card->wandev.del_if		= &del_if;
	card->wandev.udp_port   = conf->udp_port;

	/* Set the flag pointers for both ports to NULL */
	card->flags_off = NULL;

	/* reset the number of times the 'update()' proc has been called */
	card->u.c.update_call_count = 0;
	
	card->wandev.ttl = conf->ttl;
	card->wandev.interface = conf->interface;
	
	if (card->u.c.comm_port == WANOPT_SEC && conf->interface == WANOPT_V35 &&
	    card->type != SDLA_S514){
		DEBUG_EVENT("%s: V35 Interface not supported on S508 %s Port!\n",
				card->devname, PORT(card->u.c.comm_port));
		splx(s);
		return -EIO;
	} 

	card->wandev.clocking = conf->clocking;
	
	if (card->wandev.clocking){
		if (port_num == WANOPT_PRI){ /* For primary port */
			max_permitted_baud = 
				(card->type == SDLA_S514) ? PRI_MAX_BAUD_RATE_S514 :
						       PRI_MAX_BAUD_RATE_S508;	
		}else{ 
			if (port_num == WANOPT_SEC){ /* For secondary port */
		        	max_permitted_baud = 
			    		(card->type == SDLA_S514) ? SEC_MAX_BAUD_RATE_S514 :
				  		           SEC_MAX_BAUD_RATE_S508;	
			}
		}
		if (conf->bps > max_permitted_baud){
			conf->bps = max_permitted_baud;
			DEBUG_EVENT("%s: Baud too high!!!\n", card->devname);
			DEBUG_EVENT("%s: Baud rate set to %lu bps.\n",
						card->devname, max_permitted_baud);
		}
		card->wandev.bps = conf->bps;
	}else{
		card->wandev.bps = 0;
	}
		
	/* Setup the Port MTU */
	if (port_num == WANOPT_PRI){ /* For primary port 0 */
		card->wandev.mtu = (conf->mtu >= MIN_LGTH_CHDLC_DATA_CFG) ?
				min(conf->mtu, PRI_MAX_NO_DATA_BYTES_IN_FRAME) :
				CHDLC_DFLT_DATA_LEN;
	}else{
		if (port_num == WANOPT_SEC){ /* For secondary port 1 */
			card->wandev.mtu = (conf->mtu >= MIN_LGTH_CHDLC_DATA_CFG) ?	
					min(conf->mtu, SEC_MAX_NO_DATA_BYTES_IN_FRAME) :
					CHDLC_DFLT_DATA_LEN;
		}
	}
	/* Set up the interrupt status area. 
	** Read the CHDLC_configuration and obtain:
	** Ptr to sahred memory info struct.
	** Use this pointer to calculate the value of card->u.c.flags!
	*/
	err = chdlc_read_configure(card, &cfg);
	if (err != WAN_CMD_OK){
		splx(s);
		return -EIO;
	}

	if (card->type == SDLA_S514){
        	card->flags_off = 
			cfg.ptr_shared_mem_info_struct;
	}else{
        	card->flags_off = 
			cfg.ptr_shared_mem_info_struct % SDLA_WINDOWSIZE;
	}
    	card->intr_type_off = card->flags_off +
		offsetof(SHARED_MEMORY_INFO_STRUCT, interrupt_info_struct) +
		offsetof(INTERRUPT_INFORMATION_STRUCT, interrupt_type);
	card->intr_perm_off = card->flags_off + 
		offsetof(SHARED_MEMORY_INFO_STRUCT, interrupt_info_struct) +
		offsetof(INTERRUPT_INFORMATION_STRUCT, interrupt_permission);
	card->fe_status_off = card->flags_off + 
		offsetof(SHARED_MEMORY_INFO_STRUCT, FT1_info_struct) +
		offsetof(FT1_INFORMATION_STRUCT, parallel_port_A_input);

	/* This is for the ports link state */
	card->wandev.state	= WAN_DISCONNECTED;
	   
	splx(s);
#ifdef CHDLC_OLD_INTR_TEST
	if (!card->wandev.piggyback){
		err = chdlc_intr_test(card);
	}
#endif

	return err;
}

/******* WAN Device Driver Entry Points *************************************/

/* 
 * ============================================================================
 * Update device status & statistics (called when updating the PROC file system)
 */
static int update (wan_device_t* wandev)
{
	sdla_t* 			card = NULL;
	volatile chdlc_private_area_t* 	chdlc_priv_area = NULL;
	unsigned long 			timeout;
 
	/* sanity checks */
	if ((wandev == NULL) || (wandev->private == NULL)){
		return -EFAULT;
	}

	if(wandev->state == WAN_UNCONFIGURED){
		return -ENODEV;
	}

	card = wandev->private;

	/* more sanity checks */
	if (!card->flags_off){
		return -ENODEV;
	}

	/* Only one interface must be in interface list */
	if (wandev->if_head == NULL || 
	    wandev->if_head->ifp == NULL ||
	    wandev->if_head->ifp->if_softc == NULL){
		return -ENODEV;
	}
	chdlc_priv_area = wandev->if_head->ifp->if_softc;
    
	if (chdlc_priv_area->update_comms_stats){
		return -EAGAIN;
	}

	/* TE1_56K	
	 *  If update will need to read T1/E1/56K statistics,
	 *  change the update_comms_stats variable to 3, 
	 *  otherwise 2.
	 */
	chdlc_priv_area->update_comms_stats = 2; 
	
	card->hw_iface.enable_byte(card->hw, card->intr_perm_off, APP_INT_ON_TIMER);
	chdlc_priv_area->timer_int_enabled = TMR_INT_ENABLED_UPDATE;

	/* wait maximum of 1 second for the statistics to be updated */
	timeout = SYSTEM_TICKS;
	for(;;){
		if (chdlc_priv_area->update_comms_stats == 0){
			break;
		}
		if ((SYSTEM_TICKS - timeout) > (1*hz)){
			chdlc_priv_area->update_comms_stats = 0;
			chdlc_priv_area->timer_int_enabled &= 
						~TMR_INT_ENABLED_UPDATE;
			return -EAGAIN;
		}
	}   
	return 0;
}

/* 
 * ============================================================================
 * Create new logical channel.
 * This routine is called by the router when ROUTER_IFNEW IOCTL is being
 * handled.
 * o parse media- and hardware-specific configuration
 * o make sure that a new channel can be created
 * o allocate resources, if necessary
 * o prepare network device structure for registaration.
 *
 * Return:	0	o.k.
 *		< 0	failure (channel will not be created)
 */
static int new_if (wan_device_t* wandev, struct ifnet* ifp, wanif_conf_t* conf)
{
	sdla_t* card = NULL;
	chdlc_private_area_t* chdlc_priv_area = NULL;

	DEBUG_INIT("%s: Initialize new interface %s\n", 
					wandev->name, conf->name);	
	if ((wandev == NULL) || (wandev->private == NULL)){
		return -ENODEV;
	}
	card = wandev->private; 

	if ((conf->name[0] == '\0') || (strlen(conf->name) >= IFNAMSIZ)){ 
		DEBUG_EVENT("%s: invalid interface name!\n", card->devname);
		return -EINVAL;
	}
	
	/* allocate and initialize private data */
	chdlc_priv_area = wan_malloc(sizeof(chdlc_private_area_t));
	
	if(chdlc_priv_area == NULL){
		DEBUG_EVENT("%s: Failed allocating private data...\n", 
					card->devname); 
	        return -ENOMEM;
	}	
	bzero(chdlc_priv_area, sizeof(chdlc_private_area_t));

	chdlc_priv_area->card = card; 

	/* initialize data */
	sprintf(card->u.c.if_name, "%s%d", conf->name, 0);

	strcpy(card->u.c.if_name, conf->name);

	if (wandev->new_if_cnt > 0){
		DEBUG_EVENT("%s: Trying to initialize more than one network\n", 
								card->devname);
		DEBUG_EVENT("%s: interface to the same card!\n", 
								card->devname);
		wan_free(chdlc_priv_area);
	        return -EEXIST;	
	}

	wandev->new_if_cnt++;
	chdlc_priv_area->TracingEnabled = 0;
	chdlc_priv_area->route_status = NO_ROUTE;
	chdlc_priv_area->route_removed = 0;
     
	/* Setup protocol options */
	card->u.c.protocol_options = 0;

	if (conf->ignore_dcd == WANOPT_YES){
		card->u.c.protocol_options |= IGNORE_DCD_FOR_LINK_STAT;
	}
	if (conf->ignore_cts == WANOPT_YES){
		card->u.c.protocol_options |= IGNORE_CTS_FOR_LINK_STAT;
	}
	if (conf->ignore_keepalive == WANOPT_YES){
		card->u.c.protocol_options |= IGNORE_KPALV_FOR_LINK_STAT;
		card->u.c.kpalv_tx = MIN_Tx_KPALV_TIMER;
		card->u.c.kpalv_rx = MIN_Rx_KPALV_TIMER;
		card->u.c.kpalv_err = MIN_KPALV_ERR_TOL;
	}else{	/* Do not ignore keepalives */
 		card->u.c.kpalv_tx = 
			((conf->keepalive_tx_tmr - MIN_Tx_KPALV_TIMER) >= 0) ?
			min(conf->keepalive_tx_tmr, MAX_Tx_KPALV_TIMER) :
			DEFAULT_Tx_KPALV_TIMER;

	 	card->u.c.kpalv_rx = 
			((conf->keepalive_rx_tmr - MIN_Rx_KPALV_TIMER) >= 0) ?
			min(conf->keepalive_rx_tmr, MAX_Rx_KPALV_TIMER) :
			DEFAULT_Rx_KPALV_TIMER;

 		card->u.c.kpalv_err = 
			((conf->keepalive_err_margin - MIN_KPALV_ERR_TOL) >= 0) ?
			min(conf->keepalive_err_margin, MAX_KPALV_ERR_TOL) :
			DEFAULT_KPALV_ERR_TOL;
	}

	/* Setup slarp timer to control delay between slarps */
	card->u.c.slarp_timer = 
			((conf->slarp_timer - MIN_SLARP_REQ_TIMER) >= 0) ?
			min(conf->slarp_timer, MAX_SLARP_REQ_TIMER) :
			DEFAULT_SLARP_REQ_TIMER;

	/* If HDLC_STREAMING is enabled then IGNORE DCD, CTS and KEEPALIVES
	 * are automatically ignored 
	 */
	if (conf->hdlc_streaming == WANOPT_YES){
		DEBUG_EVENT("%s: Enabling HDLC STREAMING Mode.\n", 
						wandev->name);
		card->u.c.protocol_options = HDLC_STREAMING_MODE;
	}

	/* Setup wanpipe as arouter (WANPIPE) or as an API */
	if (strcmp(conf->usedby, "WANPIPE") == 0){
		DEBUG_EVENT("%s: Running in WANPIPE mode!\n", wandev->name);
		card->u.c.usedby = WANPIPE;
	}else{
        	if (strcmp(conf->usedby, "API") == 0){
			if (wandev->config_id == WANCONFIG_CHDLC){
				DEBUG_EVENT("%s: Running in API mode!\n", 
								wandev->name);
				card->u.c.usedby = API;
			}else{
				DEBUG_EVENT("%s: API Mode is not supported for SyncPPP!\n",
							wandev->name);
				wan_free(chdlc_priv_area);
				return -EINVAL;
			}
		}
	}
	/* Get Multicast Information */
	chdlc_priv_area->mc = conf->mc;

	/* prepare network device data space for registration */
	if (wan_netif_init(ifp, conf->name, 0)){
		DEBUG_EVENT("%s:%s: Failed to initialize network interface!\n", 
					card->devname, conf->name);
		wan_free(chdlc_priv_area);
		return -ENOMEM;
	}
	ifp->if_softc          	= chdlc_priv_area;
	ifp->if_mtu            	= wandev->mtu; /* ETHERMTU */
	ifp->if_type           	= IFT_OTHER;
	ifp->if_flags          	= IFF_POINTOPOINT;
	/* Enable Multicasting if user parameters */
	if (chdlc_priv_area->mc == WANOPT_YES){
		ifp->if_flags |= IFF_MULTICAST; 
	}
	ifp->if_output   	= chdlc_output;
	ifp->if_start          	= chdlc_start;
	ifp->if_watchdog       	= chdlc_watchdog;
	ifp->if_ioctl		= chdlc_ioctl;
#ifdef __FreeBSD__
	ifp->if_init           = (void*)chdlc_init;
#endif
	ifp->if_snd.ifq_len    = 0;
	/* ???? This members was not initialize in 3.1 */
	ifp->if_hdrlen = CHDLC_HDR_LEN;

	if (wandev->config_id == WANCONFIG_MPPP){
		if (conf->protocol == WANCONFIG_PPP){
			/* 
			** Verify that SPPP is enabled in the kernel.
			*/
			if (!WAN_SPPP_ENABLED){
				DEBUG_EVENT("%s: Please verify that SPPP is enabled in your kernel!\n",
							card->devname);
				return -EINVAL;
			}
			DEBUG_EVENT("%s: Attaching WAN interface to Kernel Sync PPP Protocol!\n",
						card->devname);
			WAN_SPPP_ATTACH(ifp);
		}else{
			DEBUG_EVENT("%s: Kernel Cisco HDLC Protocol doesn't supported!\n",
						card->devname);
			return -EINVAL;
		}
	}

	DEBUG_INIT("%s: New interface (%s) initialized\n", 
				card->devname, card->u.c.if_name); 
	return 0;
}

/* 
 * ============================================================================
 * Delete logical channel.
 */
static int del_if (wan_device_t* wandev, struct ifnet* ifp)
{
	sdla_t*	card = wandev->private;

	DEBUG_INIT("%s: Deleting CHDLC interface %s...\n", 
				card->devname, card->u.c.if_name); 
	if (card == NULL){
		return -ENODEV;
	}

	/* Initialize interrupt handler pointer */
	card->isr = NULL;

	/* Close current interface, set interrupt permission byte to 0 */ 
	chdlc_close(ifp);

	/* Initialize network interface */
	IFQ_PURGE(&ifp->if_snd);
	if (wandev->config_id == WANCONFIG_MPPP){
		DEBUG_EVENT("%s: Detaching WAN interface from Kernel Sync PPP Protocol!\n",
					card->devname);
		WAN_SPPP_FLUSH(ifp);
		WAN_SPPP_DETACH(ifp);
	}
	ifp->if_output = NULL;
	ifp->if_start = NULL;
	ifp->if_ioctl = NULL;
	ifp->if_watchdog = NULL;
	ifp->if_flags = 0;
	if (ifp->if_softc){ 
		wan_free(ifp->if_softc);
		ifp->if_softc = NULL;
	}
	wan_netif_del(ifp);

	return 0;
}

/****** Network Device Interface ********************************************/

/* 
 * cpipe_ioctl() -
 *
 */
static int
chdlc_ioctl(struct ifnet * ifp, u_long command, caddr_t data)
{
	chdlc_private_area_t* 	chdlc_priv_area = ifp->if_softc;
	sdla_t* 		card = chdlc_priv_area->card;
	struct ifaddr*	      	ifa = (struct ifaddr *)data; /* address struct */
	int 		      	s = 0;	    	    /* saved interrupt state */
	int	      		error = 0;    	    /* resulting status */

	s = splimp();

	switch(command){
	case SIOC_WANPIPE_MONITOR:
		break;
	default:
		if (card->wandev.config_id == WANCONFIG_MPPP){
			error = WAN_SPPP_IOCTL(ifp,command,data);
		}
		break;
	}
	if (error){
		goto chdlc_ioctl_end;
	}
	switch (command){ 
	case SIOCSIFADDR:
    		DEBUG_IOCTL("%s:%s: ioctl(SIOCSIFADDR) called.\n", 
				card->devname, wan_netif_name(ifp));
	    	/* check the protocol family - we currently support IP only */
    		switch(ifa->ifa_addr->sa_family){ 
	    	case AF_INET:
      			/* If chdlc_init fails it means 
			** we had an invalid IP address.
			*/
			if (chdlc_init(ifp, ifa)) 
			  	error = EADDRNOTAVAIL;
			break;
		default:
			error = EAFNOSUPPORT;
			break;
		}
		break;

	case SIOCSIFFLAGS:
		DEBUG_IOCTL("%s:%s: ioctl(SIOCSIFFLAGS) called.\n", 
				card->devname, wan_netif_name(ifp));
	    	/*
     		** If the interface is marked up - enable communications. 
	     	** If down - disable communications.  IFF_UP is taken 
		** care of before entering this function.
	     	*/
    		if ((ifp->if_flags & IFF_UP) == 0){ 
			/* bring it down */
			DEBUG_EVENT("%s:%s: Bringing interface down.\n", 
					card->devname, wan_netif_name(ifp));
			ifp->if_flags &= ~IFF_RUNNING;
		}else{ /* bring it up */ 
			DEBUG_EVENT("%s:%s: Bringing interface up.\n", 
					card->devname, wan_netif_name(ifp));
			if ((ifp->if_flags & IFF_RUNNING) == 0){ 
				/* update the up time ... */
				chdlc_priv_area->router_up_time = 
						wan_getcurrenttime();
			}
			/* Start debugging */
			WAN_DEBUG_START(card);
			/* Starting LCP (only for Multi-port PPP) */
			if (card->wandev.config_id == WANCONFIG_MPPP){
				struct sppp* sp = (struct sppp*)ifp;
				sp->pp_up(sp);
			}
		}
		break;

	case SIOC_WANPIPE_MONITOR:
		DEBUG_IOCTL("%s:%s: ioctl(SIOC_WANPIPE_MONITOR) called.\n", 
				card->devname, wan_netif_name(ifp));
		error = chdlc_udp_ioctl(ifp, data);
		break;

	default:
		DEBUG_IOCTL("%s:%s: ioctl(default) called.\n", 
				card->devname, wan_netif_name(ifp));
		error = EINVAL;
		break;
	}
chdlc_ioctl_end:
	splx(s);
	/* done */
	return error;
}

/*
 *
 * ===========================================================================
 * chdlc_watchdog()
 *
 */
static void chdlc_watchdog(struct ifnet *ifp)
{
	chdlc_private_area_t *chdlc_priv_area = ifp->if_softc;
	sdla_t *card = chdlc_priv_area->card;

	DEBUG_TIMER("%s:%s - Watchdog function called\n", 
				card->devname, wan_netif_name(ifp));
	if ((card->wandev.state != WAN_UNCONFIGURED) && card->poll){
		card->poll(card);
	}
#ifndef CHDLC_OLD_INTR_TEST
	if (chdlc_priv_area->timer_int_enabled & TMR_INT_ENABLED_TEST){
		if (Intr_test_counter < MAX_NEW_INTR_TEST_COUNTER){
			chdlc_priv_area->timer_int_enabled &=
					~TMR_INT_ENABLED_TEST;
			DEBUG_EVENT("%s: Interrupt test failed (%d)\n",
					card->devname, Intr_test_counter);
			DEBUG_EVENT("%s: Please choose another interrupt\n",
					card->devname);
			chdlc_set_intr_mode(card, 0x00);
		}
    	}
#endif
	
	return;
}

/* 
 * ============================================================================
 * Open network interface.
 * o enable communications and interrupts.
 * o prevent module from unloading by incrementing use count
 *
 * Return 0 if O.k. or errno.
 */
static int chdlc_init (struct ifnet *ifp, struct ifaddr *ifa)
{
	chdlc_private_area_t*		chdlc_priv_area = ifp->if_softc;
	sdla_t*				card = chdlc_priv_area->card;
        struct sockaddr_in*		saddr = NULL;
        unsigned long			local_ip = 0,	/* local IP address */
					remote_ip = 0,	/* remote IP address */
					ip_mask = 0;	/* IP mask	    */
	int				err = 0;

	DEBUG_INIT("%s: Initialize interface...\n", 
					card->devname);
        /* get local IP address */
        if ((saddr = (struct sockaddr_in *)ifa->ifa_addr) != NULL){ 
            local_ip = htonl(saddr->sin_addr.s_addr);
	}

        /* get remote IP address */
        if ((saddr = (struct sockaddr_in *)ifa->ifa_dstaddr) != NULL){
	    remote_ip = htonl(saddr->sin_addr.s_addr);
	}

        /* get  IP netmask address */
        if ((saddr = (struct sockaddr_in *)ifa->ifa_netmask) != NULL){
	    ip_mask = htonl(saddr->sin_addr.s_addr);
	}

        chdlc_priv_area->IP_address = local_ip;
        chdlc_priv_area->IP_netmask = ip_mask;

	if (chdlc_set_intr_mode(card, APP_INT_ON_TIMER)){ 
		return -EIO;
	}

#ifdef CHDLC_OLD_INTR_TEST
	chdlc_priv_area->timer_int_enabled |= TMR_INT_ENABLED_CONFIG;
#else
	if (!card->wandev.piggyback){
		Intr_test_counter = 0;
		chdlc_priv_area->timer_int_enabled |= TMR_INT_ENABLED_TEST;
		/* Enable watchdog timer in case if hardware i
		 * interrupt will failed */ 
		ifp->if_timer = 5;
	}else{
		chdlc_priv_area->timer_int_enabled |= TMR_INT_ENABLED_CONFIG;
	}
#endif
	card->hw_iface.enable_byte(card->hw, card->intr_perm_off, APP_INT_ON_TIMER);

	wanpipe_open(card);
	DEBUG_INIT("%s: Initialization is passed\n", 
				card->devname);
	return err;
}

/* 
 * ============================================================================
 *
 *
 *
 */
static int chdlc_config (struct ifnet *ifp)
{
	chdlc_private_area_t*		chdlc_priv_area = ifp->if_softc;
	sdla_t*				card = chdlc_priv_area->card;
	int				err = 0;

	DEBUG_INIT("%s: Configuring interface...\n", 
					card->devname);
  	if (set_chdlc_config(card)){ 
		return -EIO;
	}

	/* Initialize Rx/Tx buffer control fields */
	init_chdlc_tx_rx_buff(card, ifp);
	
	if (IS_TE1_CARD(card)) {
		DEBUG_EVENT("%s: Configuring onboard %s CSU/DSU\n",
			card->devname, 
			(IS_T1_CARD(card))?"T1":"E1");
		if (sdla_te_config(card)){
			DEBUG_EVENT("%s: Failed %s configuratoin!\n",
						card->devname,
						(IS_T1_CARD(card))?"T1":"E1");
			return 0;
		}
	}

	if (IS_56K_CARD(card)) {
		DEBUG_EVENT("%s: Configuring 56K onboard CSU/DSU\n",
					card->devname);

		if(sdla_56k_config(card)) {
			DEBUG_EVENT("%s: Failed 56K configuration!\n",
						card->devname);
			return 0;
		}
	}

	/* Set interrupt mode and mask */
	if (chdlc_set_intr_mode(card, 
                        APP_INT_ON_RX_FRAME | APP_INT_ON_GLOBAL_EXCEP_COND |
                        APP_INT_ON_TX_FRAME | APP_INT_ON_CHDLC_EXCEP_COND | 
	                APP_INT_ON_TIMER)) 
	{
		return -EIO;
	}

	/* Mask the Transmit and Timer interrupt */
	card->hw_iface.disable_byte(card->hw, card->intr_perm_off, (APP_INT_ON_TX_FRAME | APP_INT_ON_TIMER));

	/* Enable communications */
	if (chdlc_comm_enable(card)){ 
		return -EIO;
	}

	wanpipe_set_state(card, WAN_CONNECTING);
	chdlc_priv_area->common.state = WAN_CONNECTING;
	chdlc_priv_area->router_start_time = wan_getcurrenttime();

	/* Manually poll the 56K CSU/DSU to get the status */
	if (IS_56K_CARD(card)){
		/* 56K Update CSU/DSU alarms */
		sdla_56k_alarm(card, 1);
	}
	return err;
}

/* 
 * ============================================================================
 * Send a packet on a network interface.
 * chdlc_output() - The function tries to enqueue the packet and then if the 
 *		 device is not IFF_OATIVE calls chdlc_start(). The function 
 *		 returns an error value or 0 on success.
 *
 */
static int
chdlc_output(struct ifnet *ifp, struct mbuf *m, struct sockaddr *dst, struct rtentry *rt0)
{
	chdlc_private_area_t*		chdlc_priv_area = ifp->if_softc;
	sdla_t*				card = chdlc_priv_area->card;
	int 			 	udp_type = 0;
	int 			 	s= 0, error = 0; 
	struct altq_pktattr		pktattr;

	if (m == NULL){ 
		/* If we get here, some higher layer thinks we've missed an
		** tx-done interrupt.
		*/
		DEBUG_EVENT("%s: interface %s got kicked\n", 
				card->devname, card->u.c.if_name);
		return 0;
	}

	if ((ifp->if_flags & (IFF_UP|IFF_RUNNING)) != (IFF_UP|IFF_RUNNING)){
		wan_skb_free(m);	/*m_freem(m);*/
		return ENETDOWN;
	}
    
	s = splimp();

	if ((ifp->if_flags & IFF_OACTIVE) != 0){ 
		/* If our device stays busy for at least 5 seconds then we increment
		** statistics. We expect that our device never stays busy more than 
		** 5 seconds.
		*/
		if((SYSTEM_TICKS - chdlc_priv_area->tick_counter) > (5 * hz)){
			ifp->if_collisions ++;
			DEBUG_EVENT("%s: Transmit times out \n", 
							card->devname);
		}
	}

	if (card->u.c.usedby != API && dst->sa_family != AF_INET){
		DEBUG_EVENT("%s: Protocol family is not supported!\n", 
						card->devname); 
		wan_skb_free(m);	/*m_freem(m);*/
		splx(s);
		return EAFNOSUPPORT;
	}

	/* check the udp packet type */
	udp_type = udp_pkt_type(mtod(m, caddr_t), card);
	if (udp_type == UDP_CPIPE_TYPE){ 
		wan_bpf_report(ifp, m, 1);
		if (store_udp_mgmt_pkt(UDP_PKT_FRM_STACK, card, m, ifp,
    				                         chdlc_priv_area)){
			card->hw_iface.enable_byte(card->hw, card->intr_perm_off, APP_INT_ON_TIMER);
		}
		splx(s);
		return 0;
	}

	/* check to see if the source IP address is a broadcast or 
	** multicast IP address 
	*/
	if (chk_bcast_mcast_addr(card, ifp, m)){
		IF_DROP(&ifp->if_snd);
		wan_skb_free(m);	/*m_freem(m);*/
		splx(s);
	        return 0;
	}

	/* classify the packet before prepanding link-headers */
	IFQ_CLASSIFY(&ifp->if_snd, m, dst->sa_family, &pktattr);

	if (card->wandev.state != WAN_CONNECTED){
		IF_DROP(&ifp->if_snd);
		DEBUG_TX("%s: Drop the packet (Device is not connected)!\n", 
				card->devname);
        	error = ENETDOWN;	
		wan_skb_free(m);	/*m_freem(m);*/
	}else{

		IFQ_ENQUEUE(&ifp->if_snd, m, &pktattr, error);
		if (error){
			DEBUG_TX("%s: Send queue is full!\n",
					card->devname);
			splx(s);
			return error;	
		}

		if ((ifp->if_flags & IFF_OACTIVE) == 0){ 
			chdlc_start(ifp);
		}
	}

	splx(s);
	return error;
}


/* 
 * ============================================================================
 * chdlc_start() -  called by the system when a packet is added to the 
 *		    output queue for the interface and the device flags
 *		    indicate the device is not currently busy ( IFF_OACTIVE
 *                  is clear ).
 *                  We will try to transmit the packet. In case there's no
 *  		    available buffer - we will set IFF_OACTIVE, enable 
 *		    transmit interrupt and exit. And then the tx interrupt 
 *		    handler will call chdlc_start() and once its done - it 
 *		    will reset the IFF_OACTIVE.
 */
static void chdlc_start(struct ifnet * ifp)
{
	chdlc_private_area_t*   	chdlc_priv_area = ifp->if_softc;
	sdla_t*		           	card = chdlc_priv_area->card;
	CHDLC_DATA_TX_STATUS_EL_STRUCT 	txbuf;
	int 		      	   	isempty = 0, s = 0;	/* saved int. state */
	netskb_t*			opkt = NULL;		/* the packet */
	register netskb_t*		m = NULL;		/* current mbuf */
	unsigned long 	           	data_addr;		/* where to copy data*/
	unsigned short	           	data_length;		/* data length */
	char* 			   	data = NULL; 

	s = splimp();

	/* We're trying to drain the queue as much as possible here because 
	** if we manage to get more than one packet into transmit queue     
	** ( line-down/line-up scenario, or heavy load ) and take only one  
	** at a time out we get an effective delay ...		      
	** BF-0001 
	*/ 
	while(1){
		isempty = (card->wandev.config_id == WANCONFIG_CHDLC) ?
				IFQ_IS_EMPTY(&ifp->if_snd): WAN_SPPP_ISEMPTY(ifp);
		if (isempty){
			/* No more packets in send queue */
			break;
		}
						
		if (card->wandev.config_id == WANCONFIG_CHDLC){
			IFQ_POLL(&ifp->if_snd, opkt);
		}else{
			opkt = WAN_SPPP_PICK(ifp);
		}
		if (opkt == NULL){
			/* Should never happened, packet pointer is NULL */
			break;
		}
		/* calculate tx element address */
		card->hw_iface.peek(card->hw, card->u.c.txbuf_off, &txbuf, sizeof(txbuf));
		/* check if the buffer is available */
		if (txbuf.opp_flag == 0x01){ 
			/* mark an interface active */
			ifp->if_flags |= IFF_OACTIVE; 

			chdlc_priv_area->tick_counter = SYSTEM_TICKS;
			ifp->if_oerrors ++;

			/* enable transmit interrupts */
			card->hw_iface.enable_byte(card->hw, card->intr_perm_off, APP_INT_ON_TX_FRAME);
			/* restore priority level and exit */
			splx(s);
			return;
		}
  
		/* 
		** ok, now when we've got an available buffer we can transmit something
		*/
		/* Calculate packet's length */ 
		data_length = 0;
		for(m = opkt; m != NULL; m = m->m_next){ 
			data_length += m->m_len;
		}
		/* 
		** Check that packet length is less then MAX_PACKET_SIZE.
		** Copy transmit packet to tx_data buffer. 
		*/
		if (data_length > MAX_PACKET_SIZE){
			DEBUG_TX("%s: Drop the packet (too big - %d bytes)!\n",
					card->devname, data_length);
			IF_DROP(&ifp->if_snd);
		
			/* dequeue packet from IP queue */
			if (card->wandev.config_id == WANCONFIG_CHDLC){
				IFQ_DEQUEUE(&ifp->if_snd, opkt);
			}else{
				opkt = WAN_SPPP_DEQUEUE(ifp);
			}
			if (opkt){
				wan_skb_free(opkt);
			}
			continue;
		}
		/* Copy the packet inside */
		data_length = 0;
		for(m = opkt; m != NULL; m = m->m_next){ 
			bcopy(mtod(m, caddr_t), &card->tx_data[data_length], m->m_len);
			data_length += m->m_len;
		}
		card->tx_data[data_length] = '\0';	
		data = card->tx_data;
		if (card->u.c.usedby == API){
			/* Only for CHDLC */
			api_tx_hdr_t *api_tx_hdr = NULL;
			unsigned char attr;
			if (data_length <= sizeof(api_tx_hdr_t)){
				IF_DROP(&ifp->if_snd);
				/* dequeue packet from IP queue */
				if (card->wandev.config_id == WANCONFIG_CHDLC){
					IFQ_DEQUEUE(&ifp->if_snd, opkt);
				}else{
					opkt = WAN_SPPP_DEQUEUE(ifp);
				}
				if (opkt){
					wan_skb_free(opkt);
				}
				continue;
			}	
			api_tx_hdr = (api_tx_hdr_t*)card->tx_data;
			attr = api_tx_hdr->attr;
			data = &card->tx_data[sizeof(api_tx_hdr_t)];
			data_length -= sizeof(api_tx_hdr_t);    
		}

		/* initialize some of our stuff */
		data_addr 	= txbuf.ptr_data_bfr;

		/* Copy data to the board */
		card->hw_iface.poke(card->hw, data_addr, data, data_length);

		/* write data length to the transmit element - offset 1, length 2 */
		txbuf.frame_length = data_length;
		card->hw_iface.poke(card->hw, card->u.c.txbuf_off, &txbuf, sizeof(txbuf));
		txbuf.opp_flag = 1;
		card->hw_iface.poke(card->hw, card->u.c.txbuf_off, &txbuf, sizeof(txbuf));
		/* report the packet to BPF if present and attached */
		wan_bpf_report(ifp, opkt, 1);

		/* dequeue packet from IP queue (only for CHDLC) */
		if (card->wandev.config_id == WANCONFIG_CHDLC){
			IFQ_DEQUEUE(&ifp->if_snd, opkt);
		}else{
			opkt = WAN_SPPP_DEQUEUE(ifp);
		}

		if (opkt){
			wan_skb_free(opkt);	/*m_freem(opkt);*/
		}

		/* update statistics and free the packet */
		ifp->if_opackets++;
		ifp->if_obytes += data_length;
 
		/* Update transmit buffer control fields */
		card->u.c.txbuf_off += sizeof(CHDLC_DATA_TX_STATUS_EL_STRUCT);
		if (card->u.c.txbuf_off > card->u.c.txbuf_last_off){
			card->u.c.txbuf_off = card->u.c.txbuf_base_off;
		}
	} /* while there're packets in the queue */

	/* restore interrupt level and exit */
	splx(s);
	return;
}

/*
 * ============================================================================
 * Close network interface.
 * o if this is the last open, then disable communications and interrupts.
 * o reset flags.
 */
static int chdlc_close (struct ifnet* ifp)
{
	chdlc_private_area_t*		chdlc_priv_area = ifp->if_softc;
	sdla_t* 			card = chdlc_priv_area->card;
	int				s = 0;		       	/* saved int. state */

	DEBUG_INIT("%s: Close interface %s\n", 
				card->devname, card->u.c.if_name);
	s = splimp();
	/* Disable all interrupts */
	chdlc_set_intr_mode(card, 0);
	card->hw_iface.poke_byte(card->hw, card->intr_perm_off, 0x00);
	wanpipe_close(card);
	wanpipe_set_state(card, WAN_DISCONNECTED);
	chdlc_priv_area->common.state = WAN_DISCONNECTED;
	chdlc_comm_disable(card);
	if (IS_TE1_CARD(card)){
		sdla_te_unconfig(card);
	}

	splx(s);
	return 0;
}

/* 
 * ============================================================================
 * Check to see if the packet to be transmitted contains a broadcast or
 * multicast source IP address.
 * Return 0 if not broadcast/multicast address, otherwise return 1.
 */
static int chk_bcast_mcast_addr(sdla_t *card, struct ifnet* ifp, struct mbuf *m)
{
	uint32_t 		src_ip_addr = 0;
	uint32_t 		broadcast_ip_addr = 0;
	struct ifaddr*	ifa = WAN_TAILQ_FIRST(ifp);
	struct sockaddr_in*	saddr = NULL;

	/* read the IP source address from the outgoing packet */
	src_ip_addr = *(uint32_t *)(mtod(m, caddr_t) + 12);

	/* read the IP broadcast address for the device */
	if ((saddr = (struct sockaddr_in *)ifa->ifa_broadaddr) != NULL) 
        	broadcast_ip_addr = htonl(saddr->sin_addr.s_addr);

	/* check if the IP Source Address is a Broadcast address */
	if((ifp->if_flags & IFF_BROADCAST) && (src_ip_addr == broadcast_ip_addr)){ 
        	DEBUG_EVENT("%s: Broadcast Source Address silently discarded!\n",
							card->devname);
		return 1;
	} 

	/* check if the IP Source Address is a Multicast address */
	if ((ntohl(src_ip_addr) >= 0xE0000001) && (ntohl(src_ip_addr) <= 0xFFFFFFFE)){
		DEBUG_EVENT("%s: Multicast Source Address silently discarded!\n", 
					card->devname);
		return 1;
	}

	return 0;
}

#if 0
/* 
 * ============================================================================
 * Reply to UDP Management system.
 * Return length of reply.
 */
static int reply_udp( unsigned char *data, unsigned int mbox_len )
{
    wan_udp_pkt_t*	udp_pkt = (wan_udp_pkt_t*)data;
    unsigned long 	ip_temp;
    unsigned short 	len, 
			udp_length, 
			temp, 
			ip_length;
    int 		even_bound = 0;
 
    /* Set length of packet */
    len = sizeof(struct ip) + sizeof(struct udphdr) + 
	  sizeof(wan_mgmt_t) + sizeof(wan_cmd_t) +
	  sizeof(wan_trace_info_t) + mbox_len;

    /* fill in UDP reply */
    udp_pkt->wan_udp_request_reply = UDPMGMT_REPLY;

    /* fill in UDP length */
    udp_length = sizeof(struct udphdr) +
	  	 sizeof(wan_mgmt_t) + sizeof(wan_cmd_t) +
	  	 sizeof(wan_trace_info_t) + mbox_len;

    /* put it on an even boundary */
    if (udp_length & 0x0001){ 
	udp_length += 1;
	len += 1;
	even_bound = 1;
    } 

    temp = (udp_length<<8)|(udp_length>>8);
    udp_pkt->wan_udp_len = temp;

    /* swap UDP ports */
    temp = udp_pkt->wan_udp_sport;
    udp_pkt->wan_udp_sport = udp_pkt->wan_udp_dport;
    udp_pkt->wan_udp_dport = temp;

    /* add UDP pseudo header */
    temp = 0x1100;
    *((unsigned short*)(udp_pkt->wan_udp_data + mbox_len + even_bound)) = temp;
    temp = (udp_length<<8)|(udp_length>>8);
    *((unsigned short*)(udp_pkt->wan_udp_data + mbox_len + even_bound + 2)) = temp;

    /* calculate UDP checksum */
    udp_pkt->wan_udp_sum = 0;
    udp_pkt->wan_udp_sum = 
		calc_checksum(&data[UDP_OFFSET], udp_length + UDP_OFFSET);

    /* fill in IP length */
    ip_length = len;
    temp = (ip_length<<8)|(ip_length>>8);
    udp_pkt->wan_ip_len = temp;

    /* swap IP addresses */
    ip_temp = udp_pkt->wan_ip_src;
    udp_pkt->wan_ip_src = udp_pkt->wan_ip_dst;
    udp_pkt->wan_ip_dst = ip_temp;
    
    /* fill in IP checksum */
    udp_pkt->wan_ip_sum = 0;
    udp_pkt->wan_ip_sum = calc_checksum(data, sizeof(struct ip)); 

    return len;
} /* reply_udp */

unsigned short calc_checksum(char *data, int len)
{
    unsigned short temp;
    unsigned long sum = 0;
    int i;
  
    for(i=0; i < len; i += 2){
	bcopy(&data[i], &temp, 2);
	sum += (unsigned long)temp;
    }

    while(sum >> 16){
	sum = (sum & 0xffffUL) + (sum >> 16);
    }    

    temp = (unsigned short)sum;
    temp = ~temp;
 
    if (temp == 0){
	temp = 0xFFFF;
    }

    return temp;
} 
#endif

/****** Cisco HDLC Firmware Interface Functions *******************************/
/* 
 * ============================================================================
 * Read firmware code version.
 *	Put code version as ASCII string in str. 
 */
static int chdlc_read_version (sdla_t* card, char* str, int len)
{
	wan_mbox_t*	mbox = &card->wan_mbox;
	int		err;

	DEBUG_CMD("%s: Read CHDLC Version..\n", card->devname); 
	WAN_MBOX_INIT(mbox);
	mbox->wan_data_len = 0;
	mbox->wan_command = READ_CHDLC_CODE_VERSION;
	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);

#ifdef WANDEBUG_CHDLC
	chdlc_dump_mailbox(mbox);
#endif
	if (err != WAN_CMD_OK){ 
		chdlc_error(card, err, mbox);
	}else if (str) {
		if (mbox->wan_data_len < len){
			memcpy(str, mbox->wan_data, mbox->wan_data_len);
			str[mbox->wan_data_len] = '\0';
		}else{
			DEBUG_EVENT("%s: Error: Version Length greater than max %i>%i!\n",
					card->devname, mbox->wan_data_len, len);
		}
	}
	return (err);
}

/* 
 * ============================================================================
 *  Configure CHDLC firmware.
 */
static int chdlc_configure (sdla_t* card, void* data)
{
	wan_mbox_t*	mbox = &card->wan_mbox;
	int 	err;

	DEBUG_INIT("%s: Setting CHDLC configuration...\n", card->devname);	
	WAN_MBOX_INIT(mbox);
	mbox->wan_data_len = sizeof(CHDLC_CONFIGURATION_STRUCT);
	mbox->wan_command = SET_CHDLC_CONFIGURATION;
	memcpy(mbox->wan_data, data, sizeof(CHDLC_CONFIGURATION_STRUCT));
    	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	if (err != WAN_CMD_OK){
        	chdlc_error (card, err, mbox);
		DEBUG_EVENT("%s: Invalid value in offset %d\n", 
				card->devname, mbox->wan_data_len);  
	}
	return err;
}

/* 
 * ============================================================================
 * Set interrupt mode -- HDLC Version.
 */

static int chdlc_set_intr_mode (sdla_t* card, unsigned mode)
{
	wan_mbox_t*			mbox = &card->wan_mbox;
	CHDLC_INT_TRIGGERS_STRUCT* 	int_data = 
			(CHDLC_INT_TRIGGERS_STRUCT *)mbox->wan_data;
	int err;

	DEBUG_CMD("%s: Setting interrupt mode(0x%x)...\n", card->devname, mode);
	WAN_MBOX_INIT(mbox);
	mbox->wan_command = SET_CHDLC_INTERRUPT_TRIGGERS;
	mbox->wan_data_len = sizeof(CHDLC_INT_TRIGGERS_STRUCT);
	int_data->CHDLC_interrupt_triggers	= mode;
	int_data->IRQ			= card->wandev.irq;
	int_data->interrupt_timer            = 1;
	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	if (err != WAN_CMD_OK){
		chdlc_error (card, err, mbox);
	}
   
	return err;
}

/*
 * ============================================================================
 * Get interrupt mode -- HDLC Version
 */
#if 0
static int chdlc_read_intr_mode(sdla_t* card)
{
	wan_mbox_t*	mbox = &card->wan_mbox;
	int 	err = 0;
     
	WAN_MBOX_INIT(mbox);
	mbox->wan_command = READ_CHDLC_INTERRUPT_TRIGGERS;
	mbox->wan_data_len = 0x0;
	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	if (err != WAN_CMD_OK){
		chdlc_error (card, err, mbox);
	}else{
	    	CHDLC_INT_TRIGGERS_STRUCT* int_data =
				(CHDLC_INT_TRIGGERS_STRUCT *)mb->wan_data;
		return int_data->CHDLC_interrupt_triggers;	
	}
     
	return err;
}
#endif

/* 
 * ============================================================================
 * Enable communications.
 */
static int chdlc_comm_enable (sdla_t* card)
{
	wan_mbox_t*	mbox = &card->wan_mbox;
	int		err = 0;

	WAN_MBOX_INIT(mbox);
	mbox->wan_data_len = 0;
	mbox->wan_command = ENABLE_CHDLC_COMMUNICATIONS;
	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	if (err != WAN_CMD_OK){
		if (err == CHDLC_COMMS_ENABLED){
			DEBUG_CMD("%s: Communication already enabled\n", 
						card->devname);
		}else{
			chdlc_error(card, err, mbox);
		}
	}else{
		DEBUG_CMD("%s: Communication enabled\n", card->devname);	
	}

	return err;
}

/* 
 * ============================================================================
 * Disable communications.
 */
static int chdlc_comm_disable (sdla_t* card)
{
	wan_mbox_t*	mbox = &card->wan_mbox;
	int		err;

	WAN_MBOX_INIT(mbox);
	mbox->wan_data_len = 0;
	mbox->wan_command = DISABLE_CHDLC_COMMUNICATIONS;
	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	if (err != WAN_CMD_OK){
		if (err == CHDLC_COMMS_DISABLED){
			DEBUG_CMD("%s: Communication already disabled\n", 
						card->devname); 
		}else{
			chdlc_error(card,err, mbox);
		}
	}else{
		DEBUG_CMD("%s: Communication disabled\n", card->devname);
	}
	return err;
}

/* 
 * ============================================================================
 * Read communication error statistics.
 */
static int chdlc_read_comm_err_stats (sdla_t* card, COMMS_ERROR_STATS_STRUCT* stats)
{
	wan_mbox_t* 	mbox = &card->wan_mbox;
	int		err;

	WAN_MBOX_INIT(mbox);
	mbox->wan_data_len = 0;
	mbox->wan_command = READ_COMMS_ERROR_STATS;
	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	if (err != WAN_CMD_OK){
		chdlc_error(card, err, mbox);
	}else{
		DEBUG_CMD("%s: Read communication error stat...ok\n", 
					card->devname);
		memcpy(stats, mbox->wan_data, sizeof(COMMS_ERROR_STATS_STRUCT));
	}
	return err;
}

/*
 ************************************************************************
 * Read CHDLC configureation.
 */
static int chdlc_read_configure(sdla_t *card, CHDLC_CONFIGURATION_STRUCT* cfg)
{
	wan_mbox_t*	mbox = &card->wan_mbox;
	int		err;

	WAN_MBOX_INIT(mbox);
	mbox->wan_data_len = 0;
	mbox->wan_command = READ_CHDLC_CONFIGURATION;
	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);

	if(err != WAN_CMD_OK){
		chdlc_error(card, err, mbox);
	}else{
		DEBUG_CMD("%s: Read CHDLC configuration...ok\n", 
				card->devname);
		bcopy(mbox->wan_data, cfg, sizeof(CHDLC_CONFIGURATION_STRUCT));
	}
	return err;
}


/* 
 * ============================================================================
 * Read CHDLC operational statistics.
 */
static int chdlc_read_op_stats (sdla_t* card, CHDLC_OPERATIONAL_STATS_STRUCT* op_stats)
{
	wan_mbox_t*	mbox = &card->wan_mbox;
	int		err;
	
	WAN_MBOX_INIT(mbox);
    	mbox->wan_data_len = 0;
    	mbox->wan_command = READ_CHDLC_OPERATIONAL_STATS;
    	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
    	if (err != WAN_CMD_OK){
		chdlc_error(card, err, mbox);
    	}else{
		DEBUG_CMD("%s: Read CHDLC operational stat...ok\n", 
				card->devname);
		memcpy(op_stats, mbox->wan_data, sizeof(CHDLC_OPERATIONAL_STATS_STRUCT));
    	}
    	return err;
}

/* 
 * ============================================================================
 * Update communications error and general packet statistics.
 */
static int 
update_comms_stats(sdla_t* card, struct ifnet* ifp)
{
	COMMS_ERROR_STATS_STRUCT		err_stats;
	CHDLC_OPERATIONAL_STATS_STRUCT 	op_stats;
	chdlc_private_area_t* 		chdlc_priv_area = ifp->if_softc;

	DEBUG_CMD("Updating communication statistics...\n");
	/* On the first timer interrupt, read the comms error statistics */
	if (chdlc_priv_area->update_comms_stats == 3){
		/* 1. On the first timer interrupt, update T1/E1 alarms 
		 * and PMON counters (only for T1/E1 card) (TE1) 
		 */
		/* TE1 Update T1/E1 alarms */
		if (IS_TE1_CARD(card)) {	
			card->wandev.cur_te_alarm = sdla_te_alarm(card, 0); 
			/* TE1 Update T1/E1 perfomance counters */
			sdla_te_pmon(card);
		}else if (IS_56K_CARD(card)) {
			/* 56K Update CSU/DSU alarms */
			card->wandev.k56_alarm = sdla_56k_alarm(card, 1); 
		}
	}else if (chdlc_priv_area->update_comms_stats == 2){
		if (chdlc_read_comm_err_stats(card, &err_stats)){
			return 1;
		}
		ifp->if_ierrors = 	err_stats.Rx_overrun_err_count + 
					err_stats.CRC_err_count +
					err_stats.Rx_abort_count +
					err_stats.Rx_dis_pri_bfrs_full_count +
					err_stats.sec_Tx_abort_count;
	}else{ /* on second timer interrupt, read the operational statistics */
		if (chdlc_read_op_stats(card, &op_stats)){
			return 1;
		}
		ifp->if_ierrors +=	(op_stats.Rx_Data_discard_short_count +
					op_stats.Rx_Data_discard_long_count);
	}	
	return 0;
}

/* 
 * ============================================================================
 * Send packet.
 *	Return:	0 - o.k.
 *		1 - no transmit buffers available
 */
static int 
chdlc_send (sdla_t* card, void* data, unsigned len)
{
	CHDLC_DATA_TX_STATUS_EL_STRUCT	txbuf;

	card->hw_iface.peek(card->hw, card->u.c.txbuf_off, &txbuf, sizeof(txbuf));
	if (txbuf.opp_flag) return 1;

	card->hw_iface.poke(card->hw, txbuf.ptr_data_bfr, data, len);

	txbuf.frame_length = len;
	card->hw_iface.poke(card->hw, card->u.c.txbuf_off, &txbuf, sizeof(txbuf));
	txbuf.opp_flag = 1;		/* start transmission */
	card->hw_iface.poke(card->hw, card->u.c.txbuf_off, &txbuf, sizeof(txbuf));

	/* Update transmit buffer control fields */
	card->u.c.txbuf_off += sizeof(CHDLC_DATA_TX_STATUS_EL_STRUCT);
	if (card->u.c.txbuf_off > card->u.c.txbuf_last_off){
		card->u.c.txbuf_off = card->u.c.txbuf_base_off;
	}

	return 0;
}

/*============================================================================
 * Read TE1/56K Front end registers
 */
static unsigned char 
read_front_end_reg (void* card_id, unsigned short reg)
{
	sdla_t* 	card = (sdla_t*)card_id;
	wan_mbox_t*	mbox = &card->wan_mbox;
	char*		data = mbox->wan_data;
	int 		err = 0x0;

	WAN_MBOX_INIT(mbox);
    	mbox->wan_data_len = sizeof(FRONT_END_REG_STRUCT);
	mbox->wan_command = READ_FRONT_END_REGISTER;
	((FRONT_END_REG_STRUCT *)data)->register_number = (unsigned short)reg;
	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	if (err != WAN_CMD_OK){
		chdlc_error(card,err, mbox);
	}
	return (((FRONT_END_REG_STRUCT *)data)->register_value);
}

/*============================================================================
 * Write to TE1/56K Front end registers  
 */
static unsigned char 
write_front_end_reg (void* card_id, unsigned short reg, unsigned char value)
{
	sdla_t*		card = (sdla_t*)card_id;
	wan_mbox_t*	mbox = &card->wan_mbox;
	char*		data = mbox->wan_data;
	int 		err = 0x00;
	
	WAN_MBOX_INIT(mbox);
	mbox->wan_data_len = sizeof(FRONT_END_REG_STRUCT);
	mbox->wan_command = WRITE_FRONT_END_REGISTER;
	((FRONT_END_REG_STRUCT *)data)->register_number = (unsigned short)reg;
	((FRONT_END_REG_STRUCT *)data)->register_value = value;
	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	if (err != WAN_CMD_OK){
		chdlc_error(card, err, mbox);
	}
	return err;
}

/*============================================================================
 * TE1 CHDLC enable hardware timer  
 */
static void chdlc_enable_timer(void* card_id)
{
	sdla_t* 		card = (sdla_t*)card_id;
	struct ifnet*		ifp = NULL;
	chdlc_private_area_t*	chdlc_priv_area = NULL;

	ifp = card->wandev.if_head->ifp;
	chdlc_priv_area = ifp->if_softc;
	chdlc_priv_area->timer_int_enabled |= TMR_INT_ENABLED_TE;
	card->hw_iface.enable_byte(card->hw, card->intr_perm_off, APP_INT_ON_TIMER);
	return;
}


static int set_adapter_config (sdla_t* card)
{
	wan_mbox_t*			mbox = &card->wan_mbox;
	ADAPTER_CONFIGURATION_STRUCT*	cfg = (ADAPTER_CONFIGURATION_STRUCT*)mbox->wan_data;
	int 				err = 0x00;
	
	WAN_MBOX_INIT(mbox);
	mbox->wan_data_len = sizeof(ADAPTER_CONFIGURATION_STRUCT);
	mbox->wan_command = SET_ADAPTER_CONFIGURATION;
	card->hw_iface.getcfg(card->hw, SDLA_ADAPTERTYPE, &cfg->adapter_type);
	cfg->adapter_config = 0x00; 
	cfg->operating_frequency = 0x00; 
	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	if (err != WAN_CMD_OK){
		chdlc_error(card, err, mbox);
	}

	return (err);
}


/****** Firmware Error Handler **********************************************/

/* 
 * ============================================================================
 * Firmware error handler.
 *	This routine is called whenever firmware command returns non-zero
 *	return code.
 *
 * Return zero if previous command has to be cancelled.
 */
static int chdlc_error (sdla_t *card, int err, wan_mbox_t* mbox)
{
	switch (err){ 
	case WAN_CMD_TIMEOUT:
		DEBUG_EVENT("%s: command 0x%02x timed out!\n", 
					card->devname, mbox->wan_command);
		break;
	
	case S514_BOTH_PORTS_SAME_CLK_MODE:
		if (mbox->wan_command == SET_CHDLC_CONFIGURATION){
			DEBUG_EVENT("%s: Configure both ports for the same clock source\n",
						card->devname);
		}else{
			DEBUG_EVENT("%s: command 0x%02x returned 0x%02x!\n", 
						card->devname, mbox->wan_command, err);
		}
		break;
	
	default:
		DEBUG_EVENT("%s: command 0x%02x returned 0x%02x!\n", 
						card->devname, mbox->wan_command, err);
		break;
	}
	return 0;
}

/****** Interrupt Handlers **************************************************/

/* 
 * ============================================================================
 * Cisco HDLC interrupt service routine.
 */
static void wpc_isr (sdla_t* card)
{
	struct ifnet*		ifp = NULL;
	chdlc_private_area_t*	chdlc_priv_area = NULL;
	SHARED_MEMORY_INFO_STRUCT	flags;
	sdla_t*			my_card = NULL;
	int 			interrupt_serviced = 0;

	/* Check for which port the interrupt has been generated 
	 * Since Secondary Port is piggiback on the Primary the check be done here
	 */
	card->hw_iface.peek(card->hw, card->flags_off,
			&flags, sizeof(flags));
	if (!flags.interrupt_info_struct.interrupt_type){
		/* Check for second port (piggiback) */
		if ((my_card = card->next) != NULL) {
			card->hw_iface.peek(card->hw, my_card->flags_off,
					&flags, sizeof(flags));
			if (flags.interrupt_info_struct.interrupt_type){
				card = my_card;
			}
		}
	}
	if (card->wandev.if_head && card->wandev.if_head->ifp){
    		ifp = card->wandev.if_head->ifp;
		chdlc_priv_area = ifp->if_softc;
	}
	card->in_isr = 1;

	/* On a 508 Card, if critical due to if_send Major Error */
	if (card->flags_off == 0){
		goto wpc_isr_done;
	}
	interrupt_serviced = 1;
	switch(flags.interrupt_info_struct.interrupt_type){
	case RX_APP_INT_PEND:	/* 0x01: receive interrupt */
		rx_intr(card);
		break;

	case TX_APP_INT_PEND:	/* 0x02: transmit interrupt */
		card->hw_iface.disable_byte(card->hw, card->intr_perm_off, APP_INT_ON_TX_FRAME);
		/* interface is not active anymore */
		if (ifp){
			ifp->if_flags &= ~IFF_OACTIVE; 
		}
		/* Update tick_counter */
		if (chdlc_priv_area){
			chdlc_priv_area->tick_counter = SYSTEM_TICKS;
		}
       
		/* check the interface output queue */
		if (ifp && !IFQ_IS_EMPTY(&ifp->if_snd)){ 
			/* there's something to send - send it */
			chdlc_start(ifp);
		}
		break;
             
	case COMMAND_COMPLETE_APP_INT_PEND:/* 0x04: cmd cplt */
		++ Intr_test_counter;
		break;

	case TIMER_APP_INT_PEND:
		timer_intr(card);
		break;

	case GLOBAL_EXCEP_COND_APP_INT_PEND:
		process_global_exception(card);
		/* Reset the 56k or T1/E1 front end exception condition */
		if(IS_56K_CARD(card) || IS_TE1_CARD(card)) {
			FRONT_END_STATUS_STRUCT FE_status;
			card->hw_iface.peek(card->hw, 
					    card->fe_status_off,
					    &FE_status,
					    sizeof(FE_status));
			FE_status.opp_flag = 0x01;	
			card->hw_iface.poke(card->hw, 
					    card->fe_status_off,
					    &FE_status,
					    sizeof(FE_status));
		}
		break;

	case CHDLC_EXCEP_COND_APP_INT_PEND:	/* 0x20 */
		process_chdlc_exception(card);
		break;

	default:
		interrupt_serviced = 0;
		break;
	}

wpc_isr_done:
	if (!interrupt_serviced){ 
        	unsigned char str[5];
		DEBUG_EVENT("%s: spurious interrupt 0x%02x!\n", card->devname,
	        	        flags.interrupt_info_struct.interrupt_type);
		bcopy(flags.global_info_struct.codename, 
			str, 
			sizeof(flags.global_info_struct.codename));
		str[sizeof(flags.global_info_struct.codename)] = '\0';
		DEBUG_EVENT("%s: Code name:    %s\n", card->devname, str);
		bcopy(flags.global_info_struct.codeversion, 
			str, 
			sizeof(flags.global_info_struct.codeversion));
		str[sizeof(flags.global_info_struct.codeversion)] = '\0';
		DEBUG_EVENT("%s: Code version: %s\n", card->devname, str);
	}

	card->in_isr = 0;
	card->hw_iface.poke_byte(card->hw, card->intr_type_off, 0x00);

}

/* 
 * ============================================================================
 * Receive interrupt handler.
 */
static void rx_intr (sdla_t* card)
{
	struct ifnet* 	           	ifp = NULL;
	chdlc_private_area_t* 	   	chdlc_priv_area = NULL;
	SHARED_MEMORY_INFO_STRUCT 	  	flags;
	CHDLC_DATA_RX_STATUS_EL_STRUCT	rxbuf;
	struct mbuf*		 	ipkt = NULL; 
	unsigned long 		   	addr = 0;
	char* 			   	data_ptr = NULL;
	unsigned 			   	len = 0, api_hdr_len = 0;
	int 			   	udp_type;
#if defined(__OpenBSD__) || (__FreeBSD_version < 501000)
	int s = 0;
#endif
	
	card->hw_iface.peek(card->hw, card->flags_off, &flags, sizeof(flags));
	card->hw_iface.peek(card->hw, card->rxmb_off, &rxbuf, sizeof(rxbuf));
	if (rxbuf.opp_flag != 0x01){ 
		unsigned char str[5];
		DEBUG_EVENT("%s: corrupted Rx buffer @ 0x%x, flag = 0x%02x!\n", 
				card->devname, (unsigned)card->rxmb_off, rxbuf.opp_flag);
		bcopy(flags.global_info_struct.codename, str,  
		      sizeof(flags.global_info_struct.codename));
		str[sizeof(flags.global_info_struct.codename)] = '\0';
	        DEBUG_EVENT("Code name:    %s\n", str);
		bcopy(flags.global_info_struct.codeversion, str,  
		      sizeof(flags.global_info_struct.codeversion));
		str[sizeof(flags.global_info_struct.codeversion)] = '\0';
	        DEBUG_EVENT("Code version: %s\n", str);

		/* 
		** 3.10.2000 RX_CORR
		** This case not suppose to be, but if it is, disable interrupt 
		** for current router. 
		*/
		DEBUG_EVENT("%s: Critical router failure...\n", card->devname);
		DEBUG_EVENT("Please contact your Sangoma represention!!\n");
		chdlc_set_intr_mode(card, 0);
		return;
	}
   
	if (card->wandev.if_head == NULL || card->wandev.if_head->ifp == NULL){
		DEBUG_EVENT("%s: Internal Error (line no. %d)!\n", 
						card->devname, __LINE__);
		goto rx_intr_end;
	}

	ifp = card->wandev.if_head->ifp;
	chdlc_priv_area = ifp->if_softc;
	addr = rxbuf.ptr_data_bfr;
	len  = rxbuf.frame_length;
	if (card->wandev.config_id == WANCONFIG_MPPP){
		/* Take off two CRC bytes */
		len -= 2;
	}
	if (len <= 0){
		goto rx_intr_end;
	}

	api_hdr_len = 0;

	if (card->u.c.usedby == API){
		api_hdr_len = sizeof(api_rx_hdr_t);
	}

	/*
	** Check that we have anough space in rx buffer.
	*/
	if (len + api_hdr_len > MAX_PACKET_SIZE){
		DEBUG_EVENT("%s: Received packet is too long (%d bytes)!\n", 
				card->devname, len + api_hdr_len);
		goto rx_intr_end;
	}
    
	if (addr + len <= card->u.c.rx_top_off + 1){
		card->hw_iface.peek(card->hw, addr, &card->rx_data[api_hdr_len], len);
		addr += len;
	}else{
		int first_piece_len = card->u.c.rx_top_off + 1 - addr;
		int second_piece_len = len - first_piece_len;

		card->hw_iface.peek(card->hw, addr, &card->rx_data[api_hdr_len], first_piece_len);
		card->hw_iface.peek(card->hw, card->u.c.rx_base_off, 
	                  	&card->rx_data[api_hdr_len]+first_piece_len, second_piece_len);
		addr = card->u.c.rx_base_off + second_piece_len; 
	}
	if (card->u.c.usedby == API){	/* Update API header */
		((api_rx_hdr_t*)card->rx_data)->error_flag = rxbuf.error_flag;
		((api_rx_hdr_t*)card->rx_data)->time_stamp = rxbuf.time_stamp;
	}
	/* Update packet length (For API, length will include header length) */
	udp_type = udp_pkt_type(&card->rx_data[api_hdr_len], card);
	if (udp_type == UDP_CPIPE_TYPE){ 
		data_ptr = card->rx_data + api_hdr_len;
	}else{
		len += api_hdr_len;
		data_ptr = card->rx_data;
	}

	ipkt = sdla_get(ifp, data_ptr, len);

	if (ipkt == NULL){
		DEBUG_EVENT("%s: Failed to allocate mbuf (RX_INTR)!!\n", 
					card->devname);
		ifp->if_iqdrops++;	
		goto rx_intr_end;
	}

	wan_bpf_report(ifp, ipkt, 1);
	if (udp_type == UDP_CPIPE_TYPE){ 
		if (store_udp_mgmt_pkt(UDP_PKT_FRM_NETWORK, card, ipkt, ifp, 
					  chdlc_priv_area)){
			card->hw_iface.enable_byte(card->hw, card->intr_perm_off, APP_INT_ON_TIMER);
		}
	}else{ 
        	/* Ok, this is a regular packet - do what's said above */
	    
		if (card->wandev.config_id == WANCONFIG_CHDLC){
	        	/* 
		        ** Schedule IP input routine. Note that it makes much
        		** more sence to check first if the protocol queue is
		     	** full - and then to schedule the input routine, but
     			** i took a look at kernel sources for ether_input() 
	        	** [ /usr/src/sys/net/if_ethersubr.c ] and they do it
	        	** in this order - so i figured they know better.
		        */
#if (__FreeBSD_version >= 501000)
			netisr_queue(NETISR_IP, ipkt);
#else
			s = splimp();
			schednetisr(NETISR_IP);
			if (IF_QFULL(&ipintrq)){
				DEBUG_RX("%s: IP queue is full, drop packet!\n",
							card->devname);
				/* oh, no - IP queue is full - well - we'll
				** try again later 
				*/ 
				IF_DROP(&ipintrq);
				wan_skb_free(ipkt);	/*m_freem(ipkt);*/
				splx(s);
				goto rx_intr_end;
			}
 
			/* ok enqueue the packet */
			IF_ENQUEUE(&ipintrq, ipkt);
			splx(s);
#endif
        		/* update interface statistics */
			ifp->if_ipackets++;
			ifp->if_ibytes += len;
		}else{
			WAN_SPPP_INPUT(ifp,ipkt);
		}
  
    	}

rx_intr_end:

	/* Release buffer element and calculate a pointer to the next one */
	rxbuf.opp_flag = 0x00;
	card->hw_iface.poke(card->hw, card->rxmb_off, &rxbuf, sizeof(rxbuf));
	card->rxmb_off += sizeof(CHDLC_DATA_RX_STATUS_EL_STRUCT);
	if (card->rxmb_off > card->u.c.rxbuf_last_off){
		card->rxmb_off = card->u.c.rxbuf_base_off;
	}
	return;
}

/* 
 * ============================================================================
 * Miscellaneous Functions
 *	- set_chdlc_config() used to set configuration options on the board
 */
static int set_chdlc_config(sdla_t *card)
{
	struct ifnet*			ifp = NULL;
	chdlc_private_area_t*		chdlc_priv_area = NULL; 
	CHDLC_CONFIGURATION_STRUCT	cfg;

	if (card->wandev.if_head == NULL || card->wandev.if_head->ifp == NULL){
		DEBUG_EVENT("%s: Internal Error (line no. %d)!\n", 
						card->devname, __LINE__);
		return ENODEV;
	}
	ifp = card->wandev.if_head->ifp;
	chdlc_priv_area = ifp->if_softc;	
	bzero(&cfg, sizeof(CHDLC_CONFIGURATION_STRUCT));

	if (card->wandev.clocking){
		cfg.baud_rate = card->wandev.bps;
	}

	cfg.line_config_options = 
			(card->wandev.interface == WANOPT_RS232) ?
			INTERFACE_LEVEL_RS232 : INTERFACE_LEVEL_V35;

	cfg.modem_config_options	    = 0;     
	cfg.modem_status_timer	    = 100;   
	cfg.CHDLC_protocol_options	    = card->u.c.protocol_options;
	cfg.percent_data_buffer_for_Tx  = 50;
	cfg.CHDLC_statistics_options    = 0;
	cfg.max_CHDLC_data_field_length = card->wandev.mtu;
	cfg.transmit_keepalive_timer    = card->u.c.kpalv_tx;
	cfg.receive_keepalive_timer	    = card->u.c.kpalv_rx;
	cfg.keepalive_error_tolerance   = card->u.c.kpalv_err;
	cfg.SLARP_request_timer	    = card->u.c.slarp_timer;

	if (cfg.SLARP_request_timer){ 
		cfg.IP_address	= 0;
		cfg.IP_netmask	= 0;
	}else{

        	cfg.IP_address	= chdlc_priv_area->IP_address;
	        cfg.IP_netmask	= chdlc_priv_area->IP_netmask;

		/* FIXME: We must re-think this message in next release */
		/*if ((cfg->IP_address & 0x000000FF) > 2)
		 *{
		 *   log(LOG_WARN, "WARNING: %s configured win an invalid\n",
		 *						card->devname);
		 *   log(LOG_WARN, "local IP address. Slarp pragmatic will fail.\n");
		 *   log(LOG_WARN, "IP address should be of the format A.B.C.1 or \n");
		 *   log(LOG_WARN, "A.B.C.2\n");
		 *}
	 	 */
	}
	return chdlc_configure(card, &cfg);
}

/* 
 * ============================================================================
 * Process global exception condition
 */
static int process_global_exception(sdla_t *card)
{
	wan_mbox_t*	mbox = &card->wan_mbox;
	int 		err;

	WAN_MBOX_INIT(mbox);
	mbox->wan_data_len = 0;
	mbox->wan_command = READ_GLOBAL_EXCEPTION_CONDITION;
	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);

	if (err != WAN_CMD_TIMEOUT){
		switch(err) {
		case EXCEP_MODEM_STATUS_CHANGE:
			if (IS_56K_CARD(card)){
				FRONT_END_STATUS_STRUCT FE_status;

				card->hw_iface.peek(card->hw,
						    card->fe_status_off, 
						    &FE_status,
						    sizeof(FE_status));

				card->wandev.RR8_reg_56k = 
					FE_status.FE_U.stat_56k.RR8_56k;	
				card->wandev.RRA_reg_56k = 
					FE_status.FE_U.stat_56k.RRA_56k;	
				card->wandev.RRC_reg_56k = 
					FE_status.FE_U.stat_56k.RRC_56k;	
				sdla_56k_alarm(card, 0);

				break;
			}

			if (IS_TE1_CARD(card)){
				/* TE1 T1/E1 interrupt */
				sdla_te_intr(card); 
				break;
			}	
			DEBUG_EVENT("%s: Modem status change\n", card->devname);
			switch(mbox->wan_data[0] & (WAN_DCD_HIGH | WAN_CTS_HIGH)){ 
			case (WAN_DCD_HIGH):
				DEBUG_EVENT("%s: DCD high, CTS low\n", card->devname);
				break;
			case (WAN_CTS_HIGH):
				DEBUG_EVENT("%s: DCD low, CTS high\n", card->devname);       
				break;
			case ((WAN_DCD_HIGH | WAN_CTS_HIGH)):
				DEBUG_EVENT("%s: DCD high, CTS high\n", card->devname);
				break;
			default:
				DEBUG_EVENT("%s: DCD low, CTS low\n", card->devname);
				break;
			}
			break;

		case EXCEP_TRC_DISABLED:
			DEBUG_EVENT("%s: Line trace disabled\n", card->devname);
			break;

		case EXCEP_IRQ_TIMEOUT:
			DEBUG_EVENT("%s: IRQ timeout occurred\n", card->devname);
			break;

		case EXCEP_CSU_DSU_STATE_CHANGE:
			break;

		default:
			DEBUG_EVENT("%s: global exception (err=%02x)\n", card->devname, err);
			break;
		}
	}
	return 0;
}


/*
 * ============================================================================
 * Process chdlc exception condition
 */
static int process_chdlc_exception(sdla_t *card)
{
	struct ifnet*		ifp = NULL;
	chdlc_private_area_t*	chdlc_priv_area = NULL;
	wan_mbox_t* 		mbox = &card->wan_mbox;
	int 			err;

	if (card->wandev.if_head == NULL || card->wandev.if_head->ifp == NULL){
		DEBUG_EVENT("%s: Internal Error (line no. %d)!\n", 
						card->devname, __LINE__);
	        return ENODEV;
	}
	ifp = card->wandev.if_head->ifp;	
	chdlc_priv_area = ifp->if_softc;
	WAN_MBOX_INIT(mbox);
	mbox->wan_data_len = 0;
	mbox->wan_command = READ_CHDLC_EXCEPTION_CONDITION;
	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);

	if (err != WAN_CMD_TIMEOUT){ 
		switch (err){ 
		case EXCEP_LINK_ACTIVE:
			DEBUG_EVENT("%s: CHDLC link became active ...\n", 
							card->devname);
			/* protocol is up - now if the slarp timer is zero, that 
			** means that IP address was set at the beginning and 
			** we are able for data transfers, so we will bring the  
			** interface up. If the slarp timer of station is not zero 
			** - we have to wait till we discover out IP address ... 
			*/
			if (card->u.c.slarp_timer == 0){
				ifp->if_flags |= (IFF_UP|IFF_RUNNING);
				DEBUG_EVENT("%s: Router has been brought up.\n", 
							card->devname); 
				chdlc_priv_area->router_up_time = wan_getcurrenttime();
			} 	
			wanpipe_set_state(card, WAN_CONNECTED);
			chdlc_priv_area->common.state = WAN_CONNECTED;
			break;

		case EXCEP_LINK_INACTIVE_MODEM:
			DEBUG_EVENT("%s: Link is down because of invalid modem state\n",
							card->devname);
			wanpipe_set_state(card, WAN_DISCONNECTED);
			chdlc_priv_area->common.state = WAN_DISCONNECTED;
			unconfigure_ip(card);
			/* Start debugging */
			WAN_DEBUG_START(card);
			break;

		case EXCEP_LINK_INACTIVE_KPALV:
			DEBUG_EVENT("%s: link is down due to the station not receiving SLARP keepalive frames ...\n", 
							card->devname); 
			wanpipe_set_state(card, WAN_DISCONNECTED);
			chdlc_priv_area->common.state = WAN_DISCONNECTED;
			DEBUG_EVENT("%s: Keepalive timer expired.\n", card->devname); 
			unconfigure_ip(card);
			/* Start debugging */
			WAN_DEBUG_START(card);
			break;

		case EXCEP_IP_ADDRESS_DISCOVERED:
			DEBUG_EVENT("%s: IP has been discovered ...\n", card->devname);
			if (configure_ip(card)){ 
				return -1;
			}
			ifp->if_flags |= (IFF_UP|IFF_RUNNING);
			break;

		case EXCEP_LOOPBACK_CONDITION:
			DEBUG_EVENT("%s: Loopback Condition Detected.\n", card->devname);
			ifp->if_flags &= ~(IFF_UP|IFF_RUNNING);
			break;

		case NO_CHDLC_EXCEP_COND_TO_REPORT:
			DEBUG_EVENT("%s: No exceptions reported.\n", card->devname);
			break;
		}
	}
	return 0;
}

/* 
 * ============================================================================
 * Configure IP from SLARP negotiation
 * This adds dynamic routes when SLARP has provided valid addresses
 */
static int configure_ip (sdla_t* card)
{
	struct ifnet*		ifp = NULL;
	chdlc_private_area_t*	chdlc_priv_area = NULL;
	int 			err = 0;

	if (card->wandev.if_head == NULL || card->wandev.if_head->ifp == NULL){
		DEBUG_EVENT("%s: Internal Error (line no. %d)!\n", 
						card->devname, __LINE__);
		return ENODEV; 
	}
	ifp = card->wandev.if_head->ifp;
	chdlc_priv_area = ifp->if_softc;
	/* set to discover */
	if (card->u.c.slarp_timer != 0x0){
		CHDLC_CONFIGURATION_STRUCT	cfg;

	        err = chdlc_read_configure(card, &cfg);
		if (err != WAN_CMD_OK){ 
			return -1;
		}

		/* According to Cisco HDLC, if the point-to-point address is
		 * A.B.C.1, then we are the opposite (A.B.C.2), and vice-versa,
		 * ie 2 and 1 respectively 
		 */
        	chdlc_priv_area->IP_address = cfg.IP_address;
	        chdlc_priv_area->IP_netmask = cfg.IP_netmask;
	}

	/* Set flag to add route */ 
	chdlc_priv_area->route_status = ADD_ROUTE;

	/* The idea here is to add the route in the poll routine.
	 * This way, we aren't in interrupt context when adding routes 
	 */
	card->poll = process_route;		/* process_route(card); */
	ifp->if_timer = 1;
    
	return 0;
}

/* 
 * ============================================================================
 * Un-Configure IP negotiated by SLARP
 * This removes dynamic routes when the link becomes inactive.
 */
static int unconfigure_ip (sdla_t* card)
{
	struct ifnet*		ifp = NULL;
	chdlc_private_area_t*	chdlc_priv_area = NULL;

	if (card->wandev.if_head == NULL || card->wandev.if_head->ifp == NULL){
		DEBUG_EVENT("%s: Internal Error (line no. %d)!\n", 
						card->devname, __LINE__);
		return ENODEV;
	}
	ifp = card->wandev.if_head->ifp;
	chdlc_priv_area = ifp->if_softc;

	if (chdlc_priv_area->route_status == ROUTE_ADDED){ 
		chdlc_priv_area->route_status = REMOVE_ROUTE;
		/* 
		 * The idea here is to delete the route in the poll routine.
	         * This way, we aren't in interrupt context when adding routes 
		 */
        	card->poll = process_route;		/* process_route(card); */
		ifp->if_timer = 1;
	}
	return 0;
}

/* 
 * ============================================================================
 * Routine to add/remove routes 
 * Called like a polling routine when Routes are flagged to be added/removed.
 */
static void process_route (void *arg)
{
	sdla_t*			card = (sdla_t*)arg;
	struct ifnet*		ifp = NULL;
	chdlc_private_area_t*	chdlc_priv_area = NULL;
	uint32_t 		local_IP_addr = 0, 
				remote_IP_addr = 0, 
				IP_netmask = 0, 
				IP_addr = 0;
	struct ifaddr*		ifa = NULL; 
	struct sockaddr_in*	saddr = NULL;
	int 			err = 0;

	if (card->wandev.if_head == NULL || card->wandev.if_head->ifp == NULL){
		DEBUG_EVENT("%s: Internal Error (line no. %d)!\n", 
						card->devname, __LINE__);
		return; 
	}
	ifp = card->wandev.if_head->ifp;
	chdlc_priv_area = ifp->if_softc;
	ifa = WAN_TAILQ_FIRST(ifp); 
	if ((chdlc_priv_area->route_status == ADD_ROUTE) &&
		((chdlc_priv_area->IP_address & ~chdlc_priv_area->IP_netmask) > 2)){ 
	        /* check that the received IP address is A.B.C.1 or A.B.C.2 
		 */
	        DEBUG_EVENT("%s: Dynamic route failure.\n", card->devname); 
        	if(card->u.c.slarp_timer){ 
			DEBUG_EVENT("%s: Bad IP address %s received ", 
				        card->devname, 
					in_ntoa(ntohl(chdlc_priv_area->IP_address)));
			DEBUG_EVENT("from remote station.\n");
		}else{ 
			DEBUG_EVENT("%s: Bad IP address %s issued",
					card->devname, 
					in_ntoa(ntohl(chdlc_priv_area->IP_address)));
			DEBUG_EVENT("to remote station.\n Local");
			DEBUG_EVENT("IP address must be A.B.C.1");	
			DEBUG_EVENT("or A.B.C.2.\n");
		}
		/* Remove the route due to the IP address error condition */
		chdlc_priv_area->route_status = REMOVE_ROUTE;
		err = 1;
	}

	/* If we are removing a route with bad IP address, then use the 
	 * local configured IP addresses 
	 */
	if ((chdlc_priv_area->route_status == REMOVE_ROUTE) && err){
		struct sockaddr_in *saddr = NULL;
		/* do not remove a bad route that has already been removed */
		if (chdlc_priv_area->route_removed){
			card->poll = NULL;
			return;
		}
		if ((saddr = (struct sockaddr_in *)ifa->ifa_addr) != NULL){ 
	        	local_IP_addr = htonl(saddr->sin_addr.s_addr);
		}
		if ((saddr = (struct sockaddr_in *)ifa->ifa_dstaddr) != NULL){ 
			remote_IP_addr = htonl(saddr->sin_addr.s_addr);
		}
		if ((saddr = (struct sockaddr_in *)ifa->ifa_netmask) != NULL){ 
			IP_netmask = htonl(saddr->sin_addr.s_addr);
		}
	}else{
		IP_netmask = ntohl(chdlc_priv_area->IP_netmask);
		remote_IP_addr = ntohl(chdlc_priv_area->IP_address);
		local_IP_addr = (remote_IP_addr & IP_netmask) + 
						(~remote_IP_addr & ntohl(0x0003));

		if (!card->u.c.slarp_timer){ 
			IP_addr = local_IP_addr;
			local_IP_addr = remote_IP_addr;
			remote_IP_addr = IP_addr;
		}
	}
	DEBUG_EVENT("%s: local_IP_addr: %s\n", 
			card->devname, in_ntoa(local_IP_addr));
	DEBUG_EVENT("%s: remote_IP_addr: %s\n", 
			card->devname, in_ntoa(remote_IP_addr));
  
	/* ok, wicked - now we need to update the interface and routing 
	 * tables ...							
	 */
	if ((ifa = WAN_TAILQ_NEXT(ifa)) == NULL){ 
        	card->poll = NULL;
	        return;
	}

	switch (chdlc_priv_area->route_status){ 
	case ADD_ROUTE:
		if (card->u.c.slarp_timer){ 
			/* Deleting old routing entry */
			err = rtrequest((int)RTM_DELETE,	/* action      */
					ifa->ifa_dstaddr,	/* destination */
					ifa->ifa_addr,		/* gateway     */
					ifa->ifa_netmask,	/* mask        */
					RTF_HOST,		/* flag        */
					(struct rtentry **)0);

			/* set up new local and remote IP's and new netmask */
			saddr = (struct sockaddr_in *)ifa->ifa_addr;
			saddr->sin_addr.s_addr = local_IP_addr;

			saddr = (struct sockaddr_in *)ifa->ifa_dstaddr;
			saddr->sin_addr.s_addr = remote_IP_addr;

			saddr = (struct sockaddr_in *)ifa->ifa_netmask;
			saddr->sin_addr.s_addr = IP_netmask;

			/* add new routing entry */
			err = rtrequest((int)RTM_ADD, 		/* action */
					ifa->ifa_dstaddr,		/* destination */
					ifa->ifa_addr,		/* gateway */
					ifa->ifa_netmask,		/* mask */
					RTF_HOST,			/* flag */
					(struct rtentry **)0);
			if (err){
				DEBUG_EVENT("%s: Add route %s is failed (%d)\n",
						card->devname, in_ntoa(remote_IP_addr), err); 
		}else{
			chdlc_priv_area->route_status = ROUTE_ADDED;
			DEBUG_EVENT("%s: Dynamic route added\n", card->devname);
			DEBUG_EVENT("%s:   Local IP addr = %s\n",
					card->devname, in_ntoa(local_IP_addr));
			DEBUG_EVENT("%s:   Remote IP addr = %s\n",
					card->devname, in_ntoa(remote_IP_addr));
			chdlc_priv_area->route_removed = 0;	
			/* perfect! now we're ready to go - mark interface up 
			 * and running.
			 */
			ifp->if_flags |= (IFF_UP|IFF_RUNNING);
			/* update the up time ... */
			chdlc_priv_area->router_up_time = wan_getcurrenttime();
		}
	}
        break;

	case REMOVE_ROUTE:
		/* Deleting current routing entry */
		err = rtrequest((int)RTM_DELETE,	/* action      */
				ifa->ifa_dstaddr,	/* destination */
				ifa->ifa_addr,		/* gateway     */
				ifa->ifa_netmask,	/* mask        */
				RTF_HOST,		/* flag        */
				(struct rtentry **)0);
	
		if (err){
			DEBUG_EVENT("%s: Remove route %s failed (%d)\n",
					card->devname, 
					in_ntoa(remote_IP_addr), err);
		}else{
			chdlc_priv_area->route_status = NO_ROUTE;
			DEBUG_EVENT("%s: Dynamic route removed %s.\n", 
					card->devname, in_ntoa(local_IP_addr)); 
			/* mark interface inactive */
			ifp->if_flags &= ~(IFF_UP|IFF_RUNNING);
			chdlc_priv_area->route_removed = 1;	
			/* Set destination address to zero */ 
			saddr = (struct sockaddr_in *)ifa->ifa_dstaddr;
			saddr->sin_addr.s_addr = 0;
		}
		break;
	}

	/* Once we've processed the route, stop polling */
	card->poll = NULL;
}

/*
 * =============================================================================
 * Store a UDP management packet for later processing.
 */
static int 
store_udp_mgmt_pkt(char udp_pkt_src, sdla_t* card, struct mbuf *m,
                   struct ifnet *ifp, chdlc_private_area_t *chdlc_priv_area)
{
	int udp_pkt_stored = 0;
	int api_hdr_len = 0;

	/*
	 * If current card running API and UDP packet arrived from NETWORK,
	 * then remove API header, which was added in rx_intr function.
	 * Otherwise, api_hdr_len will be 0. 
	 */
	if (udp_pkt_src == UDP_PKT_FRM_NETWORK && card->u.c.usedby == API)
		api_hdr_len = sizeof(api_rx_hdr_t);

	if(!chdlc_priv_area->udp_pkt_lgth && (m->m_len <= MAX_LGTH_UDP_MGNT_PKT)){ 
		DEBUG_UDP("%s: Storing UDP packet in private area..\n", 
				card->devname); 
		chdlc_priv_area->udp_pkt_lgth = m->m_len - api_hdr_len;
		chdlc_priv_area->udp_pkt_src = udp_pkt_src;
		bcopy(mtod(m, caddr_t) + api_hdr_len, 
			chdlc_priv_area->udp_pkt_data, m->m_len - api_hdr_len);
		chdlc_priv_area->timer_int_enabled = TMR_INT_ENABLED_UDP;
		udp_pkt_stored = 1;
	}else{
		DEBUG_EVENT("%s: Previous UDP packet is still not sended!\n", 
				card->devname); 
	}

	wan_skb_free(m);	/*m_freem(m);*/
	return(udp_pkt_stored);
}

/*
 * Execute CHDLC udp commands
 */
static int chdlc_udp_ioctl(struct ifnet* ifp, caddr_t data)
{
	chdlc_private_area_t*	chdlc_priv_area = ifp->if_softc;
    	sdla_t* 		card = chdlc_priv_area->card;
	struct ifreq*		ifr = (struct ifreq*)data;
	wan_udp_hdr_t		udp_hdr;
	int 			error = 0;

	if (copyin((void*)ifr->ifr_data, (void*)&udp_hdr, sizeof(wan_udp_hdr_t)) < 0){
		DEBUG_EVENT("%s: Failed to copy data from user to kernel space\n",
				card->devname);
		return -EFAULT;
	}

	error = process_udp_cmd(ifp, &udp_hdr);	

	if (copyout((void*)&udp_hdr, (void*)ifr->ifr_data, sizeof(wan_udp_hdr_t)) < 0){
		DEBUG_EVENT("%s: Failed to copy data from kernel to user space\n",
				card->devname);
		return -EFAULT;
	}
	return 0;
}


/*
 *
 *
 */

static int process_udp_cmd(struct ifnet* ifp, wan_udp_hdr_t* udp_hdr)
{
    	chdlc_private_area_t*		chdlc_priv_area = ifp->if_softc;
    	sdla_t* 			card = chdlc_priv_area->card;
    	wan_mbox_t*			mbox = &card->wan_mbox;
    	SHARED_MEMORY_INFO_STRUCT	flags;
        unsigned long 			trace_status_cfg_addr = 0;
        TRACE_STATUS_EL_CFG_STRUCT	trace_cfg_struct;
        TRACE_STATUS_ELEMENT_STRUCT 	trace_element_struct;
    	unsigned short 			buffer_length = 0, 
					real_len = 0;
    	unsigned long 			data_ptr = 0;
	unsigned int			frames = 0;
    	unsigned 			data_length = 0;
	int				err = 0;

    	udp_hdr->wan_udphdr_opp_flag = 0;

	card->hw_iface.peek(card->hw, 
			    card->flags_off, 
			    &flags, 
			    sizeof(SHARED_MEMORY_INFO_STRUCT));
        switch(udp_hdr->wan_udphdr_command){ 
        case CPIPE_ENABLE_TRACING:
		if (!chdlc_priv_area->TracingEnabled){ 
			/* OPERATE_DATALINE_MONITOR */
			WAN_MBOX_INIT(mbox);
		        mbox->wan_data_len = sizeof(LINE_TRACE_CONFIG_STRUCT);
			mbox->wan_command = SET_TRACE_CONFIGURATION;
	
			((LINE_TRACE_CONFIG_STRUCT *)mbox->wan_data)->
				trace_config = TRACE_ACTIVE; 
			/* Trace delay mode is not used because it slows
			 * down transfer and results in a standoff situation
		         * when there is a lot of data 
			 */
			((LINE_TRACE_CONFIG_STRUCT *)mbox->wan_data)->trace_config |= 
					udp_hdr->wan_udphdr_data[0];;
			((LINE_TRACE_CONFIG_STRUCT *)mbox->wan_data)->
					trace_deactivation_timer = 4000;
			err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
			if (err != WAN_CMD_OK){ 
				chdlc_error(card, err, mbox);
				card->TracingEnabled = 0;
				udp_hdr->wan_udphdr_return_code = err;
		    		udp_hdr->wan_udphdr_data_len = 0; 
       				break;
		        } 
	
			/* Get the base address of the trace element list */
			WAN_MBOX_INIT(mbox);
			mbox->wan_data_len = 0;
			mbox->wan_command = READ_TRACE_CONFIGURATION;
			err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
			if (err != WAN_CMD_OK){ 
				chdlc_error(card, err, mbox);
				chdlc_priv_area->TracingEnabled = 0;
				udp_hdr->wan_udphdr_return_code = err;
	    			udp_hdr->wan_udphdr_data_len = 0; 
				break;
			}
			trace_status_cfg_addr =((LINE_TRACE_CONFIG_STRUCT *)
					mbox->wan_data) -> ptr_trace_stat_el_cfg_struct;
	
			card->hw_iface.peek(card->hw, trace_status_cfg_addr,
					&trace_cfg_struct, sizeof(trace_cfg_struct));
			    
			chdlc_priv_area->start_trace_addr = 
					trace_cfg_struct.base_addr_trace_status_elements;

			chdlc_priv_area->number_trace_elements = 
					trace_cfg_struct.number_trace_status_elements;

			chdlc_priv_area->end_trace_addr = 
					(unsigned long)((TRACE_STATUS_ELEMENT_STRUCT *)
					chdlc_priv_area->start_trace_addr + 
					(chdlc_priv_area->number_trace_elements - 1));

			chdlc_priv_area->base_addr_trace_buffer = 
					trace_cfg_struct.base_addr_trace_buffer;

			chdlc_priv_area->end_addr_trace_buffer = 
					trace_cfg_struct.end_addr_trace_buffer;

		 	chdlc_priv_area->curr_trace_addr = 
					trace_cfg_struct.next_trace_element_to_use;

		    	chdlc_priv_area->available_buffer_space = 
					2000 - sizeof(struct ip) - sizeof(struct udphdr) -
					sizeof(wan_mgmt_t) - sizeof(wan_cmd_t) - 
					sizeof(wan_trace_info_t); 
		}  
		udp_hdr->wan_udphdr_return_code = WAN_CMD_OK;
		udp_hdr->wan_udphdr_data_len = 0; 
		chdlc_priv_area->TracingEnabled = 1;
		break;
	   
	case CPIPE_DISABLE_TRACING:
		if (chdlc_priv_area->TracingEnabled){ 
	        	/* OPERATE_DATALINE_MONITOR */
		        mbox->wan_data_len = sizeof(LINE_TRACE_CONFIG_STRUCT);
		        mbox->wan_command = SET_TRACE_CONFIGURATION;
			((LINE_TRACE_CONFIG_STRUCT *)mbox->wan_data)->
				trace_config = TRACE_INACTIVE;
			err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
		}		

		chdlc_priv_area->TracingEnabled = 0;
		udp_hdr->wan_udphdr_return_code = WAN_CMD_OK;
		udp_hdr->wan_udphdr_data_len = 0; 
		break;
	   
        case CPIPE_GET_TRACE_INFO:
		if (!chdlc_priv_area->TracingEnabled){ 
			udp_hdr->wan_udphdr_return_code = 1;
			udp_hdr->wan_udphdr_data_len = 0; 
			break;
		}
		udp_hdr->wan_udphdr_chdlc_ismoredata = 0x0;
		buffer_length = 0;	/* offset of packet already occupied */
		for (frames=0; frames < chdlc_priv_area->number_trace_elements;
	        				frames++){ 
			wan_trace_pkt_t *trace_pkt =
                		(wan_trace_pkt_t *)&udp_hdr->wan_udphdr_data[buffer_length];
			card->hw_iface.peek(card->hw, chdlc_priv_area->curr_trace_addr,
		                               (unsigned char *)&trace_element_struct,
	        		               sizeof(TRACE_STATUS_ELEMENT_STRUCT));
			if (trace_element_struct.opp_flag == 0x00){ 
				break;
			}

			/* get pointer to real data */
			data_ptr = trace_element_struct.ptr_data_bfr;

			/* See if there is actual data on the trace buffer */
			if (data_ptr){
				data_length = trace_element_struct.trace_length;
			}else{
				data_length = 0;
				udp_hdr->wan_udphdr_chdlc_ismoredata = 0x01;
			}
	
	        	if( (chdlc_priv_area->available_buffer_space - buffer_length)
			      < ( sizeof(wan_trace_pkt_t) + data_length)){ 
	        		/* indicate there are more frames on board & exit */
	        		udp_hdr->wan_udphdr_chdlc_ismoredata = 0x01;
				break;
			}
			trace_pkt->status = trace_element_struct.trace_type;
			trace_pkt->time_stamp = trace_element_struct.trace_time_stamp;
			trace_pkt->real_length = trace_element_struct.trace_length;

			/* see if we can fit the frame into the user buffer */
			real_len = trace_pkt->real_length;

			if (data_ptr == 0){ 
				trace_pkt->data_avail = 0x00;
			}else{ 
				unsigned tmp = 0;

				/* get the data from circular buffer
				 * must check for end of buffer 
				 */
				trace_pkt->data_avail = 0x01;

				if ((data_ptr + real_len) >
						chdlc_priv_area->end_addr_trace_buffer + 1){ 
					tmp = chdlc_priv_area->end_addr_trace_buffer - 
                                        			data_ptr + 1;
					card->hw_iface.peek(card->hw, data_ptr, trace_pkt->data, tmp);
					data_ptr = chdlc_priv_area->base_addr_trace_buffer;
				}
				card->hw_iface.peek(card->hw, data_ptr,
							&trace_pkt->data[tmp], real_len - tmp);
			}	

			/* zero the opp flag to show we got the frame */
			card->hw_iface.poke_byte(card->hw, chdlc_priv_area->curr_trace_addr, 0x00);
			/* now move onto the next frame */
			chdlc_priv_area->curr_trace_addr += 
						sizeof(TRACE_STATUS_ELEMENT_STRUCT);

			/* check if we went over the last address */
			if (chdlc_priv_area->curr_trace_addr > 
		        		           chdlc_priv_area->end_trace_addr){
				chdlc_priv_area->curr_trace_addr = 
							chdlc_priv_area->start_trace_addr;
			}

			/* update buffer length */ 
			if (trace_pkt->data_avail == 0x01){ 
				buffer_length += real_len - 1;
			}

			/* for the header */
			buffer_length += sizeof(wan_trace_pkt_t);

		}  /* For Loop */
		if (frames == chdlc_priv_area->number_trace_elements){
			udp_hdr->wan_udphdr_chdlc_ismoredata = 0x01;
		}	
		udp_hdr->wan_udphdr_chdlc_num_frames = frames;

		/* set the data length */
		udp_hdr->wan_udphdr_data_len = buffer_length; 

		/* set return code */
		udp_hdr->wan_udphdr_return_code = WAN_CMD_OK; 
		break;
	
	case CPIPE_FT1_READ_STATUS:
		((unsigned char *)udp_hdr->wan_udphdr_data )[0] =
				flags.FT1_info_struct.parallel_port_A_input;

		((unsigned char *)udp_hdr->wan_udphdr_data )[1] =
				flags.FT1_info_struct.parallel_port_B_input;
		udp_hdr->wan_udphdr_return_code = WAN_CMD_OK;
		udp_hdr->wan_udphdr_data_len = 2; 
		break;
	
        case CPIPE_ROUTER_UP_TIME:
	    chdlc_priv_area->router_up_time = wan_getcurrenttime() - 
	                        chdlc_priv_area->router_start_time;
	    *(unsigned long *)&udp_hdr->wan_udphdr_data[0] = 
	                        chdlc_priv_area->router_up_time;
	    udp_hdr->wan_udphdr_data_len = sizeof(unsigned long); 
            udp_hdr->wan_udphdr_return_code = WAN_CMD_OK;
	    break;

        case FT1_MONITOR_STATUS_CTRL:
	    /* Enable FT1 MONITOR STATUS */
       	    if ((udp_hdr->wan_udphdr_data[0] & ENABLE_READ_FT1_STATUS) ||  
		(udp_hdr->wan_udphdr_data[0] & ENABLE_READ_FT1_OP_STATS)){
	        if (rCount++ != 0){  
		    udp_hdr->wan_udphdr_return_code = WAN_CMD_OK;
	            udp_hdr->wan_udphdr_data_len = 1; 
		    break;
   	         }
   	    }

     	    /* Disable FT1 MONITOR STATUS */
      	    if (udp_hdr->wan_udphdr_data[0] == 0){
      	        if (--rCount != 0){ 
		    udp_hdr->wan_udphdr_return_code = WAN_CMD_OK;
	            udp_hdr->wan_udphdr_data_len = 1; 
		    break;
    	        } 
    	    } 	
	    goto udp_dflt_cmd;
	    break;

	case WAN_FE_SET_LB_MODE:
	case WAN_GET_MEDIA_TYPE:
	case WAN_FE_GET_STAT:
	case WAN_FE_GET_CFG:
 	case WAN_FE_FLUSH_PMON:
    	    if (IS_TE1_CARD(card)){
		sdla_te_udp(card, &udp_hdr->wan_cmd, &udp_hdr->wan_udphdr_data[0]);	
	    }else if (IS_56K_CARD(card)){
		sdla_56k_udp(card, &udp_hdr->wan_cmd, &udp_hdr->wan_udphdr_data[0]);	
	    }else{
		if (udp_hdr->wan_udphdr_command == WAN_GET_MEDIA_TYPE){
	    		udp_hdr->wan_udphdr_return_code = WAN_CMD_OK;
    	    		udp_hdr->wan_udphdr_data_len = sizeof(unsigned char); 
		}else{
	    		udp_hdr->wan_udphdr_return_code = WAN_UDP_INVALID_CMD;
		}	
	    }
	    break;
	    
 	case WAN_GET_PROTOCOL:
	    udp_hdr->wan_udphdr_chdlc_num_frames = WANCONFIG_CHDLC;
	    udp_hdr->wan_udphdr_return_code = WAN_CMD_OK;
    	    udp_hdr->wan_udphdr_data_len = 1;
	    break;

 	case WAN_GET_PLATFORM:
	    udp_hdr->wan_udphdr_data[0] = WAN_FREEBSD_PLATFORM;
	    udp_hdr->wan_udphdr_return_code = WAN_CMD_OK;
    	    udp_hdr->wan_udphdr_data_len = 1;
	    break;

	default:

udp_dflt_cmd:
	    /* it's a board command */
		WAN_MBOX_INIT(mbox);
	    mbox->wan_command = udp_hdr->wan_udphdr_command;
	    mbox->wan_data_len = udp_hdr->wan_udphdr_data_len;
	    if (mbox->wan_data_len){
          	memcpy(&mbox->wan_data, (unsigned char *) udp_hdr->wan_udphdr_data, mbox->wan_data_len);
	    }

	    /* run the command on the board */
            err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	    if (err != WAN_CMD_OK) {
	        chdlc_error(card, err, mbox);
	    }

	    /* copy the result back to our buffer */
	    udp_hdr->wan_udphdr_command  = mbox->wan_command;
	    udp_hdr->wan_udphdr_data_len = mbox->wan_data_len;
	    udp_hdr->wan_udphdr_return_code = mbox->wan_return_code;

            if (mbox->wan_data_len){
		memcpy(&udp_hdr->wan_udphdr_data[0], mbox->wan_data, 
								mbox->wan_data_len);
	    }
        } /* end of switch */


	return err;
}

/* Finished
 * ============================================================================
 * Process UDP management packet.
 */
static int 
process_udp_mgmt_pkt(sdla_t* card, struct ifnet *ifp, 
		     chdlc_private_area_t *chdlc_priv_area)
{
	struct mbuf*		m = NULL; 
	int 			udp_mgmt_req_valid = 1;
	wan_udp_pkt_t*		udp_pkt = NULL;
	unsigned char*		data_addr = NULL;
	unsigned 			data_length = 0;
	int 			err;

	udp_pkt = (wan_udp_pkt_t *)chdlc_priv_area->udp_pkt_data;

	if (chdlc_priv_area->udp_pkt_src == UDP_PKT_FRM_NETWORK) { 
		switch(udp_pkt->wan_udp_command) { 
		case READ_GLOBAL_STATISTICS:
		case READ_MODEM_STATUS:
		case READ_CHDLC_LINK_STATUS:
		case CPIPE_ROUTER_UP_TIME:
		case READ_COMMS_ERROR_STATS:
		case READ_CHDLC_OPERATIONAL_STATS:
		case READ_CHDLC_CONFIGURATION:
		case READ_CHDLC_CODE_VERSION:
		case WAN_GET_MEDIA_TYPE:
		case WAN_FE_GET_STAT:
			udp_mgmt_req_valid = 1;
			break;
		default:
			udp_mgmt_req_valid = 0;
			break;
		}
	}

	udp_pkt->wan_udp_opp_flag = 0;
	if(!udp_mgmt_req_valid){ 

		/* set length to 0 */
		udp_pkt->wan_udp_data_len = 0;

		/* set return code */
		udp_pkt->wan_udp_return_code = WAN_UDP_INVALID_NET_CMD;

	}else{ 

		err = process_udp_cmd(ifp, &udp_pkt->wan_udp_hdr);

	} /* end of else */

	/* Fill UDP TTL */
	udp_pkt->wan_ip_ttl = card->wandev.ttl; 

	data_length = wan_reply_udp(card, 
				chdlc_priv_area->udp_pkt_data, 
				udp_pkt->wan_udp_data_len);

	if (chdlc_priv_area->udp_pkt_src == UDP_PKT_FRM_NETWORK){ 
		chdlc_send(card, chdlc_priv_area->udp_pkt_data, data_length);
	}else{
#if defined(__OpenBSD__) || (__FreeBSD_version < 501000)
		int s = 0;	
#endif
		/*
		 * Ok, what we've got now in chdlc_priv_area->udp_pkt_data is a 
		 * fully formatted IP packet which consists of responce to the 
		 * UDP management request. The length of the packet is 
		 * 'data_length'. We have to free the original mbuf and
		 * allocate the new one. ( when i tried to re-use the same 
		 * mbuf with m_pullup i ended up crashing the system somewhere
		 * in IP stack after i passed the frame up - looks like they 
		 * free the buffer anyways if the driver does not ).
		 */  
		/* will run through the buffer */
		data_addr  = chdlc_priv_area->udp_pkt_data;	

		/*
		 * Build a new mbuf chain 
		 */

		m = sdla_get(ifp, data_addr, data_length);
		if (m == NULL){
			DEBUG_EVENT("%s: Failed to allocate mbuf (UDP)!!\n", 
					card->devname);
			return ENOMEM;
		}
		wan_bpf_report(ifp, m, 1);
		/* 
		 * Ok, at this point we've got an mbuf chain pointed by m which contains
		 * a UDP management reply - now we just have to pass it up the stack -
		 * and that's it !!!
		 * Note that we neither report those packets to BPF nore update any 
		 * stats. Those packets are meant to be invisible ;-)
		 * Also we run inside splimp already - do there's no need to call it 
		 * again.
		 */
#if (__FreeBSD_version >= 501000)
		netisr_queue(NETISR_IP, m);
#else
		s = splimp();
		schednetisr(NETISR_IP);
		if (IF_QFULL(&ipintrq)){ 
			/* oh, no - IP queue is full - well - screw it. We cant do much 
			 * here. 
			 */
			DEBUG_RX("%s: IP queue is full!\n", 
					card->devname);
			IF_DROP(&ipintrq);
			wan_skb_free(m);	/*m_freem(m);*/
			splx(s); 
			return -EINVAL;
		}

		/* ok enqueue the packet */
		IF_ENQUEUE(&ipintrq, m);
		splx(s);
#endif
		DEBUG_UDP("%s: UDP packet passed back to the stack\n", 
				card->devname);
	}

	chdlc_priv_area->udp_pkt_lgth = 0;

	return 0;
}

/* Finished
 * ============================================================================
 * Initialize Receive and Transmit Buffers.
 */

static void init_chdlc_tx_rx_buff( sdla_t* card, struct ifnet *ifp )
{
	CHDLC_CONFIGURATION_STRUCT 		cfg;
	CHDLC_TX_STATUS_EL_CFG_STRUCT	tx_config;
	CHDLC_RX_STATUS_EL_CFG_STRUCT	rx_config;
	unsigned long	tx_config_off;
	unsigned long	rx_config_off;
	int err = 0;

    err = chdlc_read_configure(card, &cfg);
    if(err != WAN_CMD_OK){ 
	return;
    }

    if(card->type == SDLA_S514){ 
	tx_config_off = cfg.ptr_CHDLC_Tx_stat_el_cfg_struct;
       	rx_config_off = cfg.ptr_CHDLC_Rx_stat_el_cfg_struct;

        /* Setup Head and Tails for buffers */
    	card->hw_iface.peek(card->hw, tx_config_off, &tx_config, sizeof(tx_config));
	card->u.c.txbuf_base_off = tx_config.base_addr_Tx_status_elements;
        card->u.c.txbuf_last_off = card->u.c.txbuf_base_off +  
		(tx_config.number_Tx_status_elements - 1) * sizeof(CHDLC_DATA_TX_STATUS_EL_STRUCT);

      	card->hw_iface.peek(card->hw, rx_config_off, &rx_config, sizeof(rx_config));
	card->u.c.rxbuf_base_off = rx_config.base_addr_Rx_status_elements;
        card->u.c.rxbuf_last_off = card->u.c.rxbuf_base_off +
		(rx_config.number_Rx_status_elements - 1) * sizeof(CHDLC_DATA_RX_STATUS_EL_STRUCT);

        /* Set up next pointer to be used */
        card->u.c.txbuf_off = tx_config.next_Tx_status_element_to_use;
        card->rxmb_off = rx_config.next_Rx_status_element_to_use;
    }else{
        tx_config_off = cfg.ptr_CHDLC_Tx_stat_el_cfg_struct % SDLA_WINDOWSIZE; 
	rx_config_off = cfg.ptr_CHDLC_Rx_stat_el_cfg_struct % SDLA_WINDOWSIZE;

        /* Setup Head and Tails for buffers */
       	card->hw_iface.peek(card->hw, tx_config_off, &tx_config, sizeof(tx_config));
	card->u.c.txbuf_base_off = tx_config.base_addr_Tx_status_elements % SDLA_WINDOWSIZE;
        card->u.c.txbuf_last_off = card->u.c.txbuf_base_off +  
		(tx_config.number_Tx_status_elements - 1) * sizeof(CHDLC_DATA_TX_STATUS_EL_STRUCT);

      	card->hw_iface.peek(card->hw, rx_config_off, &rx_config, sizeof(rx_config));
	card->u.c.rxbuf_base_off = rx_config.base_addr_Rx_status_elements % SDLA_WINDOWSIZE;
        card->u.c.rxbuf_last_off = card->u.c.rxbuf_base_off +
		(rx_config.number_Rx_status_elements - 1) * sizeof(CHDLC_DATA_RX_STATUS_EL_STRUCT);

        /* Set up next pointer to be used */
        card->u.c.txbuf_off = tx_config.next_Tx_status_element_to_use % SDLA_WINDOWSIZE;
        card->rxmb_off = rx_config.next_Rx_status_element_to_use % SDLA_WINDOWSIZE;
    }

    /* Setup Actual Buffer Start and end addresses */
    card->u.c.rx_base_off = rx_config.base_addr_Rx_buffer;
    card->u.c.rx_top_off  = rx_config.end_addr_Rx_buffer;

}

#ifdef CHDLC_OLD_INTR_TEST
/* 
 * ============================================================================
 * Perform Interrupt Test by running READ_CHDLC_CODE_VERSION command MAX_INTR
 * _TEST_COUNTER times.
 */
static int chdlc_intr_test(sdla_t* card)
{
	int 	err = 0, i;

	Intr_test_counter = 0;

	err = chdlc_set_intr_mode(card, APP_INT_ON_COMMAND_COMPLETE); 
	if (err == WAN_CMD_OK){ 
		wan_mbox_t*	mbox = &card->wan_mbox;
		WAN_MBOX_INIT(mbox);
		for (i = 0; i < MAX_INTR_TEST_COUNTER; i ++){ 
			mbox->wan_data_len  = 0;
			mbox->wan_command = READ_CHDLC_CODE_VERSION;
			err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
			if (err != WAN_CMD_OK){ 
				chdlc_error(card, err, mbox);
			}
		}
		err = chdlc_set_intr_mode(card, 0x0);
	}

	if (err || (Intr_test_counter < MAX_INTR_TEST_COUNTER)){
		DEBUG_EVENT("%s: Interrupt test failed (%d)\n",
					card->devname, Intr_test_counter);
		DEBUG_EVENT("%s: Please choose another interrupt\n",
					card->devname);
		return -EIO;
	}
	DEBUG_EVENT("%s: Interrupt test passed(%d)\n",
					card->devname, Intr_test_counter);
	return 0;
}
#endif

/* 
 * =============================================================================
 * Determine what type of UDP call it is. CPIPEAB ?
 */
static int udp_pkt_type(caddr_t ptr, sdla_t* card)
{
    wan_udp_pkt_t* 	udp_pkt = (wan_udp_pkt_t*)ptr; 
    unsigned int 	udp_port = card->wandev.udp_port;

    if ((!strncmp(udp_pkt->wan_udp_signature,CHDLC_UDP_SIGNATURE,CHDLC_UDP_SIGNATURE_LEN) ||
        !strncmp(udp_pkt->wan_udp_signature,GLOBAL_UDP_SIGNATURE,GLOBAL_UDP_SIGNATURE_LEN)) &&
	(udp_pkt->wan_ip_p == UDPMGMT_UDP_PROTOCOL) &&
	(udp_pkt->wan_udp_request_reply == UDPMGMT_REQUEST))
    {
	if (udp_pkt->wan_udp_dport == ntohs(udp_port)){
		return UDP_CPIPE_TYPE;
	}
	if (udp_pkt->wan_udp_command == WAN_GET_PROTOCOL){
		return UDP_CPIPE_TYPE;
	}
    }
    return UDP_INVALID_TYPE;
}


/* Finished
 * ============================================================================
 * Timer interrupt handler.
 */
void timer_intr(sdla_t *card)
{
	struct ifnet*		ifp = NULL;
	chdlc_private_area_t* 	chdlc_priv_area = NULL;

	if (card->wandev.if_head == NULL || card->wandev.if_head->ifp == NULL){
		DEBUG_EVENT("%s: Internal Error (line no. %d)!\n", 
						card->devname, __LINE__);
		return;
	} 
	ifp = card->wandev.if_head->ifp;
	chdlc_priv_area = ifp->if_softc;
	/* process a udp call if pending */
#ifndef CHDLC_OLD_INTR_TEST
	if (chdlc_priv_area->timer_int_enabled & TMR_INT_ENABLED_TEST){
		Intr_test_counter++;
    		if (Intr_test_counter >= MAX_NEW_INTR_TEST_COUNTER){
    			DEBUG_EVENT("%s: Interrupt test passed(%d)\n",
					card->devname, Intr_test_counter);
        		chdlc_priv_area->timer_int_enabled &= ~TMR_INT_ENABLED_TEST;
			chdlc_priv_area->timer_int_enabled |= TMR_INT_ENABLED_CONFIG;
		}
	}
#endif
 
	/* process a udp call if pending */
	if (chdlc_priv_area->timer_int_enabled & TMR_INT_ENABLED_UDP){
		process_udp_mgmt_pkt(card, ifp, chdlc_priv_area);
        	chdlc_priv_area->timer_int_enabled &= ~TMR_INT_ENABLED_UDP;
	}
 
	/* TE1 T1/E1 timer interrupt */
	if (chdlc_priv_area->timer_int_enabled & TMR_INT_ENABLED_TE){
		sdla_te_polling(card);
        	chdlc_priv_area->timer_int_enabled &= ~TMR_INT_ENABLED_TE;
	}

	/* Chdlc configure */
	if (chdlc_priv_area->timer_int_enabled & TMR_INT_ENABLED_CONFIG){
		chdlc_config(ifp);
        	chdlc_priv_area->timer_int_enabled &= ~TMR_INT_ENABLED_CONFIG;
	}

	/* read communicationstatistics if required */
	if (chdlc_priv_area->timer_int_enabled & TMR_INT_ENABLED_UPDATE){
		update_comms_stats(card, ifp);
		if (!(--chdlc_priv_area->update_comms_stats)){
        		chdlc_priv_area->timer_int_enabled &= ~TMR_INT_ENABLED_UPDATE;
		}
	}
	/* Only disable the timer interrupt if there are no udp or statistics 
	 * updates pending
	 */
	if (!chdlc_priv_area->timer_int_enabled){
		card->hw_iface.disable_byte(card->hw, card->intr_perm_off, APP_INT_ON_TIMER);
	}
}



/*
****************************************************************************
**
**
**
*/
static int chdlc_debugging(sdla_t* card)
{
	CHDLC_OPERATIONAL_STATS_STRUCT	op_stats;
	static unsigned long	rx_keepalives = 0;

	chdlc_read_op_stats(card, &op_stats);
	if (card->wan_debugging_state == WAN_DEBUGGING_CONT){
		rx_keepalives	= op_stats.CHDLC_SLARP_REPLY_Rx_count;
		card->wan_debugging_state = WAN_DEBUGGING_PROTOCOL;
		/*
		** Check number of keepalives frames after 15 seconds.
		*/
		return 15;
	}

	if (op_stats.CHDLC_SLARP_REPLY_Rx_count != rx_keepalives){
		if (card->wan_debug_last_msg != WAN_DEBUG_CHDLC_UNKNWN_MSG){
			DEBUG_EVENT("%s: Unknown CHDLC problem!\n",
						card->devname);
			DEBUG_EVENT("%s:    Contact your dealer.\n",
						card->devname);
		}
		card->wan_debug_last_msg = WAN_DEBUG_CHDLC_UNKNWN_MSG;
	}else{
		if (card->wan_debug_last_msg != WAN_DEBUG_CHDLC_KPLV_MSG){
			DEBUG_EVENT("%s: No replies to Keepalive packets!\n",
						card->devname);
			DEBUG_EVENT("%s:    Check remote router or ISP.\n",
						card->devname);
		}
		card->wan_debug_last_msg = WAN_DEBUG_CHDLC_KPLV_MSG;
	}
	return 0;
}

static unsigned long chdlc_crc_frames(sdla_t* card)
{
	COMMS_ERROR_STATS_STRUCT	stats;

 	if (!chdlc_read_comm_err_stats(card, &stats)){
		return stats.CRC_err_count;	
	}
	return 0;
}

static unsigned long chdlc_abort_frames(sdla_t * card)
{
	COMMS_ERROR_STATS_STRUCT	stats;

 	if (!chdlc_read_comm_err_stats(card, &stats)){
		return stats.Rx_abort_count;	
	}
	return 0;
}

static unsigned long chdlc_tx_underun_frames(sdla_t* card)
{
	COMMS_ERROR_STATS_STRUCT	stats;

 	if (!chdlc_read_comm_err_stats(card, &stats)){
		return stats.sec_Tx_abort_msd_Tx_int_count;	
	}

	return 0;
}

static void chdlc_handle_front_end_state(void* card_id)
{
	sdla_t*	card = (sdla_t*)card_id;
	
	WAN_ASSERT1(card == NULL);
	/* FIXME: Add handle front end state changes */

	return;
}

#ifdef WANDEBUG_CHDLC
static void 
chdlc_dump_mailbox(wan_mbox_t *mbox)
{
    DEBUG_EVENT("Opp Flag      = 0x%x\n", mbox->wan_opp_flag);
    DEBUG_EVENT("Command       = 0x%x\n", mbox->wan_command);
    DEBUG_EVENT("Buffer Length = 0x%x\n", mbox->wan_data_len);
    DEBUG_EVENT("Return code   = 0x%x\n", mbox->wan_return_code);
}
static void
chdlc_dump_udp(wan_udp_pkt_t* udp_pkt, unsigned int udp_port)
{
    if (udp_pkt->wan_udp_signature){
        DEBUG_EVENT("Signature     = %s(%s)\n", 
		udp_pkt->wan_udp_signature, CHDLC_UDP_SIGNATURE);
    }
    DEBUG_EVENT("UDP port      = %d(%d)\n", 
		udp_pkt->wan_udp_dport, ntohs(udp_port));
    DEBUG_EVENT("Protocol      = %d(%d)\n", 
		udp_pkt->wan_ip_p, UDPMGMT_UDP_PROTOCOL);
    DEBUG_EVENT("Request reply = %d(%d)\n", 
		udp_pkt->wan_udp_request_reply, UDPMGMT_REQUEST);
}
#endif


/****** End ****************************************************************/
