/*
 * Copyright (c) 1997, 1998, 1999
 *	Alex Feldman <al.feldman@sangoma.com>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Alex Feldman.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Alex Feldman AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Alex Feldman OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: wanmain.c,v 1.23 2003/09/24 20:25:27 sangoma Exp $
 */

/*****************************************************************************
 * wanmain.c	WAN Multiprotocol Router Module. Main code.
 *
 *		This module is completely hardware-independent and provides
 *		the following common services for the WAN Link Drivers:
 *		 o WAN device managenment (registering, unregistering)
 *		 o Network interface management
 *		 o Physical connection management (dial-up, incomming calls)
 *		 o Logical connection management (switched virtual circuits)
 *
 * Author:	Alex Feldman	<al.feldmang@sangoma.com>
 *
 * ============================================================================
 * Dec 12, 1999	Alex Feldman	Updated for FreeBSD 3.2 
 * Dec 27, 1996	Gene Kozin	Initial version (based on Sangoma's WANPIPE)
 ******************************************************************************
 */

/*
 ****************************************************************************** 
 *				INCLUDE FILES 				      *
 ******************************************************************************
 */ 
#include <net/wanpipe_config.h>
#include <net/wanpipe_version.h>
#include <net/wanpipe_includes.h>
#include <net/wanpipe_defines.h>
#include <net/wanpipe_debug.h>
#include <net/wanpipe_common.h>
#include <net/wanproc.h>
#include <net/wanpipe.h>

/*
 ****************************************************************************** 
 *				DEFINES and MACROS			      *
 ******************************************************************************
 */ 

#ifndef	min
#define min(a,b) (((a)<(b))?(a):(b))
#endif
#ifndef	max
#define max(a,b) (((a)>(b))?(a):(b))
#endif

#if defined(__FreeBSD__)
# define IFP2AC(ifp)	((struct arpcom*)ifp)
# if (__FreeBSD_version < 410000)
#  define ETHER_IFATTACH(ifp)	if_attach(ifp); ether_ifattach(ifp)
#  define ETHER_IFDETACH(ifp)	if_detach(ifp)
# elif (__FreeBSD_version > 500000)
#  define ETHER_IFATTACH(ifp)	ether_ifattach(ifp, IFP2AC(ifp)->ac_enaddr)
#  define ETHER_IFDETACH(ifp)	ether_ifdetach(ifp)
# else
#  define ETHER_IFATTACH(ifp)	ether_ifattach(ifp, ETHER_BPF_SUPPORTED)
#  define ETHER_IFDETACH(ifp)	ether_ifdetach(ifp, ETHER_BPF_SUPPORTED)
# endif
# define IFATTACH(ifp)		if_attach(ifp)
# define IFDETACH(ifp)		if_detach(ifp)
# if (__FreeBSD_version > 500000)
typedef struct thread*	wan_proc_ptr_t;
# else
typedef struct proc*	wan_proc_ptr_t;
# endif
#elif defined(__OpenBSD__)
#  define ETHER_IFATTACH(ifp)	if_attach(ifp); ether_ifattach(ifp)
#  define ETHER_IFDETACH(ifp)	ether_ifdetach(ifp); if_detach(ifp)
# if defined(OpenBSD3_2) || defined(OpenBSD3_3)
#  define IFATTACH(ifp)		if_attach(ifp);if_alloc_sadl(ifp)
#  define IFDETACH(ifp)		if_free_sadl(ifp); if_detach(ifp)
# else
#  define IFATTACH(ifp)		if_attach(ifp)
#  define IFDETACH(ifp)		if_detach(ifp)
# endif
typedef struct proc*	wan_proc_ptr_t;
#else
# error "IFATTACH/IFDETACH/ETHER_IFDETACH macros doesn't supported yet!"
#endif

#ifdef NBPFILTER
# if defined(__FreeBSD__)
#  define BPF_ATTACH(ifp,type)	bpfattach(ifp, type, 4)
# elif defined(__OpenBSD__)
#  define BPF_ATTACH(ifp,type)	bpfattach(&ifp->if_bpf, ifp, DLT_NULL, 4)
# else
#  error "BPF_ATTACH macro doesn't supported yet!"
# endif
#else
# define BPF_ATTACH(ifp, type)
#endif

/*
 ****************************************************************************** 
 *			TYPEDEF and STRUCTURES				      *
 ******************************************************************************
 */ 
#if defined(__FreeBSD__)
# define CDEV_MAJOR 139
d_open_t wanrouter_open;
d_close_t wanrouter_close;
d_ioctl_t wanrouter_ioctl;

# if (__FreeBSD_version >= 501000)
static struct cdevsw wanrouter_devsw = 
{
	CDEV_MAJOR,
	D_NOGIANT,	/* FIXME: ???? */
	"wanrouter",
	wanrouter_open,
	wanrouter_close,
	noread,
	nowrite,
	wanrouter_ioctl,
	NULL,
	nommap,
	nostrategy,
	NULL,
	nopsize,
};
# elif (__FreeBSD_version > 400000)
static struct cdevsw wanrouter_devsw = 
{
	wanrouter_open,
	wanrouter_close,
	noread,
	nowrite,
	wanrouter_ioctl,
	NULL,
	nommap,
	nostrategy,
	"wanrouter",
	CDEV_MAJOR,
	NULL,
	nopsize,
	0,
	0
};
# else
static struct cdevsw wanrouter_devsw = 
{
	wanrouter_open,
	wanrouter_close,
	noread,
	nowrite,
	wanrouter_ioctl,
	nostop,
	nullreset,
	nodevtotty,
	NULL,
	nommap,
	NULL,
	"wanrouter",
	NULL,
	CDEV_MAJOR
};
# endif
#else

int wanrouter_open __P((dev_t dev, int oflags, int devtype, struct proc* p));
int wanrouter_close __P((dev_t dev, int oflags, int devtype, struct proc* p));
int wanrouter_ioctl __P((dev_t dev, u_long cmd, caddr_t data, int fflag, struct proc* p));

static struct cdevsw wanrouter_devsw = {
	wanrouter_open,
	wanrouter_close,
	NULL,
	NULL,
	wanrouter_ioctl,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};

#endif

/* 
 *	WAN device IOCTL handlers 
 */

static int device_setup(wan_device_t *wandev, wandev_conf_t *u_conf);
static int device_stat(wan_device_t *wandev, wandev_stat_t *u_stat);
static int device_shutdown(wan_device_t *wandev, wandev_conf_t* u_conf);
static int device_debugging(wan_device_t *wandev, char* u_ifname);
static int device_new_if(wan_device_t *wandev, wanif_conf_t *u_conf);
static int device_del_if(wan_device_t *wandev, char *u_name);
static int device_procfs(wan_device_t* wandev, wan_procfs_t* procfs);

/* 
 *	Miscellaneous 
 */

static wan_device_t *find_device (char *name); 
static void wanpipe_attach(struct ifnet *ifp, wan_device_t* wandev);
static void wanpipe_detach(struct ifnet *ifp, wan_device_t* wandev);
static int delete_interface (wan_device_t *wandev, char *name, int force);

/*
 *	Global Data
 */
wan_device_t* router_devlist 	= NULL;	/* list of registered devices */
static int devcnt 		= 0;

/********************* Kernel Loadable Module Entry Points *******************/
#ifdef __FreeBSD__
static int wanrouter_load __P((module_t, int, void*));
#else
MOD_DEV("wanrouter", LM_DT_CHAR, -1, &wanrouter_devsw);
extern int sdladrv_handle(int cmd);
extern int wanpipe_handle(int cmd);
#endif

/* 
 *=============================================================================
 * 	Module entry point.
 *
 * 	Return:	0	Ok
 *		< 0	error.
 * 	Context:	process
 */
#ifdef __FreeBSD__
int wanrouter_load (module_t mod, int cmd, void *arg)
#else
int wanrouter_handle(struct lkm_table* lkmtp, int cmd)
#endif
{
	int err = 0;
#ifdef __FreeBSD__
# if (__FreeBSD_version > 400000)
	static dev_t wanpipe_dev;
# endif 
#endif

#ifdef __OpenBSD__
	sdladrv_handle(cmd);
	wanpipe_handle(cmd);
#endif
	switch(cmd){
	case WAN_MOD_LOAD:
        	/* 'insert' entry point.
	         * 	o print announcement
        	 * 	o initialize static data
	         */
		DEBUG_MOD("Loading WANPIPE module...\n");
#ifdef __FreeBSD__
# if (__FreeBSD_version > 400000)
		wanpipe_dev = make_dev(&wanrouter_devsw, 
				       0, 
				       UID_ROOT, 
			       	       GID_WHEEL, 
				       0600, 
				       "wanrouter");
# endif 
#endif
        	break;
	
	case WAN_MOD_UNLOAD:
	        /*
        	 * 'remove' entry point.
	         */
#ifdef __FreeBSD__
# if (__FreeBSD_version > 400000)
		destroy_dev(wanpipe_dev);
# endif 
#endif
		break;
	}
	return err;
}
#ifdef __FreeBSD__
# if (__FreeBSD_version > 400000)
DEV_MODULE(wanrouter, wanrouter_load, NULL);
MODULE_VERSION(wanrouter,WANROUTER_MAJOR_VER);
MODULE_DEPEND(wanrouter,sdladrv,1,SDLADRV_MAJOR_VER,SDLADRV_MAJOR_VER);
# else
CDEV_MODULE(wanrouter, CDEV_MAJOR, wanrouter_devsw, wanrouter_load, NULL);
# endif
#else
int wanrouter(lkmtp, cmd, ver)
	struct lkm_table*	lkmtp;
	int 			cmd;
	int 			ver;
{
    DISPATCH(lkmtp, cmd, ver, wanrouter_handle, wanrouter_handle, lkm_nofunc)
}
#endif

/*
 * 	Kernel APIs
 */

/* 
 * ============================================================================
 * Register WAN device.
 * o verify device credentials
 * o create an entry for the device in the /proc/net/router directory(LINUX)
 * o initialize internally maintained fields of the wan_device structure
 * o link device data space to a singly-linked list
 * o if it's the first device, then start kernel 'thread'
 * o increment module use count
 *
 * Return:
 * 0	Ok
 * < 0	error.
 *
 * Context:	process
 */
int register_wan_device(wan_device_t *wandev)
{
	int namelen;

	WAN_ASSERT(wandev == NULL);
	WAN_ASSERT(wandev->magic != ROUTER_MAGIC);
	WAN_ASSERT(wandev->name == NULL);
 
	namelen = strlen(wandev->name);
	if (!namelen || (namelen > WAN_DRVNAME_SZ)){
		return -EINVAL;
	}
	
	if (find_device(wandev->name) != NULL){
		return -EEXIST;
	}

	DEBUG_INIT("%s: registering WAN device...\n", 
					wandev->name);

	/*
	 * Initialize fields of the wan_device structure 
	 * maintained by the router and update local data.
     	 */
 
	wandev->ndev    = 0;
	wandev->if_head = NULL;
	wandev->next    = router_devlist;
	router_devlist  = wandev;
	++devcnt;
	return 0;
}

/* 
 * ============================================================================
 *	Unregister WAN device.
 *	o shut down device
 *	o unlink device data space from the linked list
 *	o delete device entry in the /proc/net/router directory
 *	o decrement module use count
 *
 *	Return:		0	Ok
 *			<0	error.
 *	Context:	process
 */

int unregister_wan_device(char *name)
{
	wan_device_t *wandev = NULL, *prev = NULL;

	WAN_ASSERT(name == NULL);	

	for (wandev = router_devlist, prev = NULL;
		wandev && strcmp(wandev->name, name);
		prev = wandev, wandev = wandev->next);
	if (wandev == NULL){
		return -ENODEV;
	}

	DEBUG_INIT("%s: unregistering WAN device\n", 
					wandev->name);

	if (wandev->state != WAN_UNCONFIGURED){ 
		struct ifnet *ifp = NULL;
	        while(wandev->if_head){
			if (wandev->if_head && 
			    wandev->if_head->ifp){
				ifp = wandev->if_head->ifp;   
				delete_interface(wandev, wan_netif_name(ifp), 1);
			}else{ /* Vary bad case */
				struct ifnet_list *tmp = wandev->if_head;
				wandev->if_head = wandev->if_head->pnext;
				if (tmp){
					wan_free(tmp);
				}
			}
		}
		if (wandev->shutdown){	
			wandev->shutdown(wandev, NULL);
		}
	}
    	if (prev){
	        prev->next = wandev->next;
	}else{
		router_devlist = wandev->next;
	}
	--devcnt;
	return 0;
}

int wanrouter_open(dev_t dev, int flag, int otyp, wan_proc_ptr_t procp)
{
	return (0);
}
int wanrouter_close(dev_t dev, int flag, int otyp, wan_proc_ptr_t procp)
{   
	return (0);
}
/*
 * ============================================================================
 * 	WAN device IOCTL.
 *	o find WAN device associated with this node
 *	o execute requested action or pass command to the device driver
 */
int wanrouter_ioctl( dev_t dev, u_long cmd, caddr_t arg, int mode, wan_proc_ptr_t procp)
{
	int err = 0;
	wan_device_t *wandev = NULL;
	wan_conf_t *cfg = NULL;

	WAN_ASSERT(((cmd >> 8) & ROUTER_IOCTL) != ROUTER_IOCTL);
	WAN_ASSERT(arg == NULL);
    
	/* The arg is points already to kernel space */
	cfg = (wan_conf_t*)arg;

	wandev = find_device(cfg->devname);
	if (!((wandev != NULL && wandev->magic == ROUTER_MAGIC) || 
	     (wandev == NULL && cmd == ROUTER_PROCFS))){
        	return -EINVAL;
	}
		
	switch(cmd){ 
    	case ROUTER_SETUP:
		DEBUG_CFG("%s: ROUTER_SETUP \n", cfg->devname);
		err = device_setup(wandev, cfg->arg);
		break;

	case ROUTER_DOWN:
		DEBUG_CFG("%s: ROUTER_DOWN\n", cfg->devname);
		err = device_shutdown(wandev, cfg->arg);
		break;

	case ROUTER_DEBUGGING:
		DEBUG_CFG("%s: ROUTER_DEBUGGING\n", cfg->devname);
		err = device_debugging(wandev, cfg->arg);
		break;

	case ROUTER_STAT:
		DEBUG_CFG("%s: ROUTER_STAT\n", cfg->devname); 
		err = device_stat(wandev, cfg->arg);
		break;

	case ROUTER_IFNEW:
		DEBUG_CFG("%s: ROUTER_IFNEW\n", cfg->devname);
		err = device_new_if(wandev, cfg->arg);
		break;

	case ROUTER_IFDEL:
		DEBUG_CFG("%s: ROUTER_IFDEL\n", cfg->devname);
		err = device_del_if(wandev, cfg->devname);
		break;

	case ROUTER_VER:
		DEBUG_CFG("%s: ROUTER_VER\n", cfg->devname);
		if (copyout((void*)WANPIPE_VERSION, 
			    (void*)cfg->arg, 
			    sizeof(WANPIPE_VERSION))){
			return -EFAULT;
		}
		break;

	case ROUTER_IFSTAT:
        	break;

	case ROUTER_PROCFS:
        	DEBUG_CFG("wanpipe?: ROUTER_PROCFS\n");
		err = device_procfs(wandev, cfg->arg);
		break;

	case ROUTER_DEBUG_READ:
        	DEBUG_CFG("wanpipe?: ROUTER_DEBUG_READ\n");
		if (wandev->debug_read){
			err = wandev->debug_read(wandev, cfg->arg);
		}else{
			err = -EINVAL;
		}
		break;

	default:
		DEBUG_CFG("%s: default\n", cfg->devname);
		if (wandev->ioctl){
			err = wandev->ioctl(wandev, cmd, cfg->arg);
		}else{
			err = -EINVAL;
		}
	}
	return err;
}

/*
 *	WAN Driver IOCTL Handlers
 */
/* 
 * ============================================================================
 *	Setup WAN link device.
 *	o verify user address space
 *	o allocate kernel memory and copy configuration data to kernel space
 *	o if configuration data includes extension, copy it to kernel space too
 *	o call driver's setup() entry point
 */

static int device_setup (wan_device_t *wandev, wandev_conf_t *u_conf)
{
	void *data = NULL;
	wandev_conf_t *conf = NULL;
	int err = 0;

	DEBUG_CFG("%s: Device Setup...\n", wandev->name);
	if (wandev->setup == NULL){	/* Nothing to do ? */
		return 0;
	}

	conf = wan_malloc(sizeof(wandev_conf_t));
	if (conf == NULL){
		DEBUG_EVENT("%s: Cannot allocate memory for configuration structure!\n",
				wandev->name);
		return -ENOMEM;
	}

	if (copyin((void *)u_conf, (void*)conf, sizeof(wandev_conf_t))){
		wan_free(conf);
		return -EFAULT;
	}
	
	if (conf->magic != ROUTER_MAGIC){ 
        	DEBUG_EVENT("%s: Wrong configuration structure on Setup!\n", 
								wandev->name);	
		wan_free(conf);
		return -EINVAL; 
	}

	if (conf->data_size && conf->data){ 
		if (conf->data_size > 128000 || conf->data_size < 0){ 
			DEBUG_EVENT("%s: Large configuration structure\n", 
								wandev->name);
			wan_free(conf);
			return -EINVAL;
		}
		data = wan_malloc(conf->data_size);
		if (data){ 
			if (!copyin((void*)conf->data, data, conf->data_size)){
				conf->data = data;
				err = wandev->setup(wandev, conf);
			}else{
				err = -EFAULT;
			}
		}else{
			DEBUG_EVENT("%s: Cannot allocate data buffer\n", 
								wandev->name);
			err = -ENOMEM;
		}
		if (data){
			wan_free(data);
		}
	}else{
		if (conf->card_type == WANOPT_ADSL || conf->config_id == WANCONFIG_DEBUG){
			err = wandev->setup(wandev, conf);
		}else{
			DEBUG_EVENT("%s: ERROR: No firmware found (size = %i)!\n",
					wandev->name, conf->data_size);
		}
	}

	if (conf){
		wan_free(conf);
	}
	DEBUG_CFG("%s: Device setup finished (%d)\n", wandev->name, err);
	return err;
}

/* 
 * ============================================================================
 *	Shutdown WAN device.
 *	o delete all not opened logical channels for this device
 *	o call driver's shutdown() entry point
 */
 
static int device_shutdown (wan_device_t *wandev, wandev_conf_t* u_conf)
{
 	struct ifnet_list*	ifnode = NULL, *prev = NULL;
	struct ifnet*		ifp = NULL;
	wandev_conf_t*		conf = NULL;
	int 			error = 0;

	DEBUG_CFG("%s: device_shutdown called\n", wandev->name);	
	if (wandev->state == WAN_UNCONFIGURED){
		return 0;
	}

	if (u_conf){
		conf = wan_malloc(sizeof(wandev_conf_t));
		if (conf == NULL){
			DEBUG_EVENT("%s: Cannot allocate memory for configuration structure\n",
					wandev->name);
			return -ENOMEM;
		}

		if (copyin((void *)u_conf, (void*)conf, sizeof(wandev_conf_t))){
			wan_free(conf);
			return -EFAULT;
		}
		
		if (conf->magic != ROUTER_MAGIC){ 
        		DEBUG_EVENT("%s: Wrong configuration structure on Shutdown!\n", 
								wandev->name);	
			wan_free(conf);
			return -EINVAL; 
		}
	}

	/* Delete all interfaces */
	ifnode = wandev->if_head;
	prev = NULL;
	while(ifnode){
		ifp = ifnode->ifp;
		if (ifp){
			if (delete_interface(wandev, wan_netif_name(ifp), 0)){
#ifdef ENABLE_IPV6
				return -EINVAL;
#else
				prev = ifnode;
#endif
			}
		}else{ /* Bad case or error */
			struct ifnet_list *tmp = ifnode;
			if (prev){
				prev->pnext = ifnode->pnext;
			}else{	
				wandev->if_head = ifnode->pnext;
			}
			if (tmp){
				wan_free(tmp);
			}
		}
		ifnode = (prev) ? prev->pnext : wandev->if_head;
	}
	if (wandev->ndev){
		return -EBUSY;	/* there are opened interfaces  */
	}
		
	if (wandev->shutdown){
		error = wandev->shutdown(wandev, conf);
	}

	if (u_conf){

		if (copyout((void*)conf, (void*)u_conf, sizeof(wandev_conf_t))){
			return -EFAULT;
		}
		if (conf){
			wan_free(conf);
		}
	}

	DEBUG_CFG("%s: device_shutdown end\n", wandev->name);	
	return error;
}

/* 
 * ============================================================================
 *	Debugging WAN device.
 */
 
static int device_debugging (wan_device_t *wandev, char* u_ifname)
{
	int 			error = 0;

	DEBUG_CFG("%s: device_debugging called\n", wandev->name);	
	if (wandev->state == WAN_UNCONFIGURED){
		return 0;
	}

	if (wandev->debugging){
		error = wandev->debugging(wandev);
	}

	DEBUG_CFG("%s: device_debugging end\n", wandev->name);	
	return 0;
}

/* 
 * ============================================================================
 *	Get WAN device status & statistics.
 */

static int device_stat (wan_device_t *wandev, wandev_stat_t *u_stat)
{
	wandev_stat_t stat;

	bzero((void*)&stat, sizeof(stat));

	/* Ask device driver to update device statistics */
	if ((wandev->state != WAN_UNCONFIGURED) && wandev->update){
		wandev->update(wandev);
	}

	/* Fill out structure */
	stat.ndev  = wandev->ndev;
	stat.state = wandev->state;

	if (copyout((void*)&stat, (void*)u_stat, sizeof(stat))){
		return -EFAULT;
	}

	return 0;
}

/* 
 * ============================================================================
 *	Create new WAN interface.
 *	o verify user address space
 *	o copy configuration data to kernel address space
 *	o allocate network interface data space
 *	o call driver's new_if() entry point
 *	o make sure there is no interface name conflict
 *	o register network interface
 */

static int device_new_if (wan_device_t *wandev, wanif_conf_t *u_conf)
{
	wanif_conf_t conf;
    	struct ifnet *ifp = NULL;
    	char ifname[IFNAMSIZ+1];
    	int interface_exist = 0;
    	int err;
	sdla_t*	card = (sdla_t*)wandev->private;

    	DEBUG_EVENT("%s: Creating new WAN interface %s...\n", 
				wandev->name, u_conf->name);
    	if ((wandev->state == WAN_UNCONFIGURED) || (wandev->new_if == NULL)){
		return -ENODEV;
	}

	if (copyin((void*)u_conf, (void*)&conf, sizeof(wanif_conf_t))){
		return -EFAULT;
	}
		
	if (conf.magic != ROUTER_MAGIC){
		return -EINVAL;
	}

	/* The network interface contanins of the interface in two forms:
	 * character string plus a unit number for the driver and binary
     	 * system wide index number. The user define the network interface 
	 * name, he responsible that these names must be different. The
	 * unit number will always 0 (I cannot do nothing with this.
	 */
	strcpy(ifname, conf.name);
	ifp = ifunit(ifname);
	if (ifp){
		interface_exist = 1;    
	}else{
		if (wandev->config_id == WANCONFIG_ADSL){
			if(card->u.adsl.EncapMode == RFC_MODE_PPP_VC ||
			   card->u.adsl.EncapMode == RFC_MODE_PPP_LLC){
				ifp = (struct ifnet*)wan_malloc(sizeof(struct sppp));
			}else{
				ifp = (struct ifnet*)wan_malloc(sizeof(struct arpcom));
			}
		}else{
        		ifp = (struct ifnet*)wan_malloc(sizeof(struct sppp));
#if 0
HDLC_SPPP
        		ifp = wan_malloc(sizeof(struct ifnet));
#endif
		}
		if (ifp == NULL){
			DEBUG_EVENT("%s: Failed allocate memory for interface %s!\n", 
						wandev->name, ifname);
			return -ENOMEM;
		}
		if (wandev->config_id == WANCONFIG_ADSL){
			if(card->u.adsl.EncapMode == RFC_MODE_PPP_VC ||
			   card->u.adsl.EncapMode == RFC_MODE_PPP_LLC){
				bzero((struct sppp*)ifp, sizeof(struct sppp));
			}else{
				bzero((struct arpcom*)ifp, sizeof(struct arpcom));
			}
		}else{
			bzero((struct sppp*)ifp, sizeof(struct sppp));
#if 0
HDLC_SPPP
			bzero(ifp, sizeof(struct ifnet));
#endif
		}
		interface_exist = 0;
	}
 
	err = wandev->new_if(wandev, ifp, &conf);
	if (!err){ 
		struct ifnet_list *if_node = NULL;
		if_node = wan_malloc(sizeof(struct ifnet_list)); 
		if (if_node){
			int s = splimp();
			if_node->ifp = ifp;	
			if_node->pnext = wandev->if_head;
			wandev->if_head = if_node;
			++wandev->ndev;
			splx(s);
			DEBUG_EVENT("%s: Attaching network interface %s...\n", 
					wandev->name, wan_netif_name(ifp));
			if (!interface_exist){
				wanpipe_attach(ifp, wandev);
	    		}
			return 0;	/* done !!! */
		}else{
	   	        DEBUG_EVENT("%s: Failed allocate memory for network interface list!\n", 
					wandev->name);
			err = -ENOMEM;
		}
	}
	wan_free(ifp);
	DEBUG_EVENT("%s: Failed to create new interface (%d)!\n",
						wandev->name, err);
	return err;
}


/* 
 * ============================================================================
 *	Delete WAN logical channel.
 *	 o verify user address space
 *	 o copy configuration data to kernel address space
 */
static int device_del_if (wan_device_t *wandev, char *u_name)
{
	char name[WAN_IFNAME_SZ + 1];

	if (wandev->state == WAN_UNCONFIGURED){
		return -ENODEV;
	}
	
	bzero((void*)name, sizeof(name));

	if (copyin((void*)u_name, (void*)name, WAN_IFNAME_SZ)){
		return -EFAULT;
	}

	return delete_interface(wandev, name, 0);
}


/*
 *	Miscellaneous Functions
 */

/* 
 * ===========================================================================
 *	Find WAN device by name.
 *	Return pointer to the WAN device data space or NULL if device not found.
 */

static wan_device_t *find_device(char *name)
{
	wan_device_t *wandev = NULL;

	for (wandev = router_devlist;wandev && strcmp(wandev->name, name);
		wandev = wandev->next);
	return wandev;
}

/* 
 * ============================================================================
 *	Delete WAN logical channel identified by its name.
 *	o find logical channel by its name
 *	o call driver's del_if() entry point
 *	o unregister network interface
 *	o unlink channel data space from linked list of channels
 *	o release channel data space
 *
 *	Return:	0		success
 *		-ENODEV		channel not found.
 *		-EBUSY		interface is open
 *
 *	Note: If (force != 0), then device will be destroyed even if interface
 *	associated with it is open. It's caller's responsibility to make
 *	sure that opened interfaces are not removed!
 */

static int delete_interface (wan_device_t *wandev, char *name, int force)
{
	struct ifnet*		ifp = NULL;
	struct ifnet_list*	if_node = NULL,
			      	*prev    = NULL;
	int s = 0;

	DEBUG_CFG("%s: Deleting interface %s...\n", wandev->name, name);
	/* Searching for all interfaces */
	for(if_node = wandev->if_head, prev = NULL; if_node && if_node->ifp; 
		prev = if_node, if_node = if_node->pnext){
		if ((strcmp(name, wan_netif_name(if_node->ifp)) == 0)){
			break;
		}
	}
	if (if_node == NULL){
		DEBUG_CFG("%s: Network interface %s is not found.\n", 
						wandev->name, name);
		return -ENODEV;
	} 
	ifp = if_node->ifp;
	if ((ifp->if_flags & IFF_UP) == IFF_UP){ 
		if (force){ 
			DEBUG_CFG("%s: deleting opened interface %s!\n", 
						wandev->name, name);
		}else{
			return -EBUSY;	/* interface in use */
		}
		ifp->if_flags &= ~IFF_UP;
	}

	wanpipe_detach(ifp, wandev);
#ifndef ENABLE_IPV6
	if (wandev->del_if){
		wandev->del_if(wandev, ifp);
	}
	if_node->ifp = NULL;
#endif
	s = splimp(); 	/***** critical section start *****/
    
	if (prev){
		prev->pnext = if_node->pnext;
	}else{
		wandev->if_head = if_node->pnext;
	}
	--wandev->ndev;

	splx(s); 		/****** critical section end ******/

#ifdef ENABLE_IPV6
	if (wandev->del_if){
		wandev->del_if(wandev, ifp);
	}
	if_node->ifp = NULL;
#endif
    
#if (defined __FreeBSD__) && (__FreeBSD_version < 400000)
/*
 * Do not enable IF_DETACH for FreeBSD-3.x (kernel bug). 
 */
# ifdef IF_DETACH
	wan_free(ifp);
# endif
#else
	wan_free(ifp);
#endif
	wan_free(if_node);
	DEBUG_CFG("%s: Interface %s deleted.\n", 
					wandev->name, name);

	return 0;
}

static void wanpipe_attach(struct ifnet *ifp, wan_device_t* wandev)
{
    	IFQ_SET_READY(&ifp->if_snd);
	if (wandev->config_id == WANCONFIG_ADSL){
		sdla_t*	card = (sdla_t*)wandev->private;
		if(card->u.adsl.EncapMode == RFC_MODE_PPP_VC ||
		   card->u.adsl.EncapMode == RFC_MODE_PPP_LLC){
    			IFATTACH(ifp);
    			BPF_ATTACH(ifp, DLT_PPP);
		}else{
			ETHER_IFATTACH(ifp);
		}
	}else{
    		IFATTACH(ifp);
    		BPF_ATTACH(ifp, (wandev->config_id == WANCONFIG_MPPP) ? DLT_PPP : DLT_RAW);
	}
}

static void wanpipe_detach(struct ifnet *ifp, wan_device_t* wandev)
{
#if defined(__FreeBSD__)
#if  (__FreeBSD_version < 400000)
    	struct ifaddr* ifa = NULL, *dst_ifa = NULL;
    	int s = splnet();

    	/* Remove addresses from routing tables */
    	ifa = TAILQ_FIRST(&ifp->if_addrhead);
    	for(; ifa; ifa = TAILQ_NEXT(ifa, ifa_link)){
		rtinit(ifa, (int)RTM_DELETE, 0);
	}
    	if ((ifa = TAILQ_FIRST(&ifp->if_addrhead)) != NULL){
		if ((dst_ifa = TAILQ_NEXT(ifa, ifa_link)) != NULL){
	    		TAILQ_REMOVE(&ifp->if_addrhead, dst_ifa, ifa_link);
		}
	}
/*
 * Do not enable IF_DETACH for FreeBSD-3.x (kernel bug). 
 */
#  ifdef IF_DETACH
    	TAILQ_REMOVE(&ifnet, ifp, if_link);
#  endif 
    	splx(s);
# endif /* __FreeBSD_version */
#endif  /* __FreeBSD__ */

	if (wandev->config_id == WANCONFIG_ADSL){
		sdla_t*	card = (sdla_t*)wandev->private;
		if(card->u.adsl.EncapMode == RFC_MODE_PPP_VC ||
		   card->u.adsl.EncapMode == RFC_MODE_PPP_LLC){
    			IFDETACH(ifp);
		}else{
			ETHER_IFDETACH(ifp);
		}
	}else{
    		IFDETACH(ifp);
	}
}

static int device_procfs(wan_device_t* wandev, wan_procfs_t* u_procfs)
{
	wan_procfs_t	procfs;
	char*		data = NULL;
	int		cnt = 0;

	if (copyin((void*)u_procfs, (void*)&procfs, sizeof(wan_procfs_t))){
		return -EFAULT;
	}
		
	if (procfs.magic != ROUTER_MAGIC){
		return -EINVAL;
	}
	if ((data = wan_malloc(procfs.max_len)) == NULL){
		return -ENOMEM;
	}
	switch(procfs.cmd){
	case WANPIPE_PROCFS_CONFIG:
		DEBUG_PROCFS("WANPIPE_PROCFS_CONFIG!\n");
		cnt = config_get_info(data, NULL, procfs.offs, procfs.max_len);
		break;	

	case WANPIPE_PROCFS_HWPROBE:
		DEBUG_PROCFS("WANPIPE_PROCFS_HWPROBE!\n");
		cnt = probe_get_info(data, NULL, procfs.offs, procfs.max_len);
		break;	

	case WANPIPE_PROCFS_STATUS:
		DEBUG_PROCFS("WANPIPE_PROCFS_STATUS!\n");
		cnt = status_get_info(data, NULL, procfs.offs, procfs.max_len);
		break;	

	case WANPIPE_PROCFS_INTERFACES:
		DEBUG_PROCFS("WANPIPE_PROCFS_INTERFACES!\n");
		cnt = interfaces_get_info(data, NULL, procfs.offs, procfs.max_len);
		break;	

	case WANPIPE_PROCFS_DEV:
#if 0
		DEBUG_PROCFS("WANPIPE_PROCFS_DEV!\n");
		if (wandev){
			cnt = wandev_get_info(data, wandev, procfs.offs, procfs.max_len);
		}
#endif
		break;	

	}
	procfs.is_more = (cnt) ? 1 : 0;
	procfs.offs = cnt;
	if (cnt){
		if (copyout((void*)data, (void*)procfs.data, cnt)){
			return -EFAULT;
		}
		if (data){
			wan_free(data);
		}
	}
	if (copyout((void*)&procfs, (void*)u_procfs, sizeof(wan_procfs_t))){
		return -EFAULT;
	}
	return 0;
}

/************************ End of Wanmain.c ******************************/
