/*****************************************************************************
* wanpipe_abstr.c WANPIPE(tm) Kernel Abstraction Layer.
*
* Authors: 	Nenad Corbic <ncorbic@sangoma.com>
*		Alex Feldman <al.feldman@sangoma.com
*
* Copyright:	(c) 2003 Sangoma Technologies Inc.
*
* ============================================================================
* Jan 20, 2003  Nenad Corbic	Initial version
* ============================================================================
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
*/


/*
**************************************************************************
**			I N C L U D E S  F I L E S			**
**************************************************************************
*/

#if defined(__FreeBSD__) || defined(__OpenBSD__)
# include <net/wanpipe_config.h>
# include <net/wanpipe_version.h>
# include <net/wanpipe_includes.h>
# include <net/wanpipe.h>
# include <net/wanpipe_abstr.h>
# include <net/sdla_adsl.h>
#elif defined(__KERNEL__)
# include <linux/wanpipe_version.h>
# include <linux/wanpipe_includes.h>
# include <linux/wanpipe.h>
# include <linux/if_wanpipe.h>
# include <linux/if_wanpipe_common.h>
# include <linux/wanpipe_abstr.h>
#else
# error "Unsupported Operating System!"
#endif


/*
******************************************************************************
**		Call Back function for kernel memory buffers
******************************************************************************
*/
/*
** wpabs_skb_data() - 
*/
unsigned char* wpabs_skb_data(void* skb)
{
	return wan_skb_data(skb);
}

/*
** wpabs_skb_len() - 
*/
int wpabs_skb_len(void* skb)
{
	return wan_skb_len(skb);
}

/*
** wpabs_skb_alloc() - 
*/
void* wpabs_skb_alloc(unsigned int len, void* dev)
{
	return wan_skb_alloc(len,dev);
}

/*
** wpabs_skb_free() - 
*/
void wpabs_skb_free(void* skb)
{
	wan_skb_free(skb);
}

/*
** wpabs_skb_copyback() - 
*/
void wpabs_skb_copyback(void* skb, int off, int len, unsigned long cp)
{
	wan_skb_copyback(skb, off, len, (caddr_t)cp);	
}

void wpabs_skb_copyback_user(void* skb, int off, int len, unsigned long cp)
{
	wan_skb_copyback_user(skb, off, len, (caddr_t)cp);	
}


/*
** wpabs_skb_pull() - 
*/
unsigned char* wpabs_skb_pull(void* skb, int len)
{
	return wan_skb_pull(skb, len); 
}

/*
** wpabs_skb_put() - 
*/
unsigned char* wpabs_skb_put(void* skb, int len)
{
	return wan_skb_put(skb, len); 
}


/*
** wpabs_skb_trim() - Trim from tail
*/
void wpabs_skb_trim(void* skb, unsigned int len)
{
	wan_skb_trim(skb, len);
}

/*
** wpabs_skb_tailroom() - Skb tail room
*/
int wpabs_skb_tailroom(void* skb)
{
	return wan_skb_tailroom(skb);
}

/*
** wpabs_skb_queue_len() - Length of skb queue
*/

int wpabs_skb_queue_len(void *queue)
{
	return wan_skb_queue_len(queue);
}

/*
** wpabs_skb_queue_tail() - Length of skb queue
*/

void wpabs_skb_queue_tail(void *queue,void *skb)
{
	wan_skb_queue_tail(queue,skb);
}



/*
** wpabs_skb_init() - Init data pointer
*/
void wpabs_skb_init(void* skb, unsigned int len)
{
	wan_skb_init(skb, len);
}


/*
**
*/
void wpabs_skb_set_dev(void *skb_new_ptr, void *dev)
{
#ifdef __LINUX__
	struct sk_buff *skb_new=(struct sk_buff *)skb_new_ptr;

	if (dev){
		skb_new->dev=dev;	
	}
	return;
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
	/* Do nothing for FreeBSD/OpenBSD */
#else
# error "Undefined function wpabs_skb_set_dev()"
#endif
}



/*
** wpabs_netif_queue_stopped
*/
int wpabs_netif_queue_stopped(void* dev)
{
	return WAN_NETIF_QUEUE_STOPPED((netdevice_t*)dev);
}

/*
** wpabs_netif_queue_stopped
*/
int wpabs_netif_dev_up(void* dev)
{
	return WAN_NETIF_UP((netdevice_t*)dev);
}



/*
** wpabs_netif_wake_queue
*/
void wpabs_netif_wake_queue(void* dev)
{
	WAN_NETIF_WAKE_QUEUE((netdevice_t*)dev);
}

void* wpabs_timer_alloc(void)
{
	return wan_malloc(sizeof(wan_timer_t));	
}

/*
** wpabs_add_timer() - Set timer
*/
void wpabs_add_timer(void* timer_info, unsigned long delay)
{
	wan_timer_t*	timer = (wan_timer_t*)timer_info;

	WAN_ASSERT1(timer == NULL);
	wan_add_timer(timer, delay);
}

/*
** wpabs_init_timer
*/
void wpabs_init_timer(void* timer_info, void* timer_func, unsigned long data)
{
	wan_timer_t*	timer = (wan_timer_t*)timer_info;
	
	WAN_ASSERT1(timer == NULL);
	timer->MiniportTimerFunction = (wan_timer_func_t)timer_func;
	timer->MiniportAdapterHandle = (void*)data;
	timer->MiniportTimerContext  = (void*)data;
	
	wan_init_timer(timer, timer_func, data);
}

/*
** wpabs_del_timer
*/
void wpabs_del_timer(void* timer_info)
{
	wan_timer_t*	timer = (wan_timer_t*)timer_info;

	WAN_ASSERT1(timer == NULL);
	wan_del_timer(timer);
}


/*void* wpabs_dma_alloc(void* pcard, DMA_DESCRIPTION* dma_descr)
*/
void* wpabs_dma_alloc(void* pcard, unsigned long max_length)
{
	sdla_t*			card = (sdla_t*)pcard;
	wan_dma_descr_t*	dma_descr = NULL;

	dma_descr = wan_malloc(sizeof(wan_dma_descr_t));
	if (dma_descr == NULL){
		return NULL;
	}
	dma_descr->max_length = max_length;
#ifdef __OpenBSD__
	card->hw_iface.getcfg(card->hw, SDLA_DMATAG, &dma_descr->dmat);
#endif
	if (wan_dma_alloc(card->hw, dma_descr)){
		return NULL;
	}
	return dma_descr;
}

int wpabs_dma_free(void* pcard, void* dma_descr)
{
	sdla_t*	card = (sdla_t*)pcard;
	return wan_dma_free(card->hw, (wan_dma_descr_t*)dma_descr);
}

unsigned long* wpabs_dma_get_vaddr(void* pcard, void* dma_descr)
{
	wan_dma_descr_t*	dma = (wan_dma_descr_t*)dma_descr;
	return dma->vAddr;
}

unsigned long wpabs_dma_get_paddr(void* pcard, void* dma_descr)
{
	wan_dma_descr_t*	dma = (wan_dma_descr_t*)dma_descr;
	return wan_virt2bus(dma->vAddr);
}
void* wpabs_malloc(int size)
{
	return wan_malloc(size);
}

void wpabs_free(void* ptr)
{
	wan_free(ptr);
}

/*
** wpabs_virt2bus
*/
unsigned long wpabs_virt2bus(unsigned long* ptr)
{
	return wan_virt2bus(ptr);
}

/*
** wpabs_bus2virt
*/
unsigned long* wpabs_bus2virt(unsigned long ptr)
{
	return wan_bus2virt(ptr);
}

/*
**
*/
unsigned char wpabs_bus_read_1(void* cardp, int offset)
{
	sdla_t*		card = (sdla_t*)cardp;
	unsigned char	data = 0x00;

	WAN_ASSERT(card == NULL);
	card->hw_iface.peek(card->hw, offset, (void*)&data, 1);
	return data;
}

/*
**
*/
unsigned long wpabs_bus_read_4(void* cardp, int offset)
{
	sdla_t*		card = (sdla_t*)cardp;
	unsigned long	data = 0x00;

	WAN_ASSERT(card == NULL);
	card->hw_iface.peek(card->hw, offset, (void*)&data, 4);
	return data;
}

/*
**
*/
void wpabs_bus_write_1(void* cardp, int offset, unsigned char data)
{
	sdla_t*	card = (sdla_t*)cardp;

	WAN_ASSERT1(card == NULL);
	card->hw_iface.poke_byte(card->hw, offset, data);
}

/*
**
*/
void wpabs_bus_write_4(void* cardp, int offset, unsigned long data)
{
	sdla_t*	card = (sdla_t*)cardp;

	WAN_ASSERT1(card == NULL);
	card->hw_iface.poke(card->hw, offset, (void*)&data, 4);
}

/*
**
*/
void wpabs_udelay(unsigned long microsecs)
{
	DELAY(microsecs);
}

void* wpabs_spinlock_alloc(void)
{
	return wan_malloc(sizeof(wan_spinlock_t));
}
void wpabs_spinlock_free(void* lock)
{
	return wan_free(lock);
}
/*
**
*/
void wpabs_spin_lock_irqsave(void* lock)
{
	wan_spinlock_t*	SpinLock = (wan_spinlock_t*)lock;

	WAN_ASSERT1(SpinLock == NULL);
	WAN_SPIN_LOCK_IRQSAVE(SpinLock);
}
/*
**
*/
void wpabs_spin_unlock_irqrestore(void* lock)
{
	wan_spinlock_t*	SpinLock = (wan_spinlock_t*)lock;

	WAN_ASSERT1(SpinLock == NULL);
	WAN_SPIN_UNLOCK_IRQSAVE(SpinLock);
}

/*
**
*/
void wpabs_spin_lock_init(void* lock)
{
	wan_spinlock_t*	SpinLock = (wan_spinlock_t*)lock;

	WAN_ASSERT1(SpinLock == NULL);
	WAN_SPIN_LOCK_INIT(SpinLock);

}

/*
**
*/

void wpabs_debug_event(const char * fmt, ...)
{
#ifdef WAN_DEBUG_EVENT
	va_list args;
	char buf[1024];
	va_start(args, fmt);
	vsnprintf(buf, sizeof(buf), fmt, args);
	DEBUG_EVENT("%s", buf);
	va_end(args);
#endif
}

/*
**
*/
void wpabs_debug_cfg(const char * fmt, ...)
{
#ifdef WAN_DEBUG_CFG
	va_list args;
	char buf[1024];
	va_start(args, fmt);
	vsnprintf(buf, sizeof(buf), fmt, args);
	DEBUG_CFG("%s",buf);
	va_end(args);
#endif
}

/*
**
*/
void wpabs_debug_init(const char * fmt, ...)
{
#ifdef WAN_DEBUG_INIT_VAR
	va_list args;
	char buf[1024];
	va_start(args, fmt);
	vsnprintf(buf, sizeof(buf), fmt, args);
	DEBUG_INIT("%s",buf);
	va_end(args);
#endif
}

/*
**
*/
void wpabs_debug_tx(const char * fmt, ...)
{
#ifdef WAN_DEBUG_TX
	va_list args;
	char buf[1024];
	va_start(args, fmt);
	vsnprintf(buf, sizeof(buf), fmt, args);
	DEBUG_TX("%s",buf);
	va_end(args);
#endif
}

/*
**
*/
void wpabs_debug_rx(const char * fmt, ...)
{
#ifdef WAN_DEBUG_RX
	va_list args;
	char buf[1024];
	va_start(args, fmt);
	vsnprintf(buf, sizeof(buf), fmt, args);
	DEBUG_RX("%s",buf);
	va_end(args);
#endif
}

/*
**
*/
void wpabs_debug_isr(const char * fmt, ...)
{
#ifdef WAN_DEBUG_ISR
	va_list args;
	char buf[1024];
	va_start(args, fmt);
	vsnprintf(buf, sizeof(buf), fmt, args);
	DEBUG_ISR("%s",buf);
	va_end(args);
#endif
}
/*
**
*/
void wpabs_debug_timer(const char * fmt, ...)
{
#ifdef WAN_DEBUG_TIMER
	va_list args;
	char buf[1024];
	va_start(args, fmt);
	vsnprintf(buf, sizeof(buf), fmt, args);
	DEBUG_TIMER("%s",buf);
	va_end(args);
#endif
}
/*
**
*/
int wpabs_set_bit(int bit, void *ptr)
{
	set_bit(bit,ptr);
	return 0;
}

/*
**
*/
int wpabs_test_and_set_bit(int bit, void *ptr)
{
#if defined(__LINUX__)
	return test_and_set_bit(bit,ptr);
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
	if (test_bit(bit,ptr)){
		return 1;
	}
	set_bit(bit,ptr);
	return 0;
#else
# error "Error: wpabs_test_and_set_bit() not implemented!"
#endif

}


/*
**
*/
int wpabs_test_bit(int bit, void *ptr)
{
	return test_bit(bit,ptr);
}
/*
**
*/
int wpabs_clear_bit(int bit, void *ptr)
{
	clear_bit(bit,ptr);
	return 0;
}

unsigned long wpabs_get_systemticks(void)
{
	return SYSTEM_TICKS;
}

unsigned long wpabs_get_hz(void)
{
	return HZ;
}

unsigned long wpabs_htons(unsigned long data)
{
	return htons(data);
}

void * wpabs_tasklet_alloc(void)
{
	wan_tasklet_t*	task = NULL;
	task = (wan_tasklet_t*)wan_malloc(sizeof(wan_tasklet_t));
	task->running = 0;
	return (void*)task;
}

void wpabs_tasklet_init(void *task_ptr, int priority, void* func, void* arg)
{
	wan_tasklet_t*		task = (wan_tasklet_t *)task_ptr;
	wan_tasklet_func_t	task_func = (wan_tasklet_func_t)func;

	WAN_ASSERT1(task == NULL);
	WAN_TASKLET_INIT(task, priority, task_func, arg);
}


void wpabs_tasklet_schedule(void *task_ptr)
{
	wan_tasklet_t *task = (wan_tasklet_t*)task_ptr;
	
	WAN_ASSERT1(task == NULL);
	WAN_TASKLET_SCHEDULE(task);
}

void wpabs_tasklet_end(void *task_ptr)
{
	wan_tasklet_t *task = (wan_tasklet_t*)task_ptr;			

	WAN_ASSERT1(task == NULL);
	WAN_TASKLET_END(task);
}

void wpabs_tasklet_kill(void *task_ptr)
{
	wan_tasklet_t *task = (wan_tasklet_t*)task_ptr;			

	WAN_ASSERT1(task == NULL);
	WAN_TASKLET_KILL(task);
}

void *wpabs_taskq_alloc(void)
{
	return wan_malloc(sizeof(wan_taskq_t));
}

void wpabs_taskq_init(void *tq_ptr, void *func, void *data)
{
	wan_taskq_t*		tq = (wan_taskq_t *)tq_ptr;
	wan_taskq_func_t	tq_func = (wan_taskq_func_t)func;

	WAN_ASSERT1(tq==NULL);
	
	WAN_TASKQ_INIT(tq, 0, tq_func, data);
}

/*+F*************************************************************************
 * Function:
 *   wpabs_taskq_schedule_event
 *
 * Description:
 *   Schedule an event to be processed by the bottom half handler.
 *-F*************************************************************************/
void wpabs_taskq_schedule_event(unsigned int bit, unsigned long *event, void *tq_ptr)
{
	wan_taskq_t *tq = (wan_taskq_t *)tq_ptr;

	WAN_ASSERT1(tq==NULL)
	DEBUG_TX ("wpabs_wan_schedule_event: scheduling task\n");
    	set_bit(bit, event);
	WAN_TASKQ_SCHEDULE(tq);
}

void* wpabs_memset(void *b, int c, int len)
{
	return memset(b,c,len);
}

void* wpabs_memcpy(void *b, void *c, int len)
{
	return memcpy(b,c,len);
}


void wpabs_debug_print_skb(void *skb_ptr, char dir)
{
#ifdef WAN_DEBUG_TX
	netskb_t*	skb = (netskb_t*)skb_ptr;
	unsigned char*	data = NULL;
	int i;

	DEBUG_TX("%s SKB: Len=%i : ",dir?"TX":"RX",wan_skb_len(skb));
	data = wan_skb_data(skb);
	for (i=0; i < wan_skb_len(skb); i++){
		_DEBUG_TX("%02X ", data[i]);
	}
	
	_DEBUG_TX("\n");
#endif
}

/* 
 * ============================================================================
 * Get WAN device state.
 */
void wpabs_set_baud (void* card_id, unsigned int baud)
{
	wanpipe_set_baud(card_id, baud);
}


/* 
 * ============================================================================
 * Set WAN device state.
 */
void wpabs_set_state (void* card_id, int state)
{
	wanpipe_set_state(card_id, state);
}
/* 
 * ============================================================================
 * Get WAN device state.
 */
char wpabs_get_state (void* card_id)
{
	return wanpipe_get_state(card_id);
}

unsigned long wpabs_get_ip_addr(void* dev, int option)
{
	return wan_get_ip_addr(dev, option);
}

void wpabs_decode_ipaddr(unsigned long ipaddr, unsigned char *string)
{
	sprintf(string,"%u.%u.%u.%u",NIPQUAD(ipaddr));
}



#define UNKNOWN_PROT	"Unknown Prot"
#define IP_LLC_ATM	"Classic IP (LLC) over ATM"
#define IP_VC_ATM	"IP (VC) over ATM"
#define B_LLC_ETH	"Bridged LLC Ethernet over ATM"
#define B_VC_ETH	"Bridged VC Ethernet over ATM"
#define PPP_LLC_ATM	"PPP (LLC) over ATM"
#define PPP_VC_ATM	"PPP (VC) over ATM"
int wpabs_detect_prot_header(unsigned char *data,int len, char* temp)
{
	int i,cnt=0;

	if (data[0] == 0xAA && data[1] == 0xAA){
		if (data[6] == 0x08){
			if (data[7] == 0x00 || data[7] == 0x06){
				if (temp){
					strcpy(temp, IP_LLC_ATM); 
				}
				return 0;
			}
		}else if (data[6] == 0x00 && data[7] == 0x07){
			if (temp){
				strcpy(temp, B_LLC_ETH); 
			}
			return 0;
		}
		goto detect_unknown;
	}

	if (data[0] == 0xFE && data[1] == 0xFE){
		if (temp){
			strcpy(temp, PPP_LLC_ATM); 
		}
		return 0;
	}
		
	if (data[0] == 0xC0 && data[1] == 0x21){
		if (temp){
			strcpy(temp, PPP_VC_ATM); 
		}
		return 0;
	}

	if (data[0] == 0x45){
		if (temp){
			strcpy(temp, IP_VC_ATM); 
		}
		return 0;		
	}
	
detect_unknown:
	if (temp){
		for (i=0;i<len && i<10;i++){
			cnt+=sprintf(temp+cnt,"%02X ",data[i]);
		}
	}
	return 1;
}

/*+F*************************************************************************
 * Function:
 *   
 *
 * Description:
 *   Limits kernel driver output messages.
 *
 *-F*************************************************************************/
int wpabs_net_ratelimit(void)
{
#ifdef __LINUX__
	return net_ratelimit();
#elif defined(__FreeBSD__) || defined(__OpenBSD__)
	/* Do nothing for FreeBSD/OpenBSD! 
	** For these OS always print all messages.
	*/
	return 1;
#else
# error "wpabs_netrate_limit not supported"
#endif
}


/*+F*************************************************************************
 * Function:
 *   Tracing support functions 
 *
 * Description:
 *
 *
 *-F*************************************************************************/
int wpabs_trace_queue_len(void *trace_ptr)
{
	wan_trace_t*	trace = (wan_trace_t *)trace_ptr;
	wan_skb_queue_t*	trace_queue = NULL;

	WAN_ASSERT(trace == NULL);
	trace_queue = &trace->trace_queue;
	return IFQ_LEN(trace_queue);
}

#define MAX_TRACE_TIMEOUT             (HZ*10)

int wpabs_tracing_enabled(void *trace_ptr)
{
	wan_trace_t *trace_info = (wan_trace_t*)trace_ptr;

	WAN_ASSERT(trace_info == NULL);
	
	if (test_bit(0,&trace_info->tracing_enabled)){
	
		if ((SYSTEM_TICKS - trace_info->trace_timeout) > MAX_TRACE_TIMEOUT){
			DEBUG_EVENT("wanpipe: Disabling trace, timeout!\n");
			clear_bit(0,&trace_info->tracing_enabled);
			clear_bit(1,&trace_info->tracing_enabled);
			return -EINVAL;
		}

		if (wpabs_trace_queue_len(trace_info) < trace_info->max_trace_queue){
			if (test_bit(1,&trace_info->tracing_enabled)){
				return 1;
			}else if (test_bit(2,&trace_info->tracing_enabled)){
				return 2;
			}else{
				return 0;
			}
		}

		if (wpabs_net_ratelimit()){
			DEBUG_EVENT("wanpipe: Warning: trace queue overflow %d (max=%d)!\n",
					wpabs_trace_queue_len(trace_info),
					trace_info->max_trace_queue);
		}
		return -ENOBUFS;
	}
	return -EINVAL;
}

void* wpabs_trace_info_alloc(void)
{
	return wan_malloc(sizeof(wan_trace_t));
}

void wpabs_trace_info_init(void *trace_ptr, int max_trace_queue)
{
	wan_trace_t*	trace = (wan_trace_t *)trace_ptr;
	wan_skb_queue_t*	trace_queue = NULL;

	trace_queue = &trace->trace_queue;
	IFQ_INIT(trace_queue, max_trace_queue);
	trace->trace_timeout	= SYSTEM_TICKS;
	trace->tracing_enabled	= 0;
	trace->max_trace_queue	= max_trace_queue;
}

int wpabs_trace_purge (void *trace_ptr)
{
	wan_trace_t*	trace = (wan_trace_t *)trace_ptr;
	wan_skb_queue_t*	trace_queue = NULL;

	WAN_ASSERT(trace == NULL);
	trace_queue = &trace->trace_queue;
	IFQ_PURGE(trace_queue);
	return 0;
}

int wpabs_trace_enqueue(void *trace_ptr, void *skb_ptr)
{
	wan_trace_t*	trace = (wan_trace_t *)trace_ptr;
	wan_skb_queue_t*	trace_queue = NULL;
	netskb_t*		skb = (netskb_t*)skb_ptr;
	int			err = 0;

	WAN_ASSERT(trace == NULL);
	trace_queue = &trace->trace_queue;
	IFQ_ENQUEUE(trace_queue, skb, NULL, err);
	return err;
}


