/*****************************************************************************
* chdlc_send.c	CHDLC API: Transmit Module
*
* Author(s):	Gideon Hack & Nenad Corbic <ncorbic@sangoma.com>
*
* Copyright:	(c) 1995-1999 Sangoma Technologies Inc.
*
*		This program is free software; you can redistribute it and/or
*		modify it under the terms of the GNU General Public License
*		as published by the Free Software Foundation; either version
*		2 of the License, or (at your option) any later version.
* ============================================================================
*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <net/bpf.h>
#include <net/if.h>
#include <net/wanpipe.h>
#include <net/sdla_chdlc.h>
#include <netinet/in.h>

#define FALSE	0
#define TRUE	1

#define MIN_FRM_LGTH	1000 /*15354*/	
#define MAX_FRM_LGTH  	1024
#define NO_FRMS_TO_TX	5000

unsigned char Tx_data[PRI_MAX_NO_DATA_BYTES_IN_FRAME + sizeof(api_tx_hdr_t)]; 

/* Prototypes */
void open_bpf(void);
int MakeConnection( char * );
void send_socket( void);

int 			desc;
fd_set 			ready;
CMDBLOCK_STRUCT		cmdblock;
pid_t			pid_number;

/*
 * Find and open bpf interface.
 */
void open_bpf(void)
{
    char device[sizeof "/dev/bpf000"];
    int n = 0;

    /* Go through all the minors and find one that isn't in use. */
    do {
	(void)sprintf(device, "/dev/bpf%d", n++);
	desc = open(device, O_WRONLY);
    } while(desc < 0 && errno == EBUSY && n < 4);

    if (desc < 0)
    {
	fprintf(stderr, "Error: All BPF devices are abusy\n");
	exit(0);
    }
    return; 
}

/***************************************************
* MakeConnection
*
*   o Create a Socket
*   o Bind a socket to a wanpipe network interface
*       (Interface name is supplied by the user)
*/         

int MakeConnection (char *i_name ) 
{
	struct ifreq ifr;
	u_int len;

	len = sizeof(Tx_data);
	if (ioctl(desc, BIOCSBLEN, &len) < 0)
 	{
	    fprintf(stderr, "Error: Cannot set buffer length %d\n", len);
	    exit(0);
	}

	strncpy(ifr.ifr_name, i_name, sizeof(ifr.ifr_name));
	if (ioctl(desc, BIOCSETIF, &ifr) < 0)
 	{
	    fprintf(stderr, "Error: Unknown interface name %s (%d)\n", 
					i_name, errno);
	    exit(0);
	}
	printf("Socket bound to %s\n",i_name);
	return( TRUE );
}

/***************************************************
* SEND_SOCKET 
*
*   o Send data throught a socket 
*   o Create a packet using api_tx_element_t data type. 
*   o The transmit packet contains 16 bytes header 
*
*	------------------------------------------
*      |  16 bytes      |        X bytes        ...
*	------------------------------------------
* 	   Header              Data
*
*   o Data structures:
*
*	typedef struct {
*		unsigned char 	attr		PACKED;
*		unsigned char  	reserved[15]	PACKED;
*	} api_tx_hdr_t;
*
*	typedef struct {
*		api_tx_hdr_t 	api_tx_hdr	PACKED;
*		void *		data		PACKED;
*	} api_tx_element_t;
*
*/

void send_socket(void)
{
    unsigned long Tx_count;
    unsigned short Tx_lgth;
    api_tx_element_t* api_tx_el;
    int err;

    Tx_count = 0;
    Tx_lgth = MIN_FRM_LGTH; 

    for(;;) 
    {
        /* Tx_data contains the maximum packet size
         * The transmit packet contains 16 byte header
         * followed by data.
         *
         * Use the api_tx_element_t structure to construct
         * a packet.
         */          
	api_tx_el = (api_tx_element_t *)Tx_data;
 
	/* Fill in the packet data */ 

        *((unsigned char *)&api_tx_el->data) =
                        (unsigned char)(Tx_count & 0xFF);
        *((unsigned char *)&api_tx_el->data + (Tx_lgth - 1)) =
                        (unsigned char)((Tx_count + 1) & 0xFF);

	err = write(desc, Tx_data, Tx_lgth + sizeof(api_tx_hdr_t));

	/* err contains number of bytes transmited */		
	if(err != -1) 
	{
	    Tx_count ++;
       	    Tx_lgth ++;
            if (Tx_lgth > MAX_FRM_LGTH)
       	        Tx_lgth = MIN_FRM_LGTH;

	    /* Print out a transmit packet if you wish */
	    printf("Packet #%ld...ok\n", Tx_count);
	} 
	else 
	{  
	    /* A packet must be retransmited if stack 
	     * failed to send 
	     */
	    /*printf("Stack Full: Retransmitting Packet # %li (err%d)\n", 
				Tx_count, errno);*/
	}
	/* Stop after number of frames */
	if (Tx_count == NO_FRMS_TO_TX)
	    break;
    }
} 

/***************************************************************
 * Main:
 *
 *    o Make a socket connection to the driver.
 *    o Call send_socket() to transmit data
 *
 **************************************************************/   

int main(int argc, char* argv[])
{
    int proceed;
	
    if (argc != 2)
    {
	printf("Usage: tsnd <interface name>\n");
	exit(0);
    }

    open_bpf();
    proceed = MakeConnection(argv[argc-1]);
    pid_number = getpid();
    if (proceed == TRUE)
    {
	printf("Sending a command\n");
	send_socket( );
	return 0;
    }
    close(desc);
    return 1;
};
