/*****************************************************************************
* wanpipemon.c	Sangoma WANPIPE Monitor
*
* Author:       Alex Feldman <al.feldman@sangoma.com>	
* 		Nenad Corbic <ncorbic@sangoma.com>
*
* Copyright:	(c) 1995-2002 Sangoma Technologies Inc.
*
* ----------------------------------------------------------------------------
* Aug 02, 2002  Nenad Corbic	Updated all linux utilites and added 
*                               GUI features.
* May 30, 2002	Alex Feldman	Initial version based on ppipemon
*****************************************************************************/

/******************************************************************************
 * 			INCLUDE FILES					      *
 *****************************************************************************/
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include <signal.h>
#include <string.h>
#include <stdlib.h>
#include <arpa/inet.h>
#if defined(__LINUX__)
# include <linux/version.h>
# include <linux/wanpipe.h>
# include <linux/wanrouter.h>
# include <linux/sdla_fr.h>
# include <linux/if_packet.h>
# include <linux/if_wanpipe.h>
# include <linux/if_ether.h>
#else
# include <netinet/in_systm.h>
# include <netinet/in.h>
# include <netinet/ip.h>
# include <netinet/udp.h>  
# include <net/wanpipe.h>
#endif
#include "fe_lib.h"
#include "wanpipemon.h"

/******************************************************************************
 * 			DEFINES/MACROS					      *
 *****************************************************************************/
#define MDATALEN 2024

#ifndef SIOC_WANPIPE_MONITOR 
#define SIOC_WANPIPE_MONITOR SIOC_WANPIPE_PIPEMON
#endif


/******************************************************************************
 * 			GLOBAL VARIABLES				      *
 *****************************************************************************/
wan_udp_hdr_t			wan_udp;
int				sock = 0;
int 				wan_protocol = 0;
int				protocol_cb_size = sizeof(wan_mgmt_t)+sizeof(wan_cmd_t)+1;

int 				is_508;
int				gui_interface=0;


static sa_family_t		af = AF_INET;
static struct sockaddr_in	soin;
 char ipaddress[16];
int  udp_port = 9000;

short dlci_number;
char if_name[WAN_IFNAME_SZ+1];
int ip_addr = -1;
int start_xml_router=0;
int stop_xml_router=1;
int raw_data=0;
int trace_all_data=0;
int lcn_number=0;
unsigned char par_port_A_byte, par_port_B_byte;

int sys_timestamp=0;
int annexg_trace =0;

trace_prot_t trace_prot_opt[]={ 
	{"FR", FRAME},
	{"LAPB",LAPB},
	{"X25", X25},
	{"",-1},
};

trace_prot_t trace_x25_prot_opt[]={
	{"DATA",DATA},
	{"PROT",PROT},
	{"",-1},
};

/* tracing global variables */
unsigned int TRACE_DLCI=0;
unsigned char TRACE_PROTOCOL=ALL_PROT;
unsigned int TRACE_X25_LCN=0;
unsigned char TRACE_X25_OPT=ALL_X25;
char TRACE_EBCDIC=0;
char TRACE_ASCII=0;
char TRACE_HEX=0;

char *cmd[MAX_CMD_ARG];
FT1_LED_STATUS FT1_LED;

/******************************************************************************
 * 			FUNCTION PROTOTYPES				      *
 *****************************************************************************/
int		main(int, char**);
static int 	init(int, char**,char*);
static void 	usage(void);
static void usage_trace_info(void);
static void 	usage_long(void);
static int  	MakeConnection(void);
static int 	GetWANConfig(void);
#if 0
static void 	sig_handler(int sigint);
#endif
int      	fail=0;
int 		xml_output=0;
//static sa_family_t	get_if_family(char*);


struct fun_protocol function_lookup[] = {
	{ WANCONFIG_FR,		"fr",    FRConfig, FRUsage, FRMain, FRDisableTrace, 
		                         FRget_main_menu, FRget_cmd_menu, 
					 FR_read_FT1_status, FR_set_FT1_mode, 0 }, 
#if defined(__LINUX__)
	{ WANCONFIG_MFR,	"fr",    FRConfig, FRUsage, FRMain, FRDisableTrace, 
		                         FRget_main_menu, FRget_cmd_menu,
					 FR_read_FT1_status, FR_set_FT1_mode,0 },
#else
	{ WANCONFIG_MFR, 	"fr",  NULL, NULL, NULL, NULL,NULL,NULL,0 }, 
#endif
	{ WANCONFIG_PPP,	"ppp",   PPPConfig, PPPUsage, PPPMain, PPPDisableTrace, 
		                         PPPget_main_menu, PPPget_cmd_menu, 
					 PPP_read_FT1_status, PPP_set_FT1_mode,0 }, 
					 
	{ WANCONFIG_CHDLC,	"chdlc", CHDLCConfig, CHDLCUsage, CHDLCMain, CHDLCDisableTrace,
		                         CHDLCget_main_menu, CHDLCget_cmd_menu, 
					 CHDLC_read_FT1_status, CHDLC_set_FT1_mode, 2}, 
					 
	{ WANCONFIG_ATM,	"atm",   ATMConfig, ATMUsage, ATMMain, ATMDisableTrace, 
		                         ATMget_main_menu, ATMget_cmd_menu, 
					 ATM_read_FT1_status, ATM_set_FT1_mode, 2 }, 
					 
	{ WANCONFIG_ADSL,	"adsl",  ADSLConfig, ADSLUsage, ADSLMain, ADSLDisableTrace, 
		                         ADSLget_main_menu, ADSLget_cmd_menu, 
					 NULL, NULL, 2 }, 

#if defined(__LINUX__)
	{ WANCONFIG_SS7,	"ss7",   SS7Config, SS7Usage, SS7Main, SS7DisableTrace, 
		                         SS7get_main_menu, SS7get_cmd_menu, 
					 NULL, NULL, 2 },
	{ WANCONFIG_X25, 	"x25",   X25Config, X25Usage, X25Main, X25DisableTrace, 
					 X25get_main_menu, X25get_cmd_menu , 
					 NULL, NULL, 0}, 
#else
	{ WANCONFIG_SS7, 	"ss7",  NULL, NULL, NULL, NULL,NULL,NULL,0 }, 
	{ WANCONFIG_X25, 	"x25",  NULL, NULL, NULL, NULL,NULL,NULL,0 }, 
#endif
	{ WANCONFIG_BSC, 	"bsc",  NULL, NULL, NULL, NULL,NULL,NULL,0 }, 
	{ WANCONFIG_HDLC, 	"hdlc", NULL, NULL, NULL, NULL,NULL,NULL,0 }, 
	{ 0, "N/A", 0, 0, 0, 0,0 }, 
};

/******************************************************************************
 * 			FUNCTION DEFINITION				      *
 *****************************************************************************/
static int MakeConnection( void )
{
	if (ip_addr == WAN_TRUE){

		sock = socket((af == AF_INET) ? PF_INET : PF_INET6, SOCK_DGRAM, 0);
		if (sock < 0){
			perror("socket");
			return WAN_FALSE;
		} /* if */

		if (af == AF_INET){
			soin.sin_addr.s_addr = inet_addr(ipaddress);
			soin.sin_family = AF_INET;
			soin.sin_port = htons((u_short)udp_port);
			if (soin.sin_addr.s_addr < 0){
	                	printf("Error: Invalid IP address!\n");
        	        	return WAN_FALSE;
        		}
			if (connect( sock, (struct sockaddr *)&soin, sizeof(soin)) != 0){
				perror("connect");	
				printf("Make sure the IP address is correct\n");
				return( WAN_FALSE );
			} /* if */
     		}else{
#ifdef ENABLE_IPV6
			inet_pton(AF_INET6, ipaddress, &soin6.sin6_addr);
			//soin6.sin6_len = sizeof(struct sockaddr_in6);
			soin6.sin6_family = AF_INET6;
			soin6.sin6_scope_id = 0xc;
			soin6.sin6_port = htons((u_short)udp_port);
			if (connect( sock, (struct sockaddr *)&soin6, sizeof(soin6)) != 0){
				perror("connect");
				printf("Make sure the IP address is correct\n");
				return WAN_FALSE;
			} /* if */
#else
			printf("Ipv6 protocol not supported!\n");
			return WAN_FALSE;
#endif

		}
	}else{
	    	sock = socket((af == AF_INET) ? PF_INET : PF_INET6, SOCK_STREAM, IPPROTO_IP);
	    	if (sock < 0){
			perror("socket");
			return WAN_FALSE;
    		}

	}

	if (!GetWANConfig()){
		printf("Error: Unable to obtain network interface information.\n");
		printf("Make sure the IP and UDP port are correct.\n");
		close(sock);
		return WAN_FALSE;
	}   

	return WAN_TRUE;
} /* MakeConnection */

static int GetWANConfig( void )
{
	int cnt = 0, err = 0;

	if (wan_protocol != 0){

		/* If the user specified protocol, do not try to
	 	* autodetect it.  Old drivers do not have protocol
	 	* detect option. */
		goto skip_protocol_detect;
	}
	
	/* Get Protocol type */ 
	wan_udp.wan_udphdr_command	= WAN_GET_PROTOCOL;
	wan_udp.wan_udphdr_data_len 	= 0;
	wan_udp.wan_udphdr_return_code 	= WAN_UDP_TIMEOUT_CMD;
	while (DO_COMMAND(wan_udp) < 0 && ++cnt < 4) {
		if (wan_udp.wan_udphdr_return_code == WAN_UDP_TIMEOUT_CMD) {
			printf("Error: Command timeout occurred\n"); 
			return WAN_FALSE;
		}
		if (wan_udp.wan_udphdr_return_code == 0xCC ) 
			return WAN_FALSE;
	}

	if (cnt >= 4){
		return WAN_FALSE;
	}


	wan_protocol = wan_udp.wan_udphdr_data[0];


skip_protocol_detect:
	printf("Network interface running %s protocol...\n",
				DECODE_PROT(wan_protocol));

	EXEC_PROT_FUNC(config,wan_protocol,err,());
	return err;
} 

int DoCommand(wan_udp_hdr_t* wan_udp)
{
        static unsigned char id = 0;
        fd_set ready;
        struct timeval to;
	int data_len = protocol_cb_size+wan_udp->wan_udphdr_data_len;
        int x, err=0;

	if (ip_addr == WAN_TRUE){
		for( x = 0; x < 4; x += 1 ) {
			wan_udp->wan_udphdr_request_reply = 0x01;
			wan_udp->wan_udphdr_id = id;
			/* 0xAA is our special return code indicating packet timeout */
			//wan_udp.wan_udphdr_return_code = WAN_UDP_TIMEOUT_CMD;
			//err = send(sock, &wan_udp. 35 + cb.cblock.buffer_length, 0);
			err = send(sock, (caddr_t)wan_udp, data_len, 0); 
			if (err < 0){
				perror("Send: ");	
				continue;
			}
			
			FD_ZERO(&ready);
			FD_SET(sock,&ready);
			to.tv_sec = 5;
			to.tv_usec = 0;
	
			if((err = select(sock + 1,&ready, NULL, NULL, &to))){
				err = recv(sock, (caddr_t)wan_udp, sizeof(wan_udp_hdr_t), 0);
				if (err < 0){
					perror("Recv: ");
				}
        	               	break;
	               	}else{
				printf("Error: No Data received from the connected socket.\n");
               		}
		}
	}else{
		struct ifreq ifr;
		wan_udp->wan_udphdr_request_reply = 0x01;
		wan_udp->wan_udphdr_id = id;
		strncpy(ifr.ifr_name, if_name, sizeof(ifr.ifr_name));
		ifr.ifr_data = (caddr_t)wan_udp;

        	if ((err=ioctl(sock, SIOC_WANPIPE_MONITOR , &ifr)) < 0){
			perror("Ioctl: ");
			return err;
    		}
	}

#if 0
	{
		int i;
		printf("Rx Packet :\n");
		for (i=0;i<sizeof(wan_udp);i++){
			printf("%x ",*((unsigned char *)wan_udp+i));
		}
		printf("\n");
	}
#endif
        return err;

} /* DoCommand */

static int init(int argc, char *argv[], char* command)
{
	int i = 0, i_cnt = 0, u_cnt = 0, c_cnt = 0, d_cnt = 0;
	struct in_addr *ip_str = NULL;
#ifdef ENABLE_IPV6
	struct in6_addr ip6_str;
#endif

	for (i = 0; i < argc; i++){

		
		if (!strcmp(argv[i],"-h")){
			usage_long();
			return WAN_FALSE;

		}else if (!strcmp(argv[i],"-traceinfo")){
			usage_trace_info();
			return WAN_FALSE;
			
		}else if (!strcmp(argv[i],"-i")){

			if (i+1 > argc-1){
				printf("ERROR: Invalid IP or Interface Name!\n");
				printf("Type wanpipemon <cr> for help\n");
				return WAN_FALSE;
			}

			strcpy(ipaddress,argv[i+1]);
			if (inet_aton(ipaddress,ip_str) != 0 ){
				ip_addr = WAN_TRUE;
			}else{
#ifdef ENABLE_IPV6
				if (inet_pton(AF_INET6,ipaddress,(void*)&ip6_str) == 1){
					af = AF_INET6;
					ip_addr = WAN_TRUE;
				}else{		
					ip_addr = WAN_FALSE;
					strncpy(if_name, ipaddress, WAN_IFNAME_SZ);
				}
#else
				ip_addr = WAN_FALSE;
				strncpy(if_name, ipaddress, WAN_IFNAME_SZ);
#endif
			}
			i_cnt=1;

		}else if (!strcmp(argv[i],"-g")){
#ifdef WANPIPEMON_GUI			
			gui_interface=1;
			return WAN_TRUE;
#else
			printf("wanpipemon: Warning GUI interface not compiled in!\n\n");
#endif			
		}else if (!strcmp(argv[i],"-u")){

			if (i+1 > argc-1){
				printf("ERROR: Invalid UDP PORT!\n");
				printf("Type wanpipemon <cr> for help\n");
				return WAN_FALSE;
			}

		 	if(isdigit(argv[i+1][0])){
				udp_port = atoi(argv[i+1]);
			}else{
				printf("ERROR: Invalid UDP Port!\n");
				printf("Type wanpipemon <cr> for help\n");
				return WAN_FALSE;
			}
			u_cnt=1;
		}else if (!strcmp(argv[i],"-c")){

			if (i+1 > argc-1){
				printf("ERROR: Invalid Command!\n");
				printf("Type wanpipemon <cr> for help\n");
				return WAN_FALSE;
			}

			strcpy(command,argv[i+1]);
			c_cnt=1;
		}else if (!strcmp(argv[i], "-d")){

			if (i+1 > argc-1){
				printf("ERROR: Invalid Command!\n");
				printf("Type wanpipemon <cr> for help\n\n");
				return WAN_FALSE;
			}
			if (isdigit(argv[i+1][0])){
				dlci_number = atoi(argv[i+1]);
				TRACE_DLCI=dlci_number;
			}else{
				printf("ERROR: DLCI must be an integer!\n");
				printf("Type wanpipemon <cr> for help\n");
				return WAN_FALSE;
			} 
			d_cnt=1;
		
		}else if (!strcmp(argv[i],"-pv6")){
			af = AF_INET6;	// IPV6

		}else if (!strcmp(argv[i],"-p")){

			if (i+1 > argc-1){
				printf("ERROR: Invalid Command!\n");
				printf("Type wanpipemon <cr> for help\n");
				return WAN_FALSE;
			}

			if (argc == 3){
				unsigned char *name=argv[i+1];
				EXEC_NAME_FUNC(usage,name,());
				return WAN_FALSE;
			}

			if (strcmp(argv[i+1], "chdlc") == 0){
				strcpy(wan_udp.wan_udphdr_signature, UDP_CHDLC_SIGNATURE);
				wan_protocol=WANCONFIG_CHDLC;
			}else if (strcmp(argv[i+1], "fr") == 0){
				strcpy(wan_udp.wan_udphdr_signature, UDP_FR_SIGNATURE);
				wan_protocol=WANCONFIG_FR;
			}else if (strcmp(argv[i+1], "ppp") == 0){
				strcpy(wan_udp.wan_udphdr_signature, UDP_PPP_SIGNATURE);
				wan_protocol=WANCONFIG_PPP;
			}else if (strcmp(argv[i+1], "x25") == 0){
				strcpy(wan_udp.wan_udphdr_signature, UDP_X25_SIGNATURE);
				wan_protocol=WANCONFIG_X25;
			}else if (strcmp(argv[i+1], "adsl") == 0){
				strcpy(wan_udp.wan_udphdr_signature, GLOBAL_UDP_SIGNATURE);
				wan_protocol=WANCONFIG_ADSL;
			}else{
				usage();
				return WAN_FALSE;
			}		
		
		}else if (!strcmp(argv[i], "-x")){
			xml_output=1;

		}else if (!strcmp(argv[i], "-full")){
			trace_all_data=1;
		
		}else if (!strcmp(argv[i], "-systime")){
			sys_timestamp=1;

		}else if (!strcmp(argv[i], "-prot")){
			int x;
			if (i+1 > argc-1){
				printf("ERROR: Invalid Command -prot, Type fpipemon <cr> for help\n\n");
				exit(1);
			}
			TRACE_PROTOCOL=0;
			for (x=0;;x++){
				if (trace_prot_opt[x].prot_index == -1)
					break;
			
				if (strstr(argv[i+1],trace_prot_opt[x].prot_name) != NULL){
					TRACE_PROTOCOL|=trace_prot_opt[x].prot_index;
				}
			}

			if (!TRACE_PROTOCOL){
				printf("ERROR: Invalid Protocol in Command -prot, Type fpipemon <cr> for help\n\n");
				exit(1);
			}
		}else if (!strcmp(argv[i], "-lcn")){
			if (i+1 > argc-1){
				printf("ERROR: Invalid Command -lcn, Type fpipemon <cr> for help\n\n");
				exit(1);
			}

			if (isdigit(argv[i+1][0])){
				TRACE_X25_LCN=atoi(argv[i+1]);
			}else{
				printf("ERROR: LCN must be an integer,Type fpipemon <cr> for help\n\n");
				exit(0);
			} 

		}else if (!strcmp(argv[i], "-ascii")){

			TRACE_ASCII=1;
			
		}else if (!strcmp(argv[i], "-ebcdic")){

			TRACE_EBCDIC=1;

		}else if (!strcmp(argv[i], "-hex")){

			TRACE_HEX=1;
			
		}else if (!strcmp(argv[i], "-x25opt")){
			int x;
			if (i+1 > argc-1){
				printf("ERROR: Invalid Command -x25prot, Type fpipemon <cr> for help\n\n");
				exit(1);
			}
		
			TRACE_X25_OPT=0;
			for (x=0;;x++){
				if (trace_x25_prot_opt[x].prot_index == -1)
					break;
				
				if (strstr(argv[i+1],trace_x25_prot_opt[x].prot_name) != NULL){
					TRACE_X25_OPT|=trace_x25_prot_opt[x].prot_index;
				}
			}

			if (!TRACE_X25_OPT){
				printf("ERROR: Invalid X25 Option in Command -x25opt, Type fpipemon <cr> for help\n\n");
				exit(1);
			}
		}
	}

	if (!i_cnt){
		printf("ERROR: No IP address or Interface Name!\n");
		printf("Type wanpipemon <cr> for help\n");
		return WAN_FALSE;
	}
	if (!u_cnt && ip_addr == WAN_TRUE){
		printf("ERROR: No UDP Port!\n");
		printf("Type wanpipemon <cr> for help\n");
		return WAN_FALSE;
	}
	if (!c_cnt){
		printf("ERROR: No Command!\n");
		printf("Type wanpipemon <cr> for help\n");
		return WAN_FALSE;
	}
	if (!d_cnt){
		/* Default DLCI Number to 0 */
		dlci_number = 0;
	}

	return WAN_TRUE;
}

void ResetWanUdp(wan_udp_hdr_t *wan)
{
	memset((void *)&wan->wan_cmd, 0, sizeof(wan_cmd_t));
} 

void banner (char *title,int dlci_number){
	
	int len,i;
	
	len = strlen(title);
	printf("\n\t");
	for (i=0;i<(len+16);i++)
		printf("-");
	printf("\n\t\t%s",title);
	printf("\n\t");
	for (i=0;i<(len+16);i++)
		printf("-");
	printf("\n\n");

}


static unsigned char wanpipemon_usage[]="
Wanpipemon Verison 1.0
Copyright (C) 2002 - Sangoma Technologies
www.sangoma.com
This may be freely redistributed under the terms of the GNU GPL
	
Usage: 	(Options in {} brackets are optional)

	#Command prompt Local debugging
	wanpipemon -i <interface> -c <command> { -p <protocol> }
	
	#Command prompt Remote debugging
	wanpipemon -i <remote IP> -u <port> -c <command> { -p <protocol> }

	#GUI interface
	wanpipemon -g { -i <interface|IP> -u <port> -p <protocol> }  
	
	#Display usage for each protocol
	wanpipemon -p [fr|ppp|chdlc|x25|adsl]			

	#Detailed usage for each option
	wanpipemon -h

	#Detailed usage for advanced tracing
	wanpipemon -traceinfo
";

static void usage(void)
{

	printf("%s\n",wanpipemon_usage);

}

static unsigned char usage_long_buf[]= "

Wanpipemon Verison 1.0
Copyright (C) 2002 - Sangoma Technologies
www.sangoma.com
This may be freely redistributed under the terms of the GNU GPL

Usage: 	(Options in {} brackets are optional)

	#Command prompt Local debugging
	wanpipemon -i <interface> -c <command> { -p <protocol> }
	
	#Command prompt Remote debugging
	wanpipemon -i <remote IP> -u <port> -c <command> { -p <protocol> }

	#GUI interface
	wanpipemon -g { -i <interface|IP> -u <port> -p <protocol> }  
	
	#Display usage for each protocol
	wanpipemon -p [fr|ppp|chdlc|x25|adsl]			

	#Detailed usage for each option
	wanpipemon -h

-----------------------------------------------------------
Option -p: 

    wanpipemon -p [ppp|fr|chdlc|x25|adsl]

	When used by itself:
		Displays the command list for each protocol

    wanpipemon -p fr -i wp1_fr16 -c xm	
	
	When used with other commands:
		Specifies the protocol and skips the protocol
		autodetect feature.  
		
		Used for backward compatibility, since
		old drivers do not support protocol detect option.

-----------------------------------------------------------		
Option 	-g:

   Start wanpipemon in GUI mode

   wanpipemon -g
   
	For local debugging: 
		GUI will display list of active interfaces.
			
	For remote debugging:
		GUI will ask for an IP address and UDP port.
	
   wanpipemon -i wp1_fr16 -g
   
   	Specifies the interface, do not show the list of interfaces

   wanpipemon -p fr -i wp1_fr16 -g
        
   	Specify the interface and protocol. 
	
-----------------------------------------------------------
Option -i: 

   wanpipemon -i wp1_fr16
   wanpipemon -i 192.168.1.1

	Specify the Local interface name or Remote IP
	address which will be used to connect to the driver.
   
	
-----------------------------------------------------------
Option -u: 

   wanpiemon -i 192.168.1.1 -u 9000 -c xm

   	Wanpipe UDPPORT specified in /etc/wanpipe/wanpipe?.conf
	Note: This option is only valid when using Remote IP address.


-----------------------------------------------------------
Option -c: 

	Wanpipe monitor command (for list of command per protocol)
	For list of commands run: 
		wanpipemon -p [fr|ppp|chdlc|x25|adsl]


";


static void usage_long(void)
{
	printf("%s\n",usage_long_buf);
}

static unsigned char wanpipemon_trace_info[]="

Wanpipemon Verison 1.0
Copyright (C) 2002 - Sangoma Technologies
www.sangoma.com
This may be freely redistributed under the terms of the GNU GPL
	
Advanced Line Trace Options
===========================

Advanced trace options include protocol decoding
and parsing, system timestamping as well as
packet filtering.

The Advanced trace command is: 'tra'
	
	wanpipemon -i <ifname> -c tra { trace options }


trace options:
-------------	

	-prot 	[FR|LAPB|X25]	#Filter packets based on
				#protocol. Multiple protocols can
	                        #be selected:  
				# <prot>-<prot>...
			        # eg: -prot LAPB-X25
				#Default: All frames
				

	-x25opt [DATA|PROT]	#Filter x25 packets based on 
				#protcol or information frames
				#Default: All frames

	-lcn	<number>	#Filter x25 packets based on
	                        #specific lcn number
	                        #Default: All lcns
	
	-hex			#Display packet info in HEX 
                                #Default: Hex
	
	-ascii			#Display packet info in ASCII
	
	-ebcdic			#Display packet info in EBCDIC	

	-systime		#Display timestamp as system time
	                        #instead of absolute number
	
	-full			#Display packet data in full.

Examples:
---------

#Trace and decode all frames, and display packets
#in full with timestampe decoded into system time.

  wanpipemon -i wp1mp -c tra -full -systime

#Trace LAPB and X25 protocol frames. Futhtermore, 
#only decode x25 frames with LCN=1

  wanpipemon -i wp1mp -c tra -prot LAPB-X25 -lcn 1

#Trace X25 protocol frames and display x25 data
#in ASCII.

  wanpipemon -i wp1mp -c tra -prot X25 -ascii -full -systime 
	
";

static void usage_trace_info(void)
{
	printf("%s\n",wanpipemon_trace_info);
}

#if 0
static void sig_handler(int sigint)
{
	int err;
	EXEC_PROT_FUNC(disable_trace,wan_protocol,err,());
}
#endif


int main(int argc, char* argv[])
{
	char command[5];

	strcpy(wan_udp.wan_udphdr_signature, GLOBAL_UDP_SIGNATURE);
	
  	printf("\n");
   	if (argc >= 2){
		int err=0;
    
		if (init(argc, argv, command) == WAN_FALSE){
			return -EINVAL;		
		}

#ifdef WANPIPEMON_GUI
		if (gui_interface && ip_addr==-1){
			if (wan_main_gui() == WAN_FALSE){
				return -EINVAL;
			}
		}
#endif
gui_loop:		
		if (MakeConnection() == WAN_FALSE){ 
			close(sock);
			return -ENODEV;	
		}

#ifdef WANPIPEMON_GUI
		if (gui_interface && wan_protocol != 0){
			if (wan_main_gui() == WAN_FALSE){
				return 0;
			}
			wan_protocol=0;
			goto gui_loop;		
		}
#endif
		EXEC_PROT_FUNC(main,wan_protocol,err,(command,argc,argv));
     		close( sock );
   	}else{
    		usage();
   	}
  
	printf("\n");
   	return 0;
}


